#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs

# Python built-ins
from builtins import range
from difflib import SequenceMatcher as sm
import warnings
from copy import deepcopy
import timeit

# Sortal Library Imports
from sortal import (transformSet, rule, primitiveSort, point3D, point2D, coordinate, lineSegment3D, planeSegment2D, 
                    lineSegment2D, line3D, label, rWeight, disjunctiveSort, attributeSort, affineTransform,  
                    vector3D, vector2D, planeSegment3D, description, color, label, enumerative, sort, key, sdlParser,
                    compoundSort, disjunctiveSort)
                    
# User-defined sort type imports from sort_types
from ..sortal_api.setup.sortTypes import (pointSort2D, pointSort3D, pointSortP2D, pointSortP3D, 
                                          lineSegSort2D, lineSegSort3D, lineSegSortP2D, lineSegSortP3D,
                                          planeSegSort2D, planeSegSort3D, planeSegSortP2D, planeSegSortP3D, 
                                          c2D, c3D, cP2D, cP3D,
                                          e2D, e3D, eP2D, eP3D,
                                          cArc2D, cArc3D, eArc2D, eArc3D, 
                                          bez2D, bezP2D, bez3D, bezP3D,
                                          ptDSort, lsDSort, plDSort, labelSort, cDSort, cArcDSort, eDSort, eArcDSort, bezDSort,
                                          weightSort, weight255Sort, colorSort, descSort, enumSort)

"""
==========
CLASS INFO
==========
'convertSort' is a class of functions that converts a sort type data structure's geometric individuals to the target parametric-dimension combination.
Any attributes attached to the geometry or standalone external descriptions already present in the sort type, if it is a disjunctive sort, are also
carried over to the new sort type.
"""
                               
class convertSort(object):
    # Creates new sortal structure based on target parametric-dimension combination
    def __init__(self, structure = None, target = None, newSortName = ''):
        self.origSort = structure
        self.type = structure.type                      # Used to determine complexity level: primitive/attribute/disjunctive/compound
        self.target = target                            # Target dimension-parametric combination
        self.parametric = False
        self.dimension = '3D'
        self.newSort = None
        self.npDict = {'lineSeg-N2D': lineSegSort2D, 'lineSeg-N3D': lineSegSort3D, 'point-N2D': pointSort2D, 'point-N3D': pointSort3D, 'planeSeg-N2D': planeSegSort2D, 'planeSeg-N3D': planeSegSort3D, 'circle-N2D': c2D, 'circle-N3D': c3D, 'ellipse-N2D': e2D, 'ellipse-N3D': e3D, 'cArc-N2D': cArc2D, 'cArc3D': cArc3D, 'eArc-N2D': eArc2D, 'eArc-N3D': eArc3D, 'bezier-N2D': bez2D, 'bezier-N3D': bez3D}
        self.pDict = {'lineSeg-P2D': lineSegSortP2D, 'lineSeg-P3D': lineSegSortP3D, 'point-P2D': pointSortP2D, 'point-P3D': pointSortP3D, 'planeSeg-P2D': planeSegSortP2D, 'planeSeg-P3D': planeSegSortP3D, 'circle-P2D': cP2D, 'circle-P3D': cP3D, 'ellipse-P2D': eP2D, 'ellipse-P3D': eP3D, 'bezier-P2D': bezP2D, 'bezier-P3D': bezP3D}
        self.newSortName = newSortName
        
        # Setting of parametric combination
        if 'P' in target:
            self.parametric = True
            
        if '2' in target:
            self.dimension = '2D'
        
        # Entered if form is a metaform/has a disjunctive sort, i.e. contains both points and line segments
        # Structure of metaform: <metaform name>.elements contains one form of line segments, one form of points
        # To access each individual sort -> since the list stored in the metaform sort
        # is constructed as [<point form>, <line segment form>]
        # i.e. <metaform>.component -> returns list of component sorts
        if self.type == 'sort-disjunctive':
            for component in self.origSort.component:
                if target in component.baseSort().characteristic.type:
                    warnings.warn('Current sort type of sortal structure already matches the target dimension-parametric combination')
                    self.newSort = self.origSort
                    break
                    
        # Entered if form is of an attribute sort, i.e. a geometric primitive sort with attributes as its weight
        # Structure of attribute sort: <geometric base sort> ^ <disjunctive sort with attribute sorts>
        elif self.type == 'sort-attribute':
            if target in self.origSort.baseSort().characteristic.type:
                warnings.warn('Current sort type of sortal structure already matches the target dimension-parametric combination')
                self.newSort = self.origSort
    
        return
    
    # Returns the new sortal structure
    def sortType(self):
        if self.newSort == None:
            self.newSort = self.assembleNewSort(self.origSort)
        
        if self.newSort.type == 'sort-disjunctive':
            # Adds newly assessed sort into 'rhino_shapes_c'
            components_list = [self.newSort]
            compound_sort = deepcopy(sort.register['rhino_shapes_c'])
            del sort.register['rhino_shapes_c']
            for sortType in compound_sort.component:
                components_list.append(sortType)
                
            compoundSort('rhino_shapes_c', components_list)

        return self.newSort
    
    # Decomposes the old sort
    # Constructs the new sort
    def assembleNewSort(self, mainSort):
        if 'sort' not in mainSort.type:
            mainSort = mainSort.sort

        name = mainSort.name
        components_list = []

        # Entered if form is a metaform, i.e. contains both points and line segments
        # Structure of metaform: <metaform name>.elements contains one form of line segments, one form of points
        # To access each individual point or line segment -> since the list stored in the metaform 
        # is constructed as [<point form>, <line segment form>]
        # i.e. <metaform>.name.elements[0].elements -> returns list of points
        # <metaform>.name.elements[1].elements -> returns list of line segments
        if mainSort.type == 'sort-disjunctive':
            components = mainSort.component
            for component_sort in components:

                # Identifies geometric sorts
                # Replaces them with corresponding target dimension-parametric geometric sort
                if component_sort.type == 'sort-attribute' or (component_sort.type == 'sort-primitive' and 'description' not in component_sort.baseSort().characteristic.type):
                    if self.constructMainSort(component_sort) != None:
                        components_list.append(self.constructMainSort(component_sort))
                    """
                    elif component_sort.type == 'sort-disjunctive':
                        sublist = []
                        components_list = {}
                        subname = ''
                        
                        if 'P' in component_sort.component[0][-3]:
                            subname += 'P'
                        
                        if '3' in component_sort.component[0][-2]:
                            subname += '3D'
                        else:
                            subname += '2D'
                            
                        for component in component_sort.component:
                            if 'description' not in component_sort.baseSort().characteristic.type:
                                if self.constructMainSort(component_sort) != None:
                                    sublist.append(self.constructMainSort(component_sort))
                            else:
                                sublist.append(component_sort.baseSort())
                                
                        components_list[subname] = sublist
                    """
                elif component_sort.type == 'sort-primitive' and 'description' in component_sort.baseSort().characteristic.type:
                    components_list.append(component_sort)
        
        elif mainSort.type == 'sort-attribute' or mainSort.type == 'sort-primitive':
            newSort = self.constructMainSort(mainSort)

        if mainSort.type == 'sort-disjunctive':
            if disjunctiveSort.sameComponent(components_list) is False:
                disjName = ''
                
                if self.parametric:
                    disjName += 'P'
                else:
                    disjName += 'N'
                    
                disjName += self.dimension
                disjName = self.newSortName + '-' + disjName
                newSort = disjunctiveSort(disjName, components_list)
                
            else:
                newSort = disjunctiveSort.sameComponent(components_list)
        """
        elif mainSort.type == 'sort-compound':
            disjSorts = []
            
            for disjName in components_list:
                disjComponents = components_list[disjName]
                
                if disjunctiveSort.sameComponent(disjComponents) is False:
                    disjSort = disjunctiveSort(disjName, disjComponents)
                else:
                    disjSort = disjunctiveSort.sameComponent(componentsList)
                
                disjSorts.append(disjSort)
            """  
        return newSort
    
    def constructMainSort(self, sortType):
        sortTypeBase = sortType.baseSort().characteristic.type

        if 'lineSegment' in sortTypeBase:
            baseName = 'lineSeg'
        elif 'point' in sortTypeBase:
            baseName = 'point'
        elif 'planeSegment' in sortTypeBase:
            baseName = 'planeSeg'
        elif 'circle' in sortTypeBase:
            baseName = 'circle'
        elif 'ellipse' in sortTypeBase:
            baseName = 'ellipse'
        elif 'circular' in sortTypeBase:
            baseName = 'cArc'
        elif 'elliptical' in sortTypeBase:
            baseName = 'eArc'
        elif 'bezier' in sortTypeBase:
            baseName = 'bezier'
        
        if self.parametric:
            baseName += '-P'
        else:
            baseName += '-N'
            
        baseName += self.dimension

        if 'cArc-P' not in baseName and 'eArc-P' not in baseName:
            baseSort = sort.getSort(baseName)
            
            if baseSort == None:
                if self.parametric:
                    for sortTypeName in self.pDict:
                        if baseName in sortTypeName:
                            baseSort = self.pDict[baseName]
                            break
                else:
                    for sortTypeName in self.npDict:
                        if baseName in sortTypeName:
                           baseSort = self.npDict[baseName]
                           break

            sortName = baseName

            if sortType.type == 'sort-attribute':
                attributes = sortType.weight
                sortName += '_' + attributes.name
    
                if sortName not in sort.register:
                    newSort = attributeSort(sortName, baseSort, attributes)
                else:
                    newSort = sort.register[sortName]
            
            elif sortType.type == 'sort-primitive':
                if sortName not in sort.register:
                    newSort = baseSort
                else:
                    newSort = sort.register[sortName]
                
            return newSort
        else:
            warnings.warn(baseName + ' is currently not enabled in the sortal library')
            return
            
"""
==========
CLASS INFO
==========
'convertShape' is a class of functions that converts a sortal shape data structure's geometric individuals to the target sort.
Any attributes attached to the geometry or standalone external descriptions already present in the sort type, if it is a disjunctive sort, are also
carried over to the new sort type.

Potential conversion directions:
    a. attribute/primitive -> attribute/primitive, disjunctive (but this is immediately inserted into the prevailing compound sort 'rhino_shapes_c')
    b. disjunctive -> attribute/primitive, disjunctive (but this is immediately inserted into the prevailing compound sort 'rhino_shapes_c')
    c. compound (This is only possible if there is only one functioning geometric disjunctive sort in 'rhino_shapes_c' and the other sort type present is a dummy
       disjunctive sort, i.e. composed of two dummy description sorts
       -> attribute/primitive, disjunctive (but this is immediately inserted into the prevailing compound sort 'rhino_shapes_c')
"""
   
class convertShape(object):
    # Converts from sortal compound form to compound agnostic object
    def __init__(self, form, targetSort):
        self.form = form
        self.sort = form.sort
        self.type = form.type
        self.targetSort = targetSort
        self.newSorts = {'description': []}
        self.finalSort = targetSort
        self.objects = {'description': []}
            
    def toTarget(self):
        # Check if form sort is already representative of the target sort type
        if self.sort == self.targetSort:
            warnings.warn('The form sort is already in the target sort type. Exiting function.')
            return
        
        if self.sort.type == 'sort-compound' and len(self.sort.component) == 2:
            check = False
            for index, sortType in enumerate(self.sort.component):
                descCount = 0
                for element in sortType.component:
                    if 'description' in element.baseSort().characteristic.type:
                        descCount += 1
                        
                if descCount == 2:
                    check = True
                    break
            
            if check == True:
                if index == 0:
                    tempMain = self.mapSorts(self.sort.component[1])
                else:
                    tempMain = self.mapSorts(self.sort.component[0])
                    
            else:
                warnings.warn ('Level of sort type is a compound sort with more than one functioning geometric disjunctive sort. Please extract the shape to be converted first.')
                warnings.warn ('Current sort type: ' + self.sort.type, ', target sort: ' + self.targetSort.type)
                return None
        else:
            tempMain = self.mapSorts(self.sort)
        
        if self.sort.type == self.targetSort.type or ((self.sort.type == 'sort-attribute' or self.sort.type == 'sort-primitive') and self.targetSort.type == 'sort-disjunctive') or (self.sort.type == 'sort-disjunctive' and (self.targetSort.type == 'sort-attribute' or self.targetSort.type == 'sort-primitive')) or len(tempMain) > 0:
            tempTarget = self.mapSorts(self.targetSort)
        else:
            warnings.warn ('Level of sort type does not match target sort')
            warnings.warn('Current sort type: ' + str(self.sort.type) + ', target sort: ' + str(self.targetSort.type))
            return None
        self.matchSorts(tempMain, tempTarget)

        """
        elif self.sort.type == 'sort-attribute' or self.sort.type == 'sort-primitive':
            if self.target
            if self.sort.type == 'sort-compound':
            elif self.sort.type == 'sort-attribute' or self.sort.type == 'sort-primitive':
                if self.targetSort.type == 'sort-disjunctive':
                    tempMain = self.mapSorts(self.sort)
                    tempTarget = self.mapSorts(self.targetSort)
                    self.matchSorts(tempMain, tempTarget)
                else:
                    warnings.warn ('Level of sort type does not match target sort')
                    warnings.warn ('Current sort type: ' + self.sort.type, ', target sort: ' + self.targetSort.type)
                    return 
        """            
        individuals = self.form.elements
        
        # DISJUNCTIVE SORT
        if self.type == 'form-metaForm': 
            formElements = self.form.elements  
            
            # Handles disjunctive sorts inside compound object
            if formElements[0].type == 'form-metaForm':
                for formElement in formElements:
                    name = formElement.sort.name.split('_')
                    name = name[-1]
                    self.fMetaform(formElement, formElement.sort.name)
                    
            # Handles single disjunctive sort
            else:
                self.fMetaform(formElements, self.form.sort.name)
                
        # SIMPLE SORTS 
        # Entered if form only contains points 
        elif self.type == 'form-discrete':
            self.fDiscrete(self.form.elements, self.form.sort.name)
        # Entered if form only contains line segments
        # Extracts data from line segments; parses them into agnostic object data
        elif self.type == 'form-interval':
           self.fInterval(self.form.elements, self.form.sort.name)
        # Entered if form only contains plane segments 
        # Extracts data from plane segments; parses them into agnostic object data
        elif elements[ind].type == 'form-coForm' or elements[ind].type == 'form-areal':
            self.fAreal(individuals, self.form.sort.name)
        elif elements[ind].type == 'form-curve':
            self.fCurve(individuals, self.form.sort.name)
        elif elements[ind].type == 'form-cyclical':
            self.fCyclical(individuals, self.form.sort.name)
            
        newForm = self.makeForm()
        if newForm.sort.type == 'sort-disjunctive':
            newForm = self.makeCompoundForm(newForm)

        return newForm
        
    def determineGeom(self, sortType):
        baseName = ''
        
        if 'lineSegment' in sortType.baseSort().characteristic.type:
            baseName = 'lineSeg'
        elif 'point' in sortType.baseSort().characteristic.type:
            baseName = 'point'
        elif 'planeSegment' in sortType.baseSort().characteristic.type:
            baseName = 'planeSeg'
        elif 'circle' in sortType.baseSort().characteristic.type:
            baseName = 'circle'
        elif 'ellipse' in sortType.baseSort().characteristic.type:
            baseName = 'ellipse'
        elif 'circular' in sortType.baseSort().characteristic.type:
            baseName = 'circularArc'
        elif 'elliptical' in sortType.baseSort().characteristic.type:
            baseName = 'ellipticalArc'
        elif 'bezier' in sortType.baseSort().characteristic.type:
            baseName = 'bezierCurve'
        elif 'description' in sortType.baseSort().characteristic.type:
            baseName = 'description'
            
        return baseName
    
    # Maps sorts in targetSort to counterparts in original sort
    def mapSorts(self, sortType):
        targetDict = {}

        # Attribute or primitive sorts
        if sortType.type == 'sort-attribute' or sortType.type == 'sort-primitive':
            if 'description' not in sortType.baseSort().characteristic.type: 
                baseName = self.determineGeom(sortType)
                if baseName not in targetDict:
                        targetDict[baseName] = []
                        
                targetDict[baseName].append(sortType)
            
        # Disjunctive sorts: broken down into disjunctive sort's component sorts
        elif sortType.type == 'sort-disjunctive':
            for component_main in sortType.component:
                baseName = self.determineGeom(component_main)
                
                if baseName not in targetDict:
                    targetDict[baseName] = []
                    
                targetDict[baseName].append(component_main)
    
        return targetDict
        
    # Determines which sorts in the target sort type matchh the sorts in the original sort
    # Prioritization of matching:
    # 1. Semantic
    # 2. Structural
    def matchSorts(self, originalDict, targetDict):
        for geomType in originalDict:
            orig_sorts = originalDict[geomType]

            if geomType not in  targetDict:
                warnings.warn ('No sort type corresponding to ' + geomType + ' is present in the target sort')
                continue 
            else:
                target_sorts = targetDict[geomType]
                
            lastHold = None
            lastHoldName = None
            check = False

            for target_sort in target_sorts:
                # Searches for exact sort type in target sort
                if target_sort in orig_sorts:
                    self.newSorts[target_sort.name] = target_sort
                    
                # Searches for counterpart sort in original sorts
                # Matching priority:
                # 1. Similar name matching (match rate >= 0.6
                # 2. Endbit matching (N2D, N3D, P2D, P3D)
                # 3. Geometry type matching
                else:
                    for orig_sort in orig_sorts:
                        go_through = False
                        # Searches for similarly named sorts
                        if self.similar(orig_sort.name, target_sort.name) >= 0.6 or orig_sort.name[-3:] == target_sort.name[-3:] or orig_sort.name == target_sort.name:
                            if lastHoldName != None and lastHoldName in self.newSorts:
                                if self.newSorts[lastHoldName] is target_sort:
                                    del self.newSorts[lastHoldName]
                                    lastHoldName = None
                                    lastHold = None
                            
                            self.newSorts[orig_sort.name] = target_sort
                            if target_sort in target_sorts:
                                lastHold = target_sorts.pop(target_sorts.index(target_sort))
                            else:
                                lastHold = target_sort
                            lastHoldName = orig_sort.name
                            break
                            
                        # Obtains last saved sort and stores it as targetSort if no other alternative can be found
                        elif lastHold != None and check:
                            self.newSorts[orig_sort.name] = lastHold  
                            target_sorts.pop(target_sorts.index(lastHold))
                            lastHoldName = None
                            lastHold = None
                            break

                        # Searches for sort with similar geometry
                        # Accepts first one it encounters as fulfilling geometry similarity and attribute similarity
                        elif self.determineGeom(orig_sort) == self.determineGeom(target_sort):
                            if orig_sort.type == 'sort-attribute' and target_sort.type == 'sort-attribute':
                                if orig_sort.weight is target_sort.weight:
                                    self.newSorts[orig_sort.name] = target_sort
                                    if target_sort in target_sorts:
                                        lastHold = target_sorts.pop(target_sorts.index(target_sort))
                                    else:
                                        lastHold = target_sort
                                    lastHoldName = orig_sort.name
        return
    
    def similar(self, main, compare):
        return sm(None, main, compare).ratio()

    def fMetaform(self, elements, sortName):
        for individual in elements:
            if individual.sort.name in self.newSorts:
                if individual.type == 'form-metaForm' and len(individual.elements) > 0:
                    self.fMetaform(individual.elements, individual.sort.name)
                    
                # Converts points/circles/elliptical
                elif individual.type == 'form-discrete' and len(individual.elements) > 0:
                    self.fDiscrete(individual.elements, individual.sort.name)
                            
                # Converts lines
                elif individual.type == 'form-interval' and len(individual.elements) > 0:
                    self.fInterval(individual.elements, individual.sort.name)   # Extracts data from line segments; parses them into agnostic object data
                
                # Converts plane segments, circular/elliptical arcs, bezier curves
                elif individual.type == 'form-coForm' or individual.type == 'form-areal' or individual.type == 'form-cyclical' or individual.type == 'form-curve' and len(individual.elements) > 0:
                    self.fGeneral(individual.elements, individual.sort.name)     # Extracts data from plane segments; parses them into agnostic object data
                    
        return
        
    def fInterval(self, individuals, sortName):
        if len(individuals) > 0:
            indivs = self.Line(sortName, individuals)
            self.objects[sortName] =  indivs
        else:
            self.objects[sortName] = []   
        return
    
    # Handles form-discrete: Descriptions and/or Points
    def fDiscrete(self, individuals, sortName):
        if self.determineGeom(sort.register[sortName]) == 'point':
            convertTo = self.Point
        elif self.determineGeom(sort.register[sortName]) == 'circle':
            convertTo = self.Circle
        elif self.determineGeom(sort.register[sortName]) == 'ellipse': 
            convertTo = self.Ellipse
            
        if len(individuals) > 0:
            self.objects[sortName] = []
            
            for indiv in individuals:
                # For description
                if indiv.type == 'ind-description':
                    # Appends description individual to corresponding dictionary item inside self.objects['description']
                    self.objects['description'].append(indiv)

                # Extracts data from points
                else:
                    if type (indiv) == type(self.form.elements):
                        for ind in indiv:
                            self.objects[sortName].append(convertTo(sortName, ind))
                    else:
                         self.objects[sortName].append(convertTo(sortName, indiv))
        
        return
    
    # Handles fAreal, fCyclical, fCurve
    def fGeneral(self, individuals, sortName):
        geomType = self.determineGeom(sort.register[sortName])
        if geomType == 'planeSeg':
            convertTo = self.Plane
        elif geomType == 'circularArc':
            convertTo = self.cArc
        elif geomType == 'ellipticalArc': 
            convertTo = self.eArc
        elif geomType == 'bezierCurve': 
            convertTo = self.bezCurve
            
        if len(individuals) > 0:
            self.objects[sortName] = []
            
            for indiv in individuals:
                if geomType == 'bezierCurve' or geomType == 'ellipticalArc' or geomType == 'circularArc' or geomType == 'planeSeg':
                    self.objects[sortName].append(convertTo(sortName, indiv))
                else:
                    for individual in indiv.elements:
                        obj = convertTo(sortName, individual)
                        self.objects[sortName].append(obj)
        else:
            self.objects[sortName] = []
            
        return
    
    """
    ================================
    CONVERTING GEOMETRIC INDIVIDUALS
    ================================
    """
    
    def makeAttribute(self, baseSort, attribSorts, attribute = None):
        attName = ''
        att = {}

        if attribute != None:
            for ind in attribute:
                if ind.type == 'form-discrete' or ind.type == 'form-special':
                    if 'description' in ind.sort.baseSort().characteristic.type:
                        att[ind.sort.name] = ind.elements
                    else:
                        att[ind.sort.name] = ind
                    
                elif ind.type == 'form-metaForm':
                    for attrib in ind.elements:
                        att[attrib.sort.name] = attrib.elements
                        
                elif ind.type == 'sort.primitive' or ind.type == 'form-ordinal':
                    att[ind.name] = ind
     
        return att 
    
    # Extracts data from line segments, returns agnostic tuple ((coordinates), {attributes dictionary}) of line segment data
    def Line(self, sortName, line_set):
        v1, v2 = 0, 0
        lines = []
        sortType = sort.register[sortName]
        baseSort = self.newSorts[sortName]
        
        #rebuild the sort from geometric base
        # then add in necessary attributes
        # then plug in information from the old individual to make a new individual
        
        # Extracts attributes and coordinates
        for indiv in line_set:
            if '2D' in baseSort.baseSort().characteristic.type:
                if '2D' in sortType.baseSort().characteristic.type:
                    v1 = indiv.start
                    v2 = indiv.end
                elif '3D' in sortType.baseSort().characteristic.type:
                    v1 = vector2D(indiv.start.x, indiv.start.y)
                    v2 = vector2D(indiv.end.x, indiv.end.y)
            elif '3D' in baseSort.baseSort().characteristic.type:
                if '2D' in sortType.baseSort().characteristic.type:
                    v1 = vector3D(indiv.start.x, indiv.start.y, 0)
                    v2 = vector3D(indiv.end.x, indiv.end.y, 0)
                elif '3D' in sortType.baseSort().characteristic.type:
                    v1 = indiv.start
                    v2 = indiv.end
                    
            # Extracts attributes from line segments inside line_set if there are attributes inside
            # Re-enters them into newly built parametric attribute sort and parametric sort
            
            att = {}

            if indiv.hasAttribute():
                if indiv.attribute.sort.type == 'sort-attribute' or indiv.attribute.sort.type == 'sort-primitive':
                    att = self.makeAttribute(baseSort, indiv.attribute.sort, indiv.attribute)
                else:
                    att = self.makeAttribute(baseSort, indiv.attribute.sort.component, indiv.attribute)

            newIndiv = baseSort.baseSort().characteristic(baseSort, v1, v2)
            
            if len(att) > 0:
                for i in att:
                    for a in att[i]:
                        newIndiv.addAttribute(a)

            if indiv.key != None:
                key.createKey(indiv.key.name, newIndiv)
                        
            lines.append(newIndiv)
        
        return lines
        
    # Extracts data from points, returns agnostic tuple ((coordinates), {attributes dictionary}) of point data
    def Point(self, sortName, pointSource):
        att = {}
        baseSort = self.newSorts[sortName] # Target sort
        sortType = pointSource.sort        # Original sort
        
        # Handles point if it has attributes
        if pointSource.hasAttribute():
            if pointSource.attribute.sort.type == 'sort-attribute' or pointSource.attribute.sort.type == 'sort-primitive':
                att = self.makeAttribute(baseSort, pointSource.attribute.sort, pointSource.attribute)
            else:
                att = self.makeAttribute(baseSort, pointSource.attribute.sort.component, pointSource.attribute)
        
        if '2D' in baseSort.baseSort().characteristic.type: # Target is 2D
            if '2D' in sortType.baseSort().characteristic.type:
                pt = baseSort.baseSort().characteristic(baseSort, float(pointSource.value.x.value), float(pointSource.value.y.value))
            elif  '3D' in sortType.baseSort().characteristic.type:
                pt = baseSort.baseSort().characteristic(baseSort, float(pointSource.value.x.value), float(pointSource.value.y.value))
        elif '3D' in baseSort.baseSort().characteristic.type: # Target is 3D
            if '2D' in sortType.baseSort().characteristic.type:
                pt = baseSort.baseSort().characteristic(baseSort, float(pointSource.value.x.value), float(pointSource.value.y.value), 0)
            elif '3D' in sortType.baseSort().characteristic.type:
                pt = baseSort.baseSort().characteristic(baseSort, float(pointSource.value.x.value), float(pointSource.value.y.value), float(pointSource.value.z.value))
        
        if len(att) > 0:
            for i in att:
                for a in att[i]:
                    pt.addAttribute(a)
            
        if pointSource.key != None:
            key.createKey(pointSource.key.name, pt)
                
        return pt
        
    # Extracts data from points, returns agnostic tuple ((coordinates), {attributes dictionary}) of point data
    def Plane(self, sortName, planeSource):
        att = {}
        coords = {}
        outerBounds = []
        innerBounds = []
        baseSort = self.newSorts[sortName]
        sortType = planeSource.sort
        
        for count, bound in enumerate(planeSource.boundaries):
            if count == 0:
                for vector in bound.polygonal:
                    if '2D' in baseSort.baseSort().characteristic.type:
                        outerBounds.append(vector2D(vector.x, vector.y))
                    elif '3D' in baseSort.baseSort().characteristic.type:
                        if '2D' in sortType.baseSort().characteristic.type:
                            outerBounds.append(vector3D(vector.x, vector.y, 0))
                        elif '3D' in sortType.baseSort().characteristic.type:
                            outerBounds.append(vector3D(vector.x, vector.y, vector.z))
            else: 
                innerBound = []
                for vector in bound.polygonal:
                    if '2D' in baseSort.baseSort().characteristic.type:
                        newVector = vector2D(vector.x, vector.y)
                    elif '3D' in baseSort.baseSort().characteristic.type:
                        if '2D' in sortType.baseSort().characteristic.type:
                            newVector = vector3D(vector.x, vector.y, 0)
                        elif '3D' in sortType.baseSort().characteristic.type:
                            newVector = vector2D(vector.x, vector.y, vector.z)
                    
                    innerBound.append(newVector)
                    
                innerBound += (innerBound[0],)
                innerBounds.append(innerBound)
        
        # Handles point if it has attributes
        if planeSource.hasAttribute():
            if planeSource.attribute.sort.type == 'sort-attribute' or planeSource.attribute.sort.type == 'sort-primitive':
                att = self.makeAttribute(baseSort, planeSource.attribute.sort, planeSource.attribute)
            else:
                att = self.makeAttribute(baseSort, planeSource.attribute.sort.component, planeSource.attribute)
                    
        pOut = baseSort.baseSort().characteristic(self.newSorts[sortName], outerBounds)
        pIn = []

        if len(innerBounds) > 0:
            for bound in innerBounds:
                pIn.append(baseSort.baseSort().characteristic(self.newSorts[sortName], bound))
                
        # Sets value of individual to pOut for attribute adding
        tempHoldForm = self.newSorts[sortName].newForm()
        tempHoldForm.insert(pOut)
        
        # if len(pIn)>0, this means that there are holes in the plane segment
        if len(pIn) > 0:
            form2 = self.newSorts[sortName].newForm()
            for obj in pIn:    
                form2.insert(obj)
                
            # Subtract form2 planes from form1
            tempHoldForm.difference(form2)
        
        if len(att) > 0:
            if '2D' in tempHoldForm.type:
                for i in att:
                    for a in att[i]:
                        tempHoldForm(a)
                        
            elif '3D' in tempHoldForm.type:
                for planes in tempHoldForm.elements:
                    for plane in planes:
                        for i in att:
                            for a in att[i]:
                                plane.addAttribute(a)
                                
        if planeSource.key != None:
            key.createKey(planeSource.key.name, tempHoldForm)
        
        return tempHoldForm
        
    def Circle(self, sortName, circleSource):
        center = circleSource.center
        points = circleSource.getPoints()
        baseSort = self.newSorts[sortName]
        sortType = circleSource.sort
        att = {}
         
        # Handles point if it has attributes
        if circleSource.hasAttribute():
            if circleSource.attribute.sort.type == 'sort-attribute' or circleSource.attribute.sort.type == 'sort-primitive':
                att = self.makeAttribute(baseSort, circleSource.attribute.sort, circleSource.attribute)
            else:
                att = self.makeAttribute(baseSort, circleSource.attribute.sort.component, circleSource.attribute)
            
        if '2D' in baseSort.baseSort().characteristic.type:
            if '2D' in sortType.baseSort().characteristic.type:
                newCircle = baseSort.baseSort().characteristic(baseSort, center, points[0])
            else:
                center = vector2D(center.x, center.y)
                point = vector2D(points[0].x, points[0].y)
                newCircle = baseSort.baseSort().characteristic(baseSort, center, point) 
        elif '3D' in baseSort.baseSort().characteristic.type:
            if '2D' in sortType.baseSort.characteristic.type:
                center = vector3D(center.x, center.y, 0)
                point1 = vector3D(points[0].x, points[0].y, 0)
                point2 = vector3D(points[0].x, points[0].y, 0)
                newCircle = baseSort.baseSort().characteristic(baseSort, center, points[0], points[1])
            else:
                newCircle = baseSort.baseSort().characteristic(baseSort, center, points[0], points[1])
        
        if len(att) > 0:
            for i in att:
                for a in att[i]:
                    newCircle.addAttribute(a)
        
        if circleSource.key != None:
            key.createKey(circleSource.key.name, newCircle)
            
        return newCircle
        
    def Ellipse(self, sortName, ellipseSource):
        f1 = ellipseSource.foci[0]
        f2 = ellipseSource.foci[1]
        edge_pt1 = ellipseSource.getPoints()[0]
        edge_pt2 = ellipseSource.getPoints()[1]
        baseSort = self.newSorts[sortName]
        sortType = ellipseSource.sort
        att = {}
         
        # Handles point if it has attributes
        if ellipseSource.hasAttribute():
            if ellipseSource.attribute.sort.type == 'sort-attribute' or ellipseSource.attribute.sort.type == 'sort-primitive':
                att = self.makeAttribute(baseSort, ellipseSource.attribute.sort, ellipseSource.attribute)
            else:
                att = self.makeAttribute(baseSort, ellipseSource.attribute.sort.component, ellipseSource.attribute)
            
        if '2D' in baseSort.baseSort().characteristic.type:
            if '3D' in sortType.baseSort().characteristic.type:
                f1 = vector2D(f1.x, f1.y)
                f2 = vector2D(f2.x, f2.y)
                edge_pt1 = vector2D(edge_pt1.x, edge_pt1.y)
                edge_pt2 = vector2D(edge_pt2.x, edge_pt2.y)
                
        elif '3D' in baseSort.baseSort().characteristic.type:
            if '2D' in sortType.baseSort.characteristic.type:
                f1 = vector3D(f1.x, f1.y, 0)
                f2 = vector3D(f2.x, f2.y, 0)
                edge_pt1 = vector3D(edge_pt1.x, edge_pt1.y, 0)
                edge_pt2 = vector3D(edge_pt2.x, edge_pt2.y, 0)

        newEllipse = baseSort.baseSort().characteristic(baseSort, f1, f2, edge_pt1, edge_pt2)
        
        if len(att) > 0:
            for i in att:
                for a in att[i]:
                    newEllipse.addAttribute(a)
        
        if ellipseSource.key != None:
            key.createKey(ellipseSource.key.name, newEllipse)
        
        return newEllipse
        
    def cArc(self, sortName, cArcSource):
        center = cArcSource.center
        edge_pt1 = cArcSource.start
        edge_pt2 = cArcSource.end
        pin_pt = cArcSource.pin
        baseSort = self.newSorts[sortName]
        sortType = cArcSource.sort
        att = {}
        
        # Handles point if it has attributes
        if cArcSource.hasAttribute():
            if cArcSource.attribute.sort.type == 'sort-attribute' or cArcSource.attribute.sort.type == 'sort-primitive':
                att = self.makeAttribute(baseSort, cArcSource.attribute.sort, cArcSource.attribute)
            else:
                att = self.makeAttribute(baseSort, cArcSource.attribute.sort.component, cArcSource.attribute)
            
        if '2D' in baseSort.baseSort().characteristic.type:
            if '3D' in sortType.baseSort().characteristic.type:

                center = vector2D(center.x, center.y)
                edge_pt1 = vector2D(edge_pt1.x, edge_pt1.y)
                edge_pt2 = vector2D(edge_pt2.x, edge_pt2.y)
                pin_pt = vector2D(pin_pt.x, pin_pt.y)

        elif '3D' in baseSort.baseSort().characteristic.type:
            if '2D' in sortType.baseSort.characteristic.type:
                center = vector3D(center.x, center.y, 0)
                edge_pt1 = vector3D(edge_pt1.x, edge_pt1.y, 0)
                edge_pt2 = vector3D(edge_pt2.x, edge_pt2.y, 0)
                pin_pt = vector3D(pin_pt.x, pin_pt.y, 0)
        
        newCArc = baseSort.baseSort().characteristic(baseSort, center, edge_pt1, edge_pt2, pin_pt)
        
        if len(att) > 0:
            for i in att:
                for a in att[i]:
                    newCArc.addAttribute(a)        
                
        if cArcSource.key != None:
            key.createKey(cArcSource.key.name, newCArc)
            
        return newCArc
        
    def eArc(self, sortName, eArcSource):
        center = eArcSource.center
        f1 = eArcSource.foci[0]
        f2 = eArcSource.foci[1]
        edge_pt1 = eArcSource.start
        edge_pt2 = eArcSource.end
        pin_pt = eArcSource.pin
        baseSort = self.newSorts[sortName]
        sortType = eArcSource.sort
        att = {}
        
        # Handles point if it has attributes
        if eArcSource.hasAttribute():
            if eArcSource.attribute.sort.type == 'sort-attribute' or eArcSource.attribute.sort.type == 'sort-primitive':
                att = self.makeAttribute(baseSort, eArcSource.attribute.sort, eArcSource.attribute)
            else:
                att = self.makeAttribute(baseSort, eArcSource.attribute.sort.component, eArcSource.attribute)
            
        if '2D' in baseSort.baseSort().characteristic.type:
            if '3D' in sortType.baseSort().characteristic.type:
                edge_pt1 = vector2D(edge_pt1.x, edge_pt1.y)
                edge_pt2 = vector2D(edge_pt2.x, edge_pt2.y)
                f1 = vector2D(f1.x, f1.y)
                f2 = vector2D(f2.x, f2.y)
                pin_pt = vector2D(pin_pt.x, pin_pt.y)

        elif '3D' in baseSort.baseSort().characteristic.type:
            if '2D' in sortType.baseSort.characteristic.type:
                edge_pt1 = vector3D(edge_pt1.x, edge_pt1.y, 0)
                edge_pt2 = vector3D(edge_pt2.x, edge_pt2.y, 0)
                f1 = vector2D(f1.x, f1.y, 0)
                f2 = vector2D(f2.x, f2.y, 0)
                pin_pt = vector3D(pin_pt.x, pin_pt.y, 0)
        
        newEArc = baseSort.baseSort().characteristic(baseSort, f1, f2, edge_pt1, edge_pt2, pin_pt)
        
        if len(att) > 0:
            for i in att:
                for a in att[i]:
                    newEArc.addAttribute(a)        
        
        if eArcSource.key != None:
            key.createKey(eArcSource.key.name, newEArc)
            
        return newEArc
        
    def bezCurve(self, sortName, bezSource):
        start_vec = bezSource.start
        ctrl_vec = bezSource.controlPoint
        end_vec = bezSource.end
        baseSort = self.newSorts[sortName]
        sortType = bezSource.sort
        att = {}
        
        # Handles point if it has attributes
        if bezSource.hasAttribute():
            if bezSource.attribute.sort.type == 'sort-attribute' or bezSource.attribute.sort.type == 'sort-primitive':
                att = self.makeAttribute(baseSort, bezSource.attribute.sort, bezSource.attribute)
            else:
                att = self.makeAttribute(baseSort, bezSource.attribute.sort.component, bezSource.attribute)

        if '2D' in baseSort.baseSort().characteristic.type:
            if '3D' in sortType.baseSort().characteristic.type:
                start_vec = vector2D(start_vec.x, start_vec.y)
                ctrl_vec = vector2D(ctrl_vec.x, ctrl_vec.y)
                end_vec = vector2D(end_vec.x, end_vec.y)

        elif '3D' in baseSort.baseSort().characteristic.type:
            if '2D' in sortType.baseSort().characteristic.type:
                start_vec = vector3D(start_vec.x, start_vec.y, 0)
                ctrl_vec = vector3D(ctrl_vec.x, ctrl_vec.y, 0)
                end_vec = vector3D(end_vec.x, end_vec.y, 0)

        newBezier = baseSort.baseSort().characteristic(baseSort, start_vec, ctrl_vec, end_vec)
        
        if len(att) > 0:
            for i in att:
                for a in att[i]:
                    newBezier.addAttribute(a)       
        
        if bezSource.key != None:
            key.createKey(bezSource.key.name, newBezier)
        
        return newBezier
        
    """
    =================================
    CREATING FORMS AND COMPOUND FORMS
    =================================
    """
    
    def makeForm(self):
        newForms = []
        targetForm = None
        if len(self.newSorts) > 1:
            for sortName in self.objects:
                check = True
                elements = self.objects[sortName]
                targetSort = self.newSorts[sortName]
                
                if sortName != 'description':
                    newForm = targetSort.newForm()
                    
                    for element in elements:
                        if 'form' not in element.type:
                            newForm.insert(element)
                        else:
                            newForms.append(element)
                            check = False
                            
                    if check:
                        newForms.append(newForm)
       
            targetForm = self.finalSort.newForm()
    
            for newForm in newForms:
                targetForm.insert(newForm)
        
        if len(self.objects['description']) > 0:
            # In case the shape to be converted is made only from descriptions
            if targetForm == None: 
                targetForm = self.finalSort.newForm()
                
            for i in self.objects['description']:
                targetForm.insert(i)

        return targetForm
    
    def makeCompoundForm(self, shape):
        compound_sort = sort.register['rhino_shapes_c']
        if shape.sort not in compound_sort.component:
            components_list = compound_sort.component
            del sort.register['rhino_shapes_c']
            components_list.append(shape.sort)
            compound_sort = compoundSort('rhino_shapes_c', components_list)
            
        final_shape = compound_sort.newForm()
        final_shape.insert(shape)
        
        return final_shape