# -*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

import rhinoscriptsyntax as rs
import Rhino as r
import scriptcontext as sc
from builtins import range

from sortal import (transformSet, rule, primitiveSort, point3D, point2D, coordinate, lineSegment3D, lineSegment2D, 
                    line3D, label, rWeight, disjunctiveSort, attributeSort, affineTransform,  vector3D, sort)

from mpmath import pi, sin, cos, tan
from copy import deepcopy
from math import sqrt

"""
VisualA Class
Takes agnostic form as input and draws it in Rhino
Returns Rhino object GUIDs based on agnostic form

INPUT
Agnostic form dictionary, e.g. {'points': [<p1>, <p2>], 'line segments': [<ls1>, <ls2>]}

OUTPUT
Draws objects in Rhino based on agnostic object dictionary
Returns list of GUIDs corresponding to Rhino objects drawn by class methods
"""

rhdoc = r.RhinoDoc.ActiveDoc

class visualA:
    
    # Sets class instance variables to inputs and sets layer to be drawn on
    # Default value of refPoint is origin (0,0,0)
    def __init__ (self, agnosticObject, refPoint = (0,0,0), currentLayer = 'Default', preds = {}, dirs = {}):
        self.agnosticObject = agnosticObject
        self.objectIds = []
        self.currentLayer = currentLayer
        try:
            rs.CurrentLayer(self.currentLayer)
        except:
            pass
            # ('The layer you inputted does not exist')
            
        self.tag = False
        self.tags = {}
        
        for i in preds:
            self.tags[i] = preds[i]
        
        for i in dirs:
            self.tags[i] = dirs[i]
        
        for a in self.tags: 
             if len(self.tags[a]) > 0:
                 self.tag = True
                 break
      
        if refPoint != None:
            if type(refPoint) == type(()) or type(refPoint) == type([]):
                self.refPoint = r.Geometry.Point3d(refPoint[0], refPoint[1], refPoint[2])
            else:
                self.refPoint = refPoint
        else:
            self.refPoint = r.Geometry.Point3d(0,0,0)
    
    # Composes tagText for a geometric individual
    # tagName contains geometric tag to search for in predicates/directives list
    def searchTag(self, tagName):
        tagText = ''

        # Goes through dictionary of predicate/directive types
        for tagType in self.tags:
            tagList = self.tags[tagType] # Obtains list/dictionary per predicate/directive type

            if type(tagList) == type([]):
                if len(tagList) > 0:
                    for tagItem in tagList:
                        newTag = tagType 
                        avoid = 0

                        if '@' + tagName in tagItem:
                            if tagType == 'pointOL':
                                for value in tagItem[1:]:
                                    if type(value) == type(''): 
                                        if value[0] != '@':
                                            newTag += '_' + str(value)
                                    elif type(value) == type(0.0):
                                        newTag += '_' + str(value)
                            
                            fullTag = newTag
                            
                            if tagText == '':
                                tagText = fullTag
                            else:
                                tagText = tagText + '|' + fullTag
                                
                else:
                    if tagName in self.tags[tagType]:
                        if tagText == '':
                            tagText = tagType
                        else:
                            tagText = tagText + '|' + tagType

            elif type(self.tags[tagType]) == type({}):

                for name in tagList:
                    # Gets range and coordinate data
                    tagData = tagList[name]

                    # Generates pred/dir group name
                    newTag =  tagType + '^' + name
                    if tagType != 'embeds':
                        tempName = '@' + tagName
                    else:
                        tempName = tagName
                        
                    if tempName in tagData:
                        # Adds in first value
                        fullTag = newTag
                        
                        if tagType == 'normal':
                            # Line Seg Normal: {(#rhsLineTag, x, y, lengthMin, lengthMax)} or {(#rhsLineTag, x, y, length)} 
                            # Plane Seg Normal: {(#planeTag, #rhsLineTag, lengthMin, lengthMax)} or {(#planeTag, #rhsLineTag, length)}
                            if 'lineSegment' in tagData[1]:
                                if len(tagData) == 5:
                                    fullTag += '_' + tagData[1][1:] + '_' + str(tagData[2]) +',' + str(tagData[3]) + '_' + str(tagData[4])
                                elif len(tagData) == 6:
                                    fullTag += '_' + tagData[1][1:] + '_' + str(tagData[2]) +',' + str(tagData[3]) + '_' + str(tagData[4]) + '_' + str(tagData[5])
                            elif 'planeSegment' in tagData[1]:
                                if len(tagData) == 4:
                                    fullTag += '_' + tagData[1][1:] + '_' + tagData[2][1:] +'_' + str(tagData[3]) 
                                elif len(tagData) == 5:
                                    fullTag += '_' + tagData[1][1:] + '_' + tagData[2][1:] +'_' + str(tagData[3]) + '_' + str(tagData[4])
                        
                        elif tagType == 'embeds':
                            if tagData.index(tempName) == 0:
                                fullTag += '_' + tagData[1]
                            else:
                                fullTag += '_' + tagData[0]
                                
                        elif 'bound' not in tagType:
                            if tagData.index(tempName) == 0:
                                pair = tagData[1][1:]
                            else:
                                pair = tagData[0][1:]

                            fullTag = newTag + '_' +str(tagData[2])
                            if len(tagData) == 4:
                                 fullTag += '_' + str(tagData[3])

                            if type(tagData[1]) == type(0.0) or type(tagData[1]) == type(0):
                                if tagData[0] != tagData[1]:
                                    fullTag += '_' + str(tagData[1])
                                    
                            fullTag += '_' + str(pair)
                        else:
                            fullTag = name
                            
                        if tagText == '':
                            tagText = fullTag
                        else:
                            tagText += '|' + fullTag

        return tagText
        
    # Draws agnostic form as Rhino objects
    def plotAgnostic(self):
        # Extracts geometric data and attribute information, plots them in the Rhino workspace
        # Distinguishes between disjunctive sort type and attribute/primitive sort types
        # Enumerates sort types if disjunctive sort type
        for sortType in self.agnosticObject:
            if sortType != 'descriptions':
                if sort.register[sortType].type == 'sort-disjunctive':
                    for subshape in self.agnosticObject[sortType]:
                        geometry = self.agnosticObject[sortType][subshape]
    
                        if len(geometry) > 0 and subshape != 'descriptions':
                            self.processGeom(subshape, geometry)
                
                # Proceeds immediately to plotting the geometry inside the dictionary element
                elif sort.register[sortType].type == 'sort-attribute' or sort.register[sortType].type == 'sort-primitive':
                       self.processGeom(sortType, self.agnosticObject[sortType])

        # Moves set of Rhino objects using refPoint as basis; 
        # e.g. can be used to redraw a form after rule application in the same location
        rs.MoveObjects(self.objectIds, self.refPoint)

        # Deletes refPoint from Rhino window
        #rs.DeleteObject(self.refPoint)  
        
        return self.objectIds
    
    def processGeom(self, layerName, objects):
        # Note that if layerName (sort type) exists within the layer names of the Rhino workspace, 
        # Then the object is moved to that layer.
        # Plots points/text dots
        if 'point' in sort.register[layerName].baseSort().characteristic.type:
            for obj in objects:
                pt = self.makePoint(obj[0])
                attributes = obj[1]
                
                # Labels and Label Descriptions need to go first in assigning attributes
                # Then followed by color, enums, weight
                        
                #locPt =  Rhino.Geometry.Point3d(obj[0][0], obj[0][1], obj[0][2])
                #pt = (rs.AddText(totalText, locPt, height=2.0, font="Arial", font_style=0, justification=None))

                pt = self.makeAttribute('pt', pt, obj[1])
                if layerName in rs.LayerNames():
                    rs.ObjectLayer(pt, layerName)
                self.objectIds.append(pt)
        
        # Plots line segments with/without attributes
        elif 'lineSegment' in sort.register[layerName].baseSort().characteristic.type:
            for obj in objects:
                line = self.makeLine(obj[0])
                line = self.makeAttribute('ls', line, obj[1])
                if layerName in rs.LayerNames():
                    rs.ObjectLayer(line, layerName)
                self.objectIds.append(line)
        
        elif 'planeSegment' in sort.register[layerName].baseSort().characteristic.type:
            for obj in objects:
                totalPlaneSeg = self.makePlane(obj[0])    # Creates plane object in Rhino Geometry form (not yet in Rhino workspace)
                planeSeg = self.makeAttribute('pl', totalPlaneSeg, obj[1])
                if layerName in rs.LayerNames():
                    rs.ObjectLayer(planeSeg, layerName)
                self.objectIds.append(planeSeg)
                
        elif 'circle' in sort.register[layerName].baseSort().characteristic.type:
            for obj in objects:
                circle = self.makeCircle(obj[0])    # Creates plane object in Rhino Geometry form (not yet in Rhino workspace)
                circleAtt = self.makeAttribute('c', circle, obj[1])
                if layerName in rs.LayerNames():
                    rs.ObjectLayer(circleAtt, layerName)
                self.objectIds.append(circleAtt)
        
        elif 'ellipse' in sort.register[layerName].baseSort().characteristic.type:
            for obj in objects:
                ellipse = self.makeEllipse(obj[0])    # Creates plane object in Rhino Geometry form (not yet in Rhino workspace)
                ellipseAtt = self.makeAttribute('e', ellipse, obj[1])
                if layerName in rs.LayerNames():
                    rs.ObjectLayer(ellipseAtt, layerName)
                self.objectIds.append(ellipseAtt)
        
        elif 'Arc' in sort.register[layerName].baseSort().characteristic.type:
            if 'circular' in sort.register[layerName].baseSort().characteristic.type:
                for obj in objects:
                    cArc = self.makeCArc(obj[0])    # Creates plane object in Rhino Geometry form (not yet in Rhino workspace)
                    cArcAtt = self.makeAttribute('cArc', cArc, obj[1])
                    if layerName in rs.LayerNames():
                        rs.ObjectLayer(cArcAtt, layerName)
                    self.objectIds.append(cArcAtt)
        
            elif 'elliptical' in sort.register[layerName].baseSort().characteristic.type:
                #print ('Elliptical arcs are currently not visually supported')
                #sc.sticky['delete'] = True
                #"""
                for obj in objects:
                    eArc = self.makeEArc(obj[0])    # Creates plane object in Rhino Geometry form (not yet in Rhino workspace)
                    eArcAtt = self.makeAttribute('eArc', eArc, obj[1])
                    if layerName in rs.LayerNames():
                        rs.ObjectLayer(eArcAtt, layerName)
                    self.objectIds.append(eArcAtt)
                #"""
        elif 'bezier' in sort.register[layerName].baseSort().characteristic.type:
            for obj in objects:
                bezier = self.makeBezier(obj[0])    # Creates plane object in Rhino Geometry form (not yet in Rhino workspace)
                bezierAtt = self.makeAttribute('b', bezier, obj[1])
                if layerName in rs.LayerNames():
                    rs.ObjectLayer(bezierAtt, layerName)
                self.objectIds.append(bezierAtt)
                
        return
        
    def makePoint(self, obj):
        if len(obj) == 2:
            pt = rs.AddPoint(obj[0], obj[1], 0)
        elif len(obj) == 3:
            pt = rs.AddPoint(obj[0], obj[1], obj[2])
        
        return pt
        
    def makeLine(self, info):   
        if len(info[0]) == 2:
            v1 = r.Geometry.Point3d(info[0][0], info[0][1], 0)
            v2 = r.Geometry.Point3d(info[1][0], info[1][1], 0)
        elif len(info[0]) == 3:
            v1 = r.Geometry.Point3d(info[0][0], info[0][1], info[0][2])
            v2 = r.Geometry.Point3d(info[1][0], info[1][1], info[1][2])
        
        return rs.AddLine(v1, v2)

    def makePlane(self, info):
        if type(info) == type({}):
            outer = info['outer']
            inner = info['inner']
            trims = []
    
            corners = []
            boundLines = []
            
            for bound in outer:
                if len(bound) == 2:     # 2D
                    corners.append(r.Geometry.Point3d(bound[0], bound[1], 0))
                elif len(bound) == 3:   #3D
                    corners.append(r.Geometry.Point3d(bound[0], bound[1], bound[2]))
                    
            corners.append(corners[0])
            outerBound =  r.Geometry.PolylineCurve(corners)
            
            if len(inner) > 0:
                trims = []
                
                for boundSet in inner:
                    corners = []
                    
                    for bound in boundSet:
                        if len(bound) == 2:
                            corners.append(r.Geometry.Point3d(bound[0], bound[1], 0))
                        elif len(bound) == 3:
                            corners.append(r.Geometry.Point3d(bound[0], bound[1], bound[2]))
                        
                    corners.append(corners[0]) 
                    innerBound = r.Geometry.PolylineCurve(corners)
                    trims.append(innerBound)
                    
                trims.append(outerBound)
                totalPlaneSeg = r.Geometry.Brep.CreatePlanarBreps(trims)
            else:
                totalPlaneSeg = r.Geometry.Brep.CreatePlanarBreps(outerBound)
    
        elif type(info) == type([]):
            outerBound = info
            corners = []
            
            for bound in outerBound:
                corners.append(r.Geometry.Point3d(bound[0], bound[1], bound[2]))

            totalPlaneSeg = r.Geometry.Brep.CreatePlanarBreps(outerBound)
        
        return totalPlaneSeg
    
    def makeCircle(self, info):
        center = info[0]
        
        #3D Circle
        if len(info) == 5:
            pin_1 = info[2] 
            pin_2 = info[3] 
            pin_3 = info[4]
            circle = rs.AddCircle3Pt(pin_1, pin_2, pin_3)
            
        elif (len (info) == 4 or len(info) == 3) and len(info[0]) == 3:
            if len(info) == 4:
                center = r.Geometry.Point3d(info[1][0], info[1][1], info[1][2])
                pin_1 = r.Geometry.Point3d(info[3][0], info[3][1], info[3][2])
                pin_2 = r.Geometry.Point3d(info[4][0], info[4][1], info[4][2])
                
            elif len(info) == 3:
                center = r.Geometry.Point3d(info[1][0], info[1][1], info[1][2])
                pin_1 = r.Geometry.Point3d(info[2][0], info[2][1], info[2][2])
                pin_2 = r.Geometry.Point3d(info[3][0], info[3][1], info[3][2])
            
            radius = sqrt((center[0]-pin_1[0])**2 + (center[1]-pin_1[1])**2 +(center[2]-pin_1[2])**2)
            angle = random()*pi*2
            x = cos(angle)*radius
            y = sin(angle)*radius
            z = tan(angle)*radius
            pin_3 = r.Geometry.Point3d(center[0]+x, center[1]+y, center[2]+z)
            planeC = r.Geometry.Plane3d(pin_1, pin_2, pin_3)
            circle = rs.AddCircle(planeC, radius)
            
        #2D Circle
        elif len(info) == 2 or len(info[0]) == 2:
            radius = info[1]
            circle = rs.AddCircle(center, radius)

        return circle
    
    # Draws ellipses
    def makeEllipse(self, info):
        center = info[0]
        if len(info) == 4:
            shortPt = r.Geometry.Point3d(info[3][0], info[3][1], info[3][2])
            bDist = (center[0] - shortPt.X)**2 + (center[1]-shortPt.Y)**2 + (center[2]-shortPt.Z)**2
            cDist = (info[1][0] - center[0])**2 + (info[1][1] - center[1])**2 + (info[1][2] - center[2])**2
            aDist = sqrt(bDist + cDist)

            # x-coords are equal
            if info[1][0] == info[2][0]:
                longPt = (center[0], center[1] - aDist, center[2])
            
            # y-coords are equal
            elif info[1][1] == info[2][1]:
                longPt = (center[0]+ aDist, center[1], center[2])
                
        else:
            longPt = r.Geometry.Point3d(info[3][0], info[3][1], info[3][2])
            shortPt = r.Geometry.Point3d(info[4][0], info[4][1], info[4][2])
            
            if len(center) == 2:
                center = r.Geometry.Point3d(center[0], center[1], 0) 
            else:
                center = r.Geometry.Point3d(center[0], center[1], center[2]) 

        return rs.AddEllipse3Pt(center, longPt, shortPt)
        
    def makeCArc(self, info):
        start = info[1]
        end = info[2]
        mid = info[3]
        
        if len(start) == 2:
            start_pt = r.Geometry.Point2d(start[0], start[1]) 
            end_pt = r.Geometry.Point2d(end[0], end[1]) 
            mid_pt = r.Geometry.Point2d(mid[0], mid[1]) 
        else:
            start_pt = r.Geometry.Point3d(start[0], start[1], start[2]) 
            end_pt = r.Geometry.Point3d(end[0], end[1], end[2]) 
            mid_pt = r.Geometry.Point3d(mid[0], mid[1], mid[2]) 
            
        return rs.AddArc3Pt(start_pt, end_pt, mid_pt)
    
    # Draws elliptical arc
    # Agnostic data set: (<center>, <foc1>, <foc2>, <start>, <end>, <maj_axis_pt or randomly generated point serves as pin>, <minor axis quadrant point>, <major axis quadrant point>)
    def makeEArc(self, info):
        center = r.Geometry.Point3d(info[0][0], info[0][1], info[0][2])
        start = r.Geometry.Point3d(info[3][0], info[3][1], info[3][2])
        end = r.Geometry.Point3d(info[4][0], info[4][1], info[4][2])
        short_pt = r.Geometry.Point3d(info[6][0], info[6][1], info[6][2])
        long_pt = r.Geometry.Point3d(info[7][0], info[7][1], info[7][2])
        
        # Creation of base ellipse 
        ellipse = r.Geometry.Ellipse(center, long_pt, short_pt)
        
        # Conversion of ellipse to nurbs curve; necessary so that the ellipse can be trimmed
        nurb = ellipse.ToNurbsCurve() 
        start_param = nurb.ClosestPoint(start)[1]
        end_param = nurb.ClosestPoint(end)[1]
        ellArc = nurb
        
        # Trimming of ellipse to create elliptical arc
        if start_param != end_param:
            ellArc = nurb.Trim(start_param, end_param)
            if ellArc.IsValid:
                ellArc = sc.doc.Objects.AddCurve(nurb)
                sc.doc.Views.Redraw()

        ellArc = rs.TrimCurve(ellArc, [end_param, start_param], True)
        a = rs.coercecurve(ellArc)
        return ellArc

    def makeBezier(self, info):
        start = info[0]
        end = info[1]
        mid = info[2]
        
        if len(start) == 2:
            start_pt = r.Geometry.Point3d(start[0], start[1], 0) 
            end_pt = r.Geometry.Point3d(end[0], end[1], 0) 
            mid_pt = r.Geometry.Point3d(mid[0], mid[1], 0) 
        else:
            start_pt = r.Geometry.Point3d(start[0], start[1], start[2]) 
            end_pt = r.Geometry.Point3d(end[0], end[1], end[2]) 
            mid_pt = r.Geometry.Point3d(mid[0], mid[1], end[2]) 

        bez = r.Geometry.Curve.CreateControlPointCurve([start_pt, end_pt, mid_pt])
        result = sc.doc.Objects.AddCurve(bez.ToNurbsCurve())
        sc.doc.Views.Redraw()

        return result
        
    def makeAttribute(self, geomType, geometry, attributes):
        drawDot = ['ls', 'pl', 'c', 'e', 'cArc', 'eArc', 'b']

        if geomType == 'ls':
            midpt = rs.CurveMidPoint(geometry)
        elif geomType == 'pl':
            # If brep creation was successful, the object is added to the Rhino document
            # and a unique GUID is generated for it, which is stored in self.objectIds
            if geometry:
                planeGeom = deepcopy(geometry)
                geometry = sc.doc.Objects.AddBrep(geometry[0])
            sc.doc.Views.Redraw()

            # Generate coordinates for text dot
            # midpt is only used if there is attribute text for this plane
            for curve in planeGeom[0].Loops:
                if curve.LoopType == r.Geometry.BrepLoopType.Outer:
                    mainCurve = curve.To3dCurve().DuplicateSegments()[0]
                    startpt = mainCurve.PointAtStart
                    endpt = mainCurve.PointAtEnd
                    midpt = ((startpt.X + endpt.X)/2,(startpt.Y + endpt.Y)/2,(startpt.Z + endpt.Z)/2)
                    midpt =  r.Geometry.Point3d(midpt[0], midpt[1], midpt[2])
                    break
        elif geomType == 'c' or geomType == 'e' or geomType == 'b':
            midpt = rs.CurveStartPoint(geometry)
        elif geomType == 'cArc' or geomType == 'eArc':
            midpt = rs.CurveMidPoint(geometry)
            
        totalText = ''
        labelText = ''
        descText = ''
        textDot = None
        index = 0
        tag = ''
        tagName = ''
        tagText = ''
        newTagText = ''
        
        for attrib in attributes:
            content = attributes[attrib]
            
            if attrib == 'labels' or attrib == 'label_descriptions' or attrib == 'tags':
                # Variable holders for label/label description text                     
                if 'LabelDot' in rs.GetUserText(geometry):
                    textDot = rs.GetUserText(geometry, 'LabelDot')

                if attrib == 'labels':
                    for item in content:
                        rs.SetUserText(geometry, '&' + str(index), item)
                        index += 1 
    
                        if labelText == '':
                            labelText = item
                        else:
                            labelText +=  ('|'+ item)
                            
                    rs.SetUserText(geometry, 'lText', labelText)          
                elif attrib == 'label_descriptions':
                    for item in content:
                        rs.SetUserText(geometry, '#' + str(index), item)
                        index += 1 
                        newText = item
                        if item.isdigit() == False:
                            newText = item
                        
                        if descText == '':
                            descText = newText
                        else:
                            descText +=  ('|'+ newText)
                    
                    rs.SetUserText(geometry, 'ldText', descText)          
                    
                elif attrib == 'tags' and self.tag:
                    
                    for tag in attributes['tags']:
                        if 'void' in tag:
                            if tagText == '':
                                tagText = tag
                            else:
                                tagText += '|' + tag
                        else:
                            if self.searchTag(tag) not in tagText:
                                if tagText == '':
                                    tagText = self.searchTag(tag)
                                else:
                                    tagText += '|' +  self.searchTag(tag)

                            newTag = tag.split('-')[-1]

                            if 'tag' not in rs.GetUserText(geometry):
                                rs.SetUserText(geometry, 'tag', newTag)
                            
                            if newTag not in tagName and 'void' not in newTag:
                                if tagName == '':
                                    tagName = newTag
                                else:
                                    tagName += '|' + newTag

                    if tagName != '':
                        rs.SetUserText(geometry, 'tag', tagName)
                        if geomType == 'ls' or geomType == 'py':
                            tag0 = 'line('+ tag + ')'
                        elif geomType == 'pt':
                            tag0 = 'point(' + tag + ')'
                        else:
                            tag0 = 'plane(' + tag + ')'
                        totalText = tag0
                        
                    if tagText != '':
                        tagText0 = '#(' + tagText + ')'   
                        
                    tagData = tagText.split('|')
                    newTagText = ''

                    for item in tagData:
                        values = item.split('_')
                        addText = ''
                        
                        if 'pointOL' in values[0]:
                            addText = item
                        
                        elif 'normal' in values[0]:
                            newText = values[0]
                            retrieve = 'norm'
                            count = values[0].split('^')[-1]
                            rs.SetUserText(geometry, retrieve + '_end_' + count, values[1]) 
                            rs.SetUserText(geometry, retrieve + '_rhs' + count, values[2])
                            
                            if len(values) == 4:
                                rs.SetUserText(geometry, retrieve + '1_' + count, values[3]) 
                                rs.SetUserText(geometry, retrieve + '2_' + count, values[3]) 
                                if len(values[3]) > 20:
                                    values.pop(3)
                            elif len(values) == 5:
                                rs.SetUserText(geometry, retrieve + '1_' + count, values[3]) 
                                rs.SetUserText(geometry, retrieve + '2_' + count, values[4]) 
                                
                                # Removes length ranges from text dot text if they are longer than 20 characters, for visual clarity
                                if len(values[3]) > 20 or len(values[4]) > 20:
                                    values.pop(3)
                                    values.pop(3)
                            else:
                                raise ValueError('Incorrect input for normal directive')
                                
                            for textBit in values[1:]:
                                newText += '_' + textBit
                            
                            addText = newText
                            
                        # If length(s) is/are greater than 20 characters, not included in visual text dot text
                        # Paired item is included in distance directive in visual text dot text
                        elif 'distance' in values[0]:
                            retrieve = 'dist'
                            newText = ''

                            if 'planeSeg' in values[-1]:
                                pair = '8' + '-' + values[-1].split('-')[-1]
                            elif 'lineSeg' in values[-1]:
                                pair = '4' + '-' + values[-1].split('-')[-1]
                            elif 'point' in values[-1]:
                                pair = '8192' + '-' + values[-1].split('-')[-1]

                            groupName = values[0].split('^')[-1]
                            seg2name = None

                            try:
                                seg1name = deepcopy(float(values[1]))
                            except:
                                seg1name = deepcopy(str(values[1]))

                            if len(values) == 4:
                                try:
                                    seg2name = deepcopy(float(values[2]))
                                except:
                                    seg2name = deepcopy(str(values[2]))
                                
                            rs.SetUserText(geometry, retrieve + '1' + '_' + groupName, seg1name)
                            
                            if seg2name == None:
                                rs.SetUserText(geometry, retrieve + '2' + '_' + groupName, seg1name)
                            else:
                                rs.SetUserText(geometry, retrieve + '2' + '_' + groupName, seg2name)
                                
                            rs.SetUserText(geometry, 'pair' + '_' + retrieve + '_' + groupName, pair)

                            if len(seg1name) > 20:
                                values.pop(values.index(seg1name))
                                if seg2name != None:
                                    values.pop(values.index(seg2name))

                            if seg2name in values:
                                if len(seg2name) > 20:
                                    values.pop(values.index(seg1name))
                                    values.pop(values.index(seg2name))

                            for value in values:
                                if newText == '':
                                    newText = value
                                else:
                                    newText += '_' + value
                            
                            addText = newText
                            if newTagText == '':
                                newTagText = newText
                            else:
                                newTagText += '|' + newText
                                
                        elif 'embeds' in values[0]:
                            addText = values[0] + '_' + values[1]
                            if 'planeSeg' in values[1]:
                                pair = '8' + '-' + values[1].split('-')[-1]
                            elif 'lineSeg' in values[-1]:
                                pair = '4' + '-' + values[1].split('-')[-1]
                            elif 'point' in values[-1]:
                                pair = '8192' + '-' + values[1].split('-')[-1]
                                
                            rs.SetUserText(geometry, 'embeds_' + values[0].split('^')[-1], pair)
                        
                        if addText != '':
                            if newTagText == '':
                                newTagText = addText
                            else:
                                newTagText += '|' + addText
                    
                    if newTagText != '':
                        rs.SetUserText(geometry, 'tagText', newTagText)
                        
            elif attrib == 'weight':   
                if type(content) == type([]):
                    weight = float(content[0])
                else:
                    weight = float(content)
                    
                if geomType in drawDot:
                    rs.ObjectPrintWidth(geometry, weight)
                else:
                    rs.ObjectColor(geometry, (float(weight[0]), float(weight[0]), float(weight[0])))
            
            elif attrib == 'weight255':
                weight = content
                rs.ObjectColor(geometry, (float(weight[0]), float(weight[1]), float(weight[2])))
                    
            elif attrib == 'enums':
                rs.SetUserText(geometry, 'Enum', content)
                
            elif attrib == 'color':
                rs.ObjectColor(geometry, (float(content[0]), float(content[1]), float(content[2])))
        if tagName != '':
            if geomType == 'ls' or geomType == 'py':
                totalText = 'line(' + tagName + ')'
            elif geomType == 'pt':
                totalText = 'point(' + tagName + ')'
            else:
                totalText = 'plane(' + tagName + ')'
                
        if labelText != '':
            if totalText == '':
                totalText = 'l(' + labelText + ')'
            else:
                totalText += '|' + 'l(' + labelText + ')'
            
        if descText != '':
            if totalText == '':
                totalText = 'd(' + descText + ')'
            else:
                totalText += '|' + 'd(' + descText + ')'

        if newTagText != '':
            if totalText == '':
                totalText = '#(' + newTagText + ')'
            else:
                totalText += '|#(' + newTagText + ')'

        # If no textDot already exists
        # New text dot is created
        # Text dot GUID is stored inside plane segment
        if totalText != '':
            if geomType in drawDot:
                if textDot == None:
                    textDot = rs.AddTextDot(totalText, midpt)
                    rs.SetUserText(textDot, 'visual', True)
                    rs.SetUserText(geometry, 'LabelDot', textDot)
                
                # If textDot already exists
                # Text dot text is updated
                # Text dot GUID is stored inside plane segment
                else: 
                    rs.TextDotText(textDot, totalText)
                    rs.SetUserText(geometry, 'LabelDot', textDot)    
                
                rs.ObjectColor(textDot, (0,0,0))
                self.objectIds.append(textDot)
            else:
                coords = rs.PointCoordinates(geometry)
                tempPt = rs.AddTextDot(totalText, (coords.X, coords.Y, coords.Z))
                color = rs.coercecolor(rs.ObjectColor(geometry))

                if color.R != 0 or color.G != 0 or color.R != 0:
                    rs.ObjectColor(tempPt, (color.R, color.G, color.B))
                    
                for text in rs.GetUserText(geometry):
                    rs.SetUserText(tempPt, text, rs.GetUserText(geometry, text))
                
                rs.DeleteObject(geometry)
                geometry = tempPt

        return geometry