#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r

# Python library imports
from copy import deepcopy
import warnings

# Sortal library imports
from sortal import sort, rule, sort, flow, form

# API support code imports
from .convert.convert_sortal_to_agnostic import convertS
from ..rhino_specific.visualAgnostic import visualA
from .move import translation_vector

# Conversion class initialization
convertS = convertS()

"""
=============
FUNCTION INFO
=============
'apply_flow' performs a flow operation based on the rules stipulated in the corresponding flow sequence. 
The flow string and the shape it is to be applied are retrieved from their respective registers, 
and the Rhino GUIDs of the final result of the flow are returned (and hidden from the Rhino Viewport). 
If the user wishes to have the Rhino GUIDs be visible, set the last value ('hide') to False.

INPUT
flowName: Name of flow inside flow register to be applied onto shape
shapeName: Name of shape inside form register that the flow will be applied onto
refPoint: Reference point to serve as 'origin' point for the final shape after the flow; the shape will be moved to this point; the default origin point is (0,0,0)
hide: Boolean True/False value if output Rhino GUIDs are to be hidden; True (default) - hides Rhino objects from viewport, False - leaves Rhino objects in viewport

OUTPUT
finalShapeRhino: Rhino GUIDs (hidden from Viewport)
None is returned if the flow sequence is unsuccesful

WARNINGS / ERRORS
KeyError: Shape name or flow name does not exist in their respective registers
          OR if the number of elements in the reference point tuple/list is not exactly three numbers 
          OR if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
TypeError: Reference point input is a not tuple/list of three numbers, a Rhino point geometry or a GUID 
           OR the inputs for shape name and/or flow name are not text strings
Warning: If the flow sequence could not be carried out successfully

Data Flow: flow name, shape name, reference point, hide (Boolean value) 
           -> retrieval of sortal data structures for flow, shape 
           -> application of flow onto shape
           -> final result as sortal form
           -> Rhino GUIDs (hidden from Viewport by default)
"""
                 
def apply_flow(flowName, shapeName, refPoint = r.Geometry.Point3d(0,0,0), hide = True):
    if 'rhino_shapes_c' in sort.register:
        if type(flowName) == type('a') and type(shapeName) == type('a'):
    
            if shapeName in form.formRegister:
                shape = form.formRegister[shapeName]
            else:
                raise KeyError('Shape named ' + shapeName + ' does not exist in either the form register.')
    
            if flowName in flow.register:
                flowObject = flow.register[flowName]
            else:
                raise KeyError('Flow named ' + flowName + ' does not exist in either the flow register.')
            
            if type(refPoint) == type ([]) or type(refPoint) == type([]):
                if len(refPoint) == 3:
                    refPoint = r.Geometry.Point3d(refPoint[0], refPoint[1], refPoint[2])
                else:
                    raise KeyError('Insufficient number of elements for reference point list/tuple input; the input should have exactly three number elements.')
            elif type(refPoint) != type(()) and type(refPoint) != type([]) and type(refPoint) != r.Geometry.Point3d:
                try:
                    test = rs.ObjectType(refPoint)
                except:
                    raise TypeError('Invalid input for reference point; input should be a tuple/list of numbers, a Rhino point or GUID.')
                    
            finalShape = flowObject.apply(shape)
            
            if type(finalShape) == type(True):
                warnings.warn('Flow sequence was unsuccessful.')
                return None
                
            finalShapeAgnostic = convertS.toAgnostic(finalShape)
            finalShapeDraw = visualA(finalShapeAgnostic)
            finalShapeRhino = finalShapeDraw.plotAgnostic()
            bbox = rs.BoundingBox(finalShapeRhino)
            translationVec = translation_vector(refPoint, bbox, 0)
                    
            if translationVec != None:
                # Moves geometry to location with respect to reference point
                finalShapeRhino = rs.MoveObjects(finalShapeRhino, translationVec)
            
            if hide:
                rs.HideObjects(finalShapeRhino)
                
            return finalShapeRhino
            
        else:
            if type(flowName) != type('a'):
                raise TypeError('Flow name input is not a string.')
                
            if type(shapeName) != type('a'):
                raise TypeError('Flow name input is not a string.')
                
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
