#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Python imports
import warnings

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r

# Sortal library imports
from sortal import sort, form

# API support code imports
from .convert.convert_sort_and_shape import convertShape

"""
=============
FUNCTION INFO
=============
'convert_shape' converts the contents of sortalShape (a sortal shape data structure) to their counterpart geometric sorts in targetSort.
All external, standalone descriptions as well as attributes attached to the geometries are carried over, so long as they are reflected in targetSort

SUPPORT FUNCTIONS
convertShape: This class of functions does the actual conversion from one sort type to another for a sortal shape

INPUT
sortalShape: The name of the shape or the sortal shape data structure to be converted to the target sort in targetSort; 
             this may have a primitive, attribute or disjunctive sort type; if the input has a compound sort type and there is only one functioning geometric disjunctive
             sort inside the compound sort, then the shape corresponding to this functioning geometric disjunctive sort is extracted and converted to the target sort
targetSort: The name of the target sort type or the sort type data structure that is the end goal of the conversion; this may be a primitive, attribute or disjunctive sort;
            if the target sort type is a compound sort, an error will be raised.
To illustrate possible conversion cases:
    a. sortalShape: 'shape_1' with sort type meta3D, targetSort: metaP2D; this will convert the geometric individuals of 'shape_1' to P2D geometric individuals,
    and retain any attribute types carried over to metaP2D. However, if, for example, meta3D has line segments with labels, while metaP2D has primitive line segments only,
    then the labels attached to line segments in 'shape_1' will not be maintained.
    b. sortalShape: 'line1' with sort type lineSegP3D, targetSort: lineSeg2D; this will construct the shape 'line1' as non-parametric, 2D line segments. The same
    behavior regarding attribute maintenance applies here as in (a).
    c. Sort type: ellipticalArc3D, targetSort: 'P3D'; this is not possible, as elliptical arcs are not enabled for parametric behavior in the sortal library.
    An error wil be raised
    d. Sort type: meta3D (disjunctive, including pointN3D), targetSort: pointP2D; in this case, only the pointP3D form will be retrieved from meta3D and turned to its non-parametric 2D counterpart
    and outputted.
    e. Sort type: rhino_shapes_c, with only one functioning geometric disjunctive sort (N3D) because the other sort type present is a dummy disjunctive sort composed of two description sorts;
    targetSort: P2D; this will convert the shape inside sortalShape corresponding to N3D to P2D
newName: The name of the converted shape that will be registered in the shape and the form register (text string); if none is given, then the name of the original shape
         and the target sort are concatenated to create a name for the shape

OUTPUT
newSortalShape: Returns converted sortal shape with geometric individuals reflecting attribute structure, and non-parametric/parametric behavior of targetSort, if successful
Returns False, if unsuccessful

WARNINGS / ERRORS
TypeError: If the input for sortalShape, targetSort or newName is not a text string, or if targetSort is a compound sort
KeyError: If the sortal shape or target sort does not exist in the form/sort register
          OR if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
          OR if the shape entered has a compound sort type (only disjunctive and lower sort types are accepted)
Warning: If the sort type of sortalShape is already the same as the target sort

Data Flow: sortalShape (sortal shape/form), targetSort (sort type) -> 
           (-> if sortalShape/targetSort are given as text strings, they are retrieved from their registers)
           -> newSortalShape (sortal shape after conversion to targetSort)
"""
   
def convert_shape(sortalShape = None, targetSort = None, newName = None):
    if 'rhino_shapes_c' in sort.register:
        newSortalShape = False
    
        if type(newName) != type('a') and newName != None:
            raise TypeError('Input for name of converted shape is not a text string.')
        elif newName == None:
            if type(sortalShape) == type('a'):
                if type(targetSort) == type('a'):
                    newName = sortalShape +'_' + targetSort
                else:
                    newName = sortalShape +'_' + targetSort.name
            else:
                if type(targetSort) == type('a'):
                    newName = sortalShape.name +'_' + targetSort
                else:
                    newName = sortalShape.name +'_' + targetSort.name
        
        if sortalShape != None and targetSort != None:
            if type(sortalShape) == type('a'):
                if sortalShape in form.formRegister:
                    shape = form.formRegister[sortalShape]
                    if shape.sort.type == 'sort-compound':
                        raise KeyError('Shape is of the type compound sort. Please only input shapes with disjunctive or lower sort types.')
                else:
                    raise KeyError('Shape with name ' + sortalShape + ' does not exist in the form register.')
            else:
                try:
                    test = sortalShape.sort
                    shape = sortalShape
                except:
                    raise TypeError('Input for shape to be converted is neither a text string nor a sortal shape data structure.')
                    
            if type(targetSort) == type('a'):
                if targetSort in sort.register:
                    target = sort.register[targetSort]
                else:
                    raise KeyError('Sort type ' + targetSort + ' does not exist in the sort register.')
            else:
                try:
                    test = targetSort.type
                    target = targetSort
                except:
                    raise TypeError('Input for target sort type is neither a text string nor a sort type data structure.')
            
            if target.type == 'sort-compound':
                raise TypeError('Target sort type cannot be a compound sort.')
                
            if shape.sort.name != target.name:
                newSortalShape = convertShape(shape, target).toTarget()
                newSortalShape.register(newName)
            else:
                warnings.warn('Shape is already the same sort as the target sort.')
                newSortalShape = shape
                
        return newSortalShape
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
