# -*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

"""
====
NOTE
====

Please include all these imports in any new .py files you create, as they are necessary for the Sortal GI functions
"""

# Rhino Scripting
import rhinoscriptsyntax as rs
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import time

# Import for transformations
#from mpmath import pi, sin, cos
from copy import deepcopy

# Sortal GI methods import
import sortalgi as sgi

# Rhino Methods import
from sortalgi import rhino_methods as rm

# Sortal Library import
from sortal import sdlParser, sort, rule, form

# Clears the sort and rule registers in the back end to prevent any sort/rule name conflicts
sort.register.clear()
rule.register.clear()

"""
===================================================================================
DEMO STRUCTURE: STORING DATA ON DESCRIPTIONS, PREDICATES AND DIRECTIVES IN GEOMETRY
===================================================================================
1. Reading SDL Files: Reading an SDL file and setting up sorts in the sort register
2. Adding data to Rhino geometry:
    a. Descriptions
    b. Predicates 
    c. Directives
3. Creating rules: Creating rules from Rhino GUIDs
4. Rule Application: Applying rules
5. SDL Writing: Saving rules and result to SDL file

NOTE: Please ensure that the sortalgi folder is on the file hierarchy level as this code

Notation:
    d(<text>): description text attached as attributes to geometry
    l(<text>): label text attached as attributes to geometry
    line(<text>) or plane(<text>) or point(<text>): tag or key name attached to geometry
    #(<text>): predicate/directive data attached to geometry

Available Predicates/Directives   
Predicates: Void, Max Line, Bound Line, Shortest Line, Longest Line, No Label
Directives: Point On Line, Distance, Normal
"""

"""
=================
READING SDL FILES
=================
Reads an SDL file and sets up the sort types inside the SDL file in the sort register
This also uploads any rules, shapes and flows (sequences of rules with additional looping and end conditions) to the sortal library registers for easy retrieval.
"""

print ('READING AN SDL FILE')
print ('Sort register before opening SDL file:')
print (sort.register)
print ()
print ('Rule register before opening SDL file:')
print (rule.register)
print ()

sgi.read_sdl('example_rules.sdl')

print ('Sort register after opening SDL file:')
print (sort.register)
print ()
print ('Rule register after opening SDL file:')
print (rule.register)
print ()
time.sleep(3)

"""
=======================
ADDING DATA TO GEOMETRY
=======================
Adds descriptions/tag or key names/predicate or directive data to geometry. 
As can be seen in the viewport, certain descriptions or labels
are already present in the geometry beforehand.

As one may also observe, the geometries of the rule sides lie on different layers. These layers
are named accordingly for the sort types the user wishes the target geometry to be classified under upon
conversion to its sortal shape counterpart.

Note: Please refer to the file 'Example rules.docx' for a detailed illustration of the predicates and directives
involved in each example
"""

# WHL-1a
# Predicate/s to be added: Void
# Directive/s to be added: Distance

# Step 1: Add void predicate to LHS
# The user is prompted for a list of geometries by the function rm.add_void()
# Place the void predicate on the LHS shape by selecting all lines (not including the reference point)
# Note that the lines do not need to define an entirely closed polygon

# Syntax: rm.add_void(sortType = [], detectIntersection = False)
# sortType: List of sort types that must not be present within void area, inputted as a list of text strings
# detectIntersection: False - goes ahead with predicate data addition even if geometry is self-intersecting
#                     True - If you wish to avoid selecting self-intersecting geometries when using add_void
rm.add_void()

# If there are no key names or tag names present within the geometry, then the user will be prompted to give 
# the highlighted object(s) a tag name.  This will be used to recognize it in the sortal library layer on.

# Step 2: Add distance directive to RHS
# This step adds the value computed by the expression "platform.wid / 2" ('platform' is a description sort type)
# as the exact distance range to generate the new line in the RHS at.

# Syntax: rm.add_distance(dist1, dist2 = None, tagSelf = True)
# dist1: Lower end of range of values to generate new RHS geometry at
# dist2: Upper end of range of values to generate newe RHS geometry at
# tagSelf: True (default) - prompts user for tag/key name for geometry if the geometry does not yet have one; 
#          False - does not prompt user for tag/key name for geometry even if the geometry does not have one
# Note: if only one value is given (i.e. dist1 only), then the new RHS geometry is generated exactly at that distance
#       if two values are given, then the new RHS geometry may be generated anywhere within the range of values stated

# In this case, select the line segment with the text 'line(0)' ('lineSeg-0' in the sortal library) in the LHS, 
# and then one of the points lying on the plane segment with the tag 'plane(0)' ('planeSeg-0' in the sortal library) in the RHS
rm.add_distance("platform.wid / 2")

"""
============================================
RULE CREATION WITH PREDICATES AND DIRECTIVES
============================================
The rule that these two sides pertain to will now be created. 
The predicates from the LHS and the directives from the RHS are added to the rule by including them as inputs
to the rule creation function sgi.create_shape.
"""
shapeL = rs.GetObjects('Select LHS of rule WHL-1a')
predL, dirL = sgi.create_shape('shapeL', shapeL)

shapeR = rs.GetObjects('Select RHS of rule WHL-1a')
predR, dirR = sgi.create_shape('shapeR', shapeR)

ruleName = 'WHL-1a'
ruleDesc = 'Void-distance example'
# Use predicates from LHS (variable predL) and directives from RHS (variable dirR)
sgi.create_rule(ruleName, ruleDesc, 'shapeL', 'shapeR', predL, dirR, prnt = True)

print ('Rule register after rule creation of WHL-1a: ', rule.register, '\n')

"""
===============================================
ADDING OTHER TYPES OF PREDICATES AND DIRECTIVES
===============================================
The process for this rule is similar to that of the previous rule's. 
"""

# DIY-4
# Predicate/s to be added: Shortest Line, Longest Line
# Directive/s to be added: Point on Line

# Step 1: Add shortest line predicate to desired line segment, followed by longest line predicate
# The user is prompted for a list of geometries to add the predicate to
# Select the line segments in question (if geometries of other types are selected, then the function will raise an error

# Syntax: rm.add_shortest_line(tagSelf = True)
# tagSelf: True (default) - prompts user for tag/key name for geometry if the geometry does not yet have one; 
#          False - does not prompt user for tag/key name for geometry even if the geometry does not have one
# In this case, select the line with the text 'line(1)' in the LHS (this means it is tagged as 'lineSeg-1' inside the sortal library)
rm.add_shortest_line()
# If there are no key names or tag names present within the geometry, then the user will be prompted to give 
# the highlighted object(s) a tag name.  This will be used to recognize it in the sortal library layer on.

# Step 2: Add point on line directive to RHS
# This step adds the values 0.3 and 0.7 as the lower end and upper end of the range of values to generate the location
# of the new point on the line segment in the RHS
# The user is prompted for the line segment(s) to add this directive to
# Select the line(s) with the text 'l(A)' (meaning this line segment is labelled with "A"; also, this has no tag/key name, so you will be prompted for this)
rm.add_point_on_line(0.3, 0.7)
# This will add the directive to that particular line segment

# Step 3: Rule creation (this is similar to the case for WHL-1a)
shapeL = rs.GetObjects('Select LHS of rule DIY-4')
predL, dirL = sgi.create_shape('shapeL', shapeL)

shapeR = rs.GetObjects('Select RHS of rule DIY-4')
predR, dirR = sgi.create_shape('shapeR', shapeR)

ruleName = 'DIY-4'
ruleDesc = 'Short-long-PoL example'
sgi.create_rule(ruleName, ruleDesc, 'shapeL', 'shapeR', predL, dirR, prnt = True)

print ('Rule register after rule creation of DIY-4: ', rule.register, '\n')