#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# API support code imports
from convert_sortal_to_agnostic import convertS
from convert_agnostic_to_sortal import convertA
from convert_rhino_to_agnostic import convertR
from visualAgnostic import visualA as visualA
from ..sortal_api.rhino_specific_methods import rhino_methods as rm

# Rhino imports
import Rhino as r
import rhinoscriptsyntax as rs
import scriptcontext as sc
import Rhino.Geometry as rg

# Conversion code initialization
convertA = convertA()
convertR = convertR()
convertS = convertS()

"""
=============
FUNCTION INFO
=============
'convert_to_agnostic' serves as a SortalGI call to the convertR.toAgnostic function located inside the file 'convert;, for ease of use.
It converts Rhino GUIDs to an agnostic dictionary based on the structure of the target sort type (this sort type must have already been created and should
be present in the register). It also checks if all the inputs are correct before proceeding.

INPUT
shapeRhino: List of Rhino GUIDs
classification: Target sort type for the shape
descriptions: Description declaration text string 
              Example: 'label1@("A2", 1, 1234);("A3", 4, 1234)|label2@("A1", 1, 1234)|label2@("A9", 8, 1234)'
              where there are two description types, 'label1' (followed by an ampersand @; 2 individuals separated by a semicolon ;),         
              and 'label2' (2 individuals); declaration of different description types and their individuals is separated by a vertical dash |
refPoint: Reference point which the coordinates of the shape will be subtracted from; this functions as the 'origin' of the shape, and can be used for defining
          shapes that compose rule sides

OUTPUT
Four outputs, in order:
    Agnostic object form of [Rhino GUIDS + description type; classification is used as the key for the dictionary if there is more than one type of disjunctive
    sort within 'rhino_shapes_c' that is not a dummy disjunctive sort] as agnostic dictionary
    Dictionary of GUIDs corresponding to geometry stored in agnostic object, according to their target sort type
    Dictionary of predicates
    Dictionary of directives
    
WARNINGS / ERRORS
TypeError: If shapeRhino is not a list of Rhino GUIDs OR if descriptions is not None or a text string
Warning: If conversion was unsuccessful
KeyError: If 'rhino_shapes_c' is not present in the sort register

Data Flow:  Rhino GUIDs 
            -> disjunctive agnostic dictionary
"""
   
def convert_to_agnostic(shapeRhino = [], classification = None, descriptions = '', refPoint = r.Geometry.Point3d(0,0,0)):
    if 'rhino_shapes_c' in sort.register:
        if (type(shapeRhino) == type([]) or type(shapeRhino) == type({})) and (type(descriptions) == type('a') or type(descriptions) == type(None)) and (type(classification) == type('a') or classification == None):
            agnosticObject, shapeIDs, predicates, directives = convertR.toAgnostic(shapeRhino, classification, descriptions, refPoint)
            return agnosticObject, shapeIDs, predicates, directives
            
        else:
            if type(shapeRhino) != type([]):
                raise TypeError('Rhino GUIDs input is not a list')
                
            if type(descriptions) != type('a') or type(descriptions) != type(None):
                raise TypeError('Description(s) input is not a text string')
            
            warnings.warn('Conversion from Rhino GUIDs to agnostic object was unsuccessful')
            return None, None, None, None
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
            

"""
=============
FUNCTION INFO
=============
'create_shape_ag' creates a sortal shape from a collection of Rhino GUIDs. It names the shape according to the input for the shapeName variable in the sortal library.
It returns the predicate/directive dictionaries. 

INPUT
shapeName: Name of shape (text string)
shapeData: List of Rhino GUIDs or an agnostic dictionary that will compose the sortal shape
descriptions: Text string of descriptions to include in the sortal shape (if these are not yet present in the agnostic dictionary), e.g.
              'label1@("A2", 1, 1234);("A3", 4, 1234)|label2@("A1", 1, 1234)|label2@("A9", 8, 1234)'
              where there are two description types, 'label1' (followed by an ampersand @; 2 individuals separated by a semicolon ;),         
              and 'label2' (2 individuals); declaration of different description types and their individuals is separated by a vertical dash |
classification: Name of target disjunctive sort type (text string); this sort type must already be present in the sort register 
                under the compound sort 'rhino_shapes_c'
refPoint: Reference point to serve as 'origin' point for shape; the default origin point is 0,0,0
prnt: Boolean value (True/False); True - prints the resulting sortal shape according to the output of the 'printE' command in the sortal library,
      False (default) - does not print anything

OUTPUT
predicates: Predicates dictionary
directives: Directives dictionary

WARNINGS / ERRORS
TypeError: If shapeName input is not a text string, or if shapeData input is not an agnostic dictionary or a list of Rhino GUIDs, 
           OR if shapeData has any elements that are not Rhino GUIDs inside
Warning: If descriptions input is not a text string
KeyError: If 'rhino_shapes_c' is not present in the sort register

Data Flow: Rhino GUIDs + description text string -> sortal shape -> register shape in form register -> print output (optional) 
           -> return predicates, directives dictionaries
"""

def create_shape_ag(shapeName = None, shapeData = None, descriptions = '',  classification = None, refPoint = rg.Point3d(0,0,0), prnt = False):
    if 'rhino_shapes_c' in sort.register:
        if type(shapeName) != type('a'):
            raise TypeError('Shape name is not a text string') 
            
        if type(shapeData) == type({}) or type(shapeData) == type([]):
            # Handles agnostic dictionary input
            if type(shapeData) == type({}):
                # Converts shape tuple to sortal metaform
                shapeMetaform = convertA.toSortal(shapeData)
                predicates = {}
                directives = {}
                
            # Handles Rhino GUIDs list input
            elif type(shapeData) == type([]):
                try:
                    rs.ObjectType(shapeData[0])
                except:
                    raise TypeError('Shape data is not a list of Rhino GUIDs')
    
                if type(descriptions) != type('a'):
                    warnings.warn('Descriptions input is not a text string. Descriptions will not be included in the sortal shape.')
                    descriptions = ''
                   
                if refPoint is None:
                    refPoint = rg.Point3d(0,0,0)
    
                shapeAg, shapeIDs, predicates, directives = convert_to_agnostic(shapeData, classification, descriptions, refPoint)
                shapeMetaform = convertA.toSortal(shapeAg)
       
            # Registers shape object name
            shapeMetaform.register(shapeName)
            
            if prnt:
                # Outputs complete sortal metaform (geometry + external descriptions) using sortal form printE() function
                print('The following sortal shape (named ' + shapeMetaform.name + ') was created:\n')
                shapeMetaform.printE()
    
            return predicates, directives 
        
        else:
            raise TypeError('Shape agnostic object is neither an agnostic dictionary nor a list of Rhino GUIDs')
        
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')

    
"""
=============
FUNCTION INFO
=============
'maximalize_ag' returns a maximalized agnostic dictionary/list of Rhino Geometry

INPUT
shape: Shape data in the form of a list of Rhino GUIDs (target disjunctive sort for maximalized must be provided) 
       OR an agnostic dictionary OR the shape name as a text string
target: Target disjunctive sort for shape to be maximalized (text string)
rhino: Boolean value (True/False); True - returns list of Rhino GUIDs corresponding to maximalized shape; False (default) - returns agnostic dictionary form
hide: Boolean value (True/False); True - hides Rhino geometry output from viewport; False (default) - leaves Rhino geometry output visible in viewport
delete: Boolean value (True/False); if the inputs are a list of Rhino GUIDs, then True deletes the inputs, 
        and False (default) leaves the inputs still in the Rhino viewport

OUTPUT
maxShape: Maximalized agnostic dictionary/list of Rhino geometries

WARNINGS / ERRORS
TypeError: If type of input is invalid (i.e. not a list of Rhino GUIDs, an agnostic dictionary or a shape name text string)
ValueError: If input is an empty list or if shape name doest not exist in the form register
KeyError: If 'rhino_shapes_c' is not present in the sort register
    
Data Flow: agnostic/GUIDs -> sortal -> agnostic/GUIDs
"""

def maximalize_ag (shape, target = None, rhino = False, hide = False, delete = False):
    if 'rhino_shapes_c' in sort.register:
        # Maximalizing agnostic dictionaries
        if type(shape) == type({}):
            shapeSortal = convertA.toSortal(shape)
            shapeSortal.maximalize()
            maxShape = convertS.toAgnostic(shapeSortal)
        
        # Maximalizing list of Rhino GUIDs
        elif type(shape) == type([]):
            try:
                rs.ObjectType(shape[0])
            except:
                if len(shape) > 0:
                    raise TypeError('Shape input is not composed of Rhino GUIDs')
                else:
                    raise ValueError('Shape input is empty')
                    
            shapeAgs = convert_to_agnostic(shape, target)[0]
            shapeSortal = convertA.toSortal(shapeAgs)
            shapeSortal.maximalize()
            maxShape = convertS.toAgnostic(shapeSortal)
            rhino = True
        
        elif type(shape) == type('a'):
            if shape in form.formRegister:
                shapeSortal = form.formRegister[shape]
                shapeSortal.maximalize()
                maxShape = convertS.toAgnostic(shapeSortal)
            else:
                raise ValueError('Shape with name ' + shape +' is not available in the sort register')
       
        else:
            raise TypeError('Invalid input for maximalization')
                
        if rhino:
            draw_shape = visualA(maxShape)
            maxShape = draw_shape.plotAgnostic()
            
            if delete:
                rs.DeleteObjects(shape)
                
            if hide:
                rs.HideObjects(maxShape)
            
        return maxShape
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
