#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# API support code imports
from .convert.convert_agnostic_to_sortal import convertA
from .convert.convert_sortal_to_agnostic import convertS
from ..rhino_specific.visualAgnostic import visualA
from ..rhino_specific.rhino_specific_methods import rhino_methods as rm
from .part_of import part_of

# Sortal library imports
from sortal import rule, form, sort

# Conversion code initializations
convertA = convertA()
convertS = convertS()

"""
=============
FUNCTION INFO
=============
'apply_all_together' finds rule applications from a subshape and shape, or from a set of GUIDs,
obtains the sum of the results of all the rule applications, and returns a list of Rhino GUIDs corresponding to the sum.
In the case where only one rule application is found, its Rhino GUIDs is returned.

INPUT
Required
chosenRule : Rule name as text string
shape: Name of main shape as text string as recorded in sortal library's form register pertaining to shape

Optional
subshape: Name of subshape as text string as recorded in sortal library's form register pertaining to subshape
refPt: GUID of point, tuple, or Rhino Geometry point of reference point which will serve as the 'origin' with which shapes
       will be plotted in respect to (in principle, a vector)
layerName: Name of target layer as text string within Rhino workspace where the GUIDs will be drawn onto
prnt: Boolean value True/False; True - prints out description individuals as input-ready text string (to create_shape)
      as well as printE form of shape after rule application; False (default) - nothing is printed
shapeIds: List of Rhino GUIDs to clear after new shapes have been drawn

OUTPUT
appnGeometry: List of Rhino GUIDs corresponding to sum of rule applications
None, if no rule applications are found

WARNINGS / ERRORS
TypeError: if initial shape or subshape is not a list of GUIDs or a text string or is empty, or if the input for chosenRule is not a text string
KeyError: if initial shape or subshape input's name is not present in the form register, or if rule name is not present in the rule register
Error: If 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: if rule applications found are less than two, then the singular rule application is returned as the result

Data Flow:  Names of rule, shape
            -> detect rule applications within shape
            -> obtain sortal rule applications
            (-> rule applications based on subshape, subshapeAgnostic if subshape is specified)
            -> sum up results of rule applications into one sortal form
            -> draw sum in Rhino and construct its external descriptions text string
            -> return Rhino GUIDs of sum, print external description text string if prnt set to True
            
Pseudocode: Names of rule, shape, subshape (optional) -> retrieve rule and shape from their registers
            
            -> rule application detection -> check if number of rule applications generated > 0
            
            -> 
            
"""

def apply_all_together(chosenRule, shape, subshape = None, refPt = r.Geometry.Point3d(0,0,0), layerName = 'Default', prnt = False, shapeIds = None):
    if 'rhino_shapes_c' in sort.register:
        sum = None

        # RETRIEVAL OF RULE OBJECT
        # Checks if rule input is a a valid string input that corresponds to a rule in the rule register
        chosenRule = retrieve_structure(chosenRule, 'r')

        ruleAppns = []          # List of rule applications to be returned (either as agnostic dict or Rhino GUIDs)
        sum = None
        subshapeSortal = None

        # Text string reference to shape
        # Retrieves shape of the corresponding name from the sortal form register
        shapeSortal = retrieve_structure(shape, 's')
        
        # RETRIEVAL OF SUBSHAPE
        # Checks if subshape is valid (optional)
        # Retrieves subshape from sortal form register if necessary
        subshapeSortal = retrieve_structure(subshape, 's')

        # COMPATIBILITY BETWEEN SUBSHAPE AND SHAPE
        # If subshape is provided
        if subshapeSortal != None:

            # Checks if subshape is part of main shape before proceeding
            # FIX THIS
            if subshapeSortal.partOf(shapeSortal):
                # RULE APPLICATION DETECTION
                subshapeApps = chosenRule.detect(subshapeSortal)       # Generates matches, results and rule applications for subshape

                if len(subshapeApps) > 1:
                  # RULE APPLICATION ONTO ENTIRE SHAPE
                  for appn in subshapeApps:
                      after = appn.perform(shapeSortal.duplicate())  # Entire shape after rule application
                      ruleAppns.append(after)                        # Append sortal form to list of results
                else:
                  if len(subshapeApps) == 1:
                    warnings.warn('There is only one rule application detected in the shape. The GUIDs of the drawing of this shape will be returned.')
                  else:
                    warnings.warn('No rule applications found')
                    return None
            else:
                warnings.warn('Subshape is not part of main shape. Matches cannot be found. Exiting function.')
                return None

        # Subshape is not provided
        else:
            # RULE APPLICATION DETECTION
            shapeApps = chosenRule.detect(shapeSortal)       # Generates matches, results and rule applications for subshape

            if len(shapeApps) > 1:
              # RULE APPLICATION ONTO ENTIRE SHAPE
              for appn in shapeApps:
                  after = appn.perform(shapeSortal.duplicate())  # Entire shape after rule application
                  ruleAppns.append(after)                        # Append sortal form to list of results
            else:
              if len(shapeApps) == 1:
                warnings.warn('There is only one rule application detected in the shape. The GUIDs of the drawing of this shape will be returned.')
              else:
                warnings.warn('No rule applications found')
                return None

        # Sums results together to get sum form
        sum = ruleAppns[0].duplicate()

        for result in ruleAppns[1:]:
          sum.sum(result)

        # Converts sum to its agnostic dictionary form
        sumAgnostic = convertS.toAgnostic(sum)
        appnGeometry = []

        # PLOTTING OF SUM and PRINTING OF DESCRIPTIONS INSIDE
        vis = visualA(sumAgnostic, refPt, layerName)
        newShapeIDs = vis.plotAgnostic()                       # Redraws shape after rule application based on agnostic dictionary input 'ruleAppn'
        appnGeometry.append(newShapeIDs)
        descriptions = None

        if 'descriptions' in sumAgnostic:
            description_text = sumAgnostic['descriptions']
            descriptions = construct_description(description_text)

        if prnt:
            # Display text
            print ('Descriptions inside sum of rule applications for ' + chosenRule.name + 'onto ' + shape + ':')
            if descriptions != None:
                print ('\nApplication description(s) text string')
                print (descriptions + ' \n')

            convertA.toSortal(sum).printE()
            print('\n')

        # CLEARING SHAPE
        # Clears previous shape IDs inside frame; input is a list of GUIDs
        if type(shapeIds) != type(None):
            rm.clearShape(shapeIds)

        return appnGeometry

    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.')

"""
================
SUPPORT FUNCTION
================
'construct_description' constructs the string text for any external, standalone descriptions present in a rule application

INPUT
descriptions: Dictionary containing description individuals grouped according to the name of their description sort type

OUTPUT
description_text: Text string assembling description individuals based on their description sort type and data; this text string
                  may be copy-pasted as direct input for future shape creation
"""

def construct_description(descriptions):
    description_text = ''

    for desc_name in descriptions:
        desc_text = descriptions[desc_name]

        if description_text == '':
            description_text = desc_name + ':'
        else:
            description_text = description_text+ '|' + desc_name + ':'

        for count, desc in enumerate(desc_text):
            if count == 0:
                description_text = description_text + desc
            else:
                description_text = description_text +';' + desc

    return description_text

"""
================
SUPPORT FUNCTION
================
'retrieve_structure' retrieves a shape (form), subshape (form), rule object (rule) or flow (flow object)
 based on its name in its corresponding sort register:
    + shape, subshape: form.formRegister
    + rule object: rule.register

INPUT
name: Name of shape, subshape, 
sortalType: String character indicating type of sortal data structure to retrieve, i.e.
            + 's' - shape or subshape (form)
            + 'r' - rule object
            + 'f' - flow object

OUTPUT
Sortal data structure, based on the item stored in the corresponding sortal library register
"""

def retrieve_structure(name, sortalType = 's'):
    chosenObject = None
    
    if type(name) == type('a') and type(sortalType) == type('a'):
        if sortalType == 's' or sortalType == 'r' or sortalType == 'f':
            if sortalType == 's':
                if name in form.formRegister:
                    chosenObject = form.formRegister[name]
                else:
                    raise KeyError(('Shape with the name ' + name + ' does not exist in the sortal form register.'))
                    
            elif sortalType == 'r':
                if name in rule.register:
                    chosenObject = rule.register[name]
                    
                else:
                    raise KeyError('Rule name ' + name + ' is not present in the rule register.')
                
            elif sortalType == 'f':
                if name in flow.register:
                    chosenObject = flow.register[name]
                
                else:
                    raise KeyError('Flow name ' + name + ' is not present in the flow register.')
                    
            return chosenObject
        else:
            raise ValueError('Only the inputs "s" (shape object - form), "r" (rule object) and "f" (flow object) are accepted.')
    else:
        if type(name) != type('a'):
            if sortalType == 's':
                raise TypeError('Input for name of shape object is not a text string')
            elif sortalType == 'r':
                raise TypeError('Input for name of rule object is not a text string')
            elif sortalType == 'f':
                raise TypeError('Input for name of flow object is not a text string')
            
        if type(sortalType) != type('a'):
            raise TypeError('Input for type of sortal data structure is not a text string')
            