#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r

# Python library imports
import warnings

# Sortal library imports
from sortal import sort, form

"""
=============
FUNCTION INFO
=============
'extract_shape' searches for a shape inside the input sortalShape that matches the sort type of the value of targetSortType and returns this sortal shape.
It uses a method called search_shape to look for the shape inside the given shape input with the same sort type as the target sort type.

SUPPORT FUNCTIONS
search_shape: Searches shape for shape with similar sort type as target sort by going inside shape recursively

INPUT
sortalShape: Name of shape (text string) or sortal shape data structure
targetSortType: Name of sort type (text string) or sort type data structure
shapeName: Name of extracted shape; this is used to register the extracted shape in the form register, if it is successfully extracted

OUTPUT
result: Sortal shape (actual sortal shape data) corresponding to targetSortType and registered in the form register, if successful
None, if no such shape with the same sort as the target sort type can be found inside the input

WARNINGS / ERRORS
TypeError: If the input for sortalShape / targetSortType (/shapeName) is neither a text string 
           or a sortal shape data structure/sort type data structure
KeyError: If sortalShape or targetSortType is a text string input, this error is raised if they do not exist in the form/sort register
          OR if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: If sortalShape does not contain any sort type equal to targetSortType

Data Flow: Text strings or sortal data structures 
           -> sortal data structure, if successful
           (-> None, if unsuccessful)
"""
   
def extract_shape(sortalShape, targetSortType, shapeName = None):
    if 'rhino_shapes_c' in sort.register:
        # Checks for validity of sortalShape input            
        if type(sortalShape) == type('a'):
            if sortalShape in form.formRegister:
                shape = form.formRegister[sortalShape]
            else:
                raise KeyError('Shape named ' + sortalShape + ' does not exist in the form register.')
        else:
            try:
                test = sortalShape.sort
                shape = sortalShape
            except:
                raise TypeError('Invalid input for sortal shape; this must be either the name of a sortal shape present in the form register or a sortal shape data structure.')
    
        # Checks for validity of shapeName input
        # If none is given, then a default name is generated, comprised of sortalShape's name and the add-on '_subshape'
        if type(shapeName) != type('a') and shapeName != None:
            raise TypeError('Shape name input is not a text string')
        elif shapeName == None:
            if type(sortalShape) == type('a'):
                shapeName = shape.name + '_subshape'
            else:
                shapeName = 'new_subshape'
    
        targetSort = targetSortType
        
        if type(targetSort) == type('a'):
            if targetSortType in sort.register:
                targetSort = sort.register[targetSort]
            else:
                 raise KeyError('Sort named ' + targetSortType + ' does not exist in the sort register.')
        else:
            try:
                test = targetSort.type
            except:
                raise TypeError('Invalid input for target sort type; this must be either the name of a sort type present in the sort register or a sort type data structure.')
                
        if shape.sort is targetSort:
            warnings.warn('Shape input is already of the sort type ' + targetSort.name)
            return shape
        else:
            result = search_shape(shape, targetSort)
            if result == None:
                warnings.warn('Shape named ' + shape.name + ' does not contain the sort type ' + targetSortType + '.')
            else:
                form.formRegister[shapeName] = result
                result.register(shapeName)
                
            return result
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
            
"""
================
SUPPORT FUNCTION
================
'search_shape' looks for the shape inside the given shape input with the same sort type as the target sort type.

INPUT
shape: Sortal shape (form) to search for the target sort type in
targetSort: Sortal data structure corresponding to the sort type to search for in the shape

OUTPUT
Returns the sortal shape with the same sort type as targetSort, if successful
Returns None, if this cannot be found
"""

def search_shape(shape, targetSort):
    shapeSort = shape.sort
    shapeType = shape.sort.type

    if shapeType == 'sort-primitive' or shapeType == 'sort-attribute':
        if shapeSort == targetSort:
            return shape
            
        return None
        
    elif shapeType == 'sort-disjunctive':
        if shapeSort == targetSort:
            return shape
        else:
            for element in shape.elements:
                if search_shape(element, targetSort):
                    return search_shape(element, targetSort)
        
        return None
        
    elif shapeType == 'sort-compound':
        if shapeSort == targetSort:
            return shape
        else:
            for element in shape.elements:
                if search_shape(element, targetSort):
                    return search_shape(element, targetSort)
