#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import scriptcontext as sc

# Python built-ins
from builtins import range

# Sortal Library Imports
from sortal import (transformSet, rule, primitiveSort, coordinate, label, 
                    rWeight, disjunctiveSort, attributeSort, affineTransform, compoundSort,  
                    vector3D, vector2D, enumerative, sort, description,
                    point2D, point3D, pointP2D, pointP3D,
                    lineSegment2D, lineSegment3D, lineSegmentP2D, lineSegmentP3D,
                    planeSegment2D, planeSegment3D, planeSegmentP2D, planeSegmentP3D,
                    bezierCurve2D, bezierCurve3D, bezierCurveP2D, bezierCurveP3D)
                
# User-defined sort type imports from sort_types
from .sortTypes import (pointSort2D, pointSort3D, pointSortP2D, pointSortP3D, 
                        lineSegSort2D, lineSegSort3D, lineSegSortP2D, lineSegSortP3D, 
                        planeSegSort2D, planeSegSort3D, planeSegSortP2D, planeSegSortP3D, 
                        c2D, c3D, cP2D, cP3D,
                        e2D, e3D, eP2D, eP3D,
                        cArc2D, cArc3D, eArc2D, eArc3D, 
                        bez2D, bezP2D, bez3D, bezP3D,
                        labelSort, descSort, ptDSort, lsDSort, plDSort, cDSort, cArcDSort, eDSort, eArcDSort, bezDSort,
                        weightSort,weight255Sort, colorSort, enumSort, planeSegSortP3D, pointSort3D, lineSegSort3D, planeSegSort3D)

# The class 'parse_disjunctive_sort' currently supports initializing the following sort types:
# GEOMETRIC SUPPORT - point, line segment
# ATTRIBUTE SUPPORT (True/False values in this order) - label, description, weight, weight255, enumerative, color

class parse_disjunctive_sort:
    
    # Creates sort types for metaform and forms of individuals to be placed inside the metaform
    # Based on user input
    def parseDisjSort(self, point = (False, False, False, False, False, False, False), 
                            line = (False, False, False, False, False, False, False), 
                            planeSeg = (False, False, False, False, False, False, False),
                            circle = (False, False, False, False, False, False, False),
                            ellipse = (False, False, False, False, False, False, False),
                            cArc = (False, False, False, False, False, False, False),
                            bezier = (False, False, False, False, False, False, False),
                            descriptions = False, parametric = False, dimension = False):

        check = True
        mainSorts = []   # Completed sorts
        attribSorts = [] # Attribute sorts for corresponding index of completed sorts
        geomSorts = [point, line, planeSeg, circle, ellipse, cArc, bezier, descriptions]
        self.parametric = parametric
        self.classification = 'N3D'
        disjList = []   # Collects all disjunctives for creation of compound sort
        
        # Check if any pre-existing rhino_shapes disjunctive sorts exist
        # Used to create compound sort later on too
        #for sortType in sort.register:
        #    if sort.register[sortType].__class__ == disjunctiveSort and ('2D' in sortType or '3D' in sortType):
        #        disjList.append(sort.register[sortType])
            
        if self.parametric:
            if dimension == True:
                self.pointSort = pointSortP2D
                self.lineSegSort = lineSegSortP2D
                self.planeSegSort = planeSegSortP2D
                self.cSort = cP2D
                self.eSort = eP2D
                self.bezSort = bezP2D
                self.classification = 'P2D'
                
            else:
                self.pointSort = pointSortP3D
                self.lineSegSort = lineSegSortP3D
                self.planeSegSort = planeSegSortP3D
                self.cSort = cP3D
                self.eSort = eP3D
                self.bezSort = bezP3D
                self.classification = 'P3D'

        else:
            if dimension == True:
                self.pointSort = pointSort2D
                self.lineSegSort = lineSegSort2D
                self.planeSegSort = planeSegSort2D
                self.cSort = c2D
                self.cArcSort = cArc2D
                self.eSort = e2D
                self.eArcSort = eArc2D
                self.bezSort = bez2D
                self.classification = 'N2D'
                
            else:
                self.pointSort = pointSort3D
                self.lineSegSort = lineSegSort3D
                self.planeSegSort = planeSegSort3D
                self.cSort = c3D
                self.cArcSort = cArc3D
                self.eSort = e3D
                self.eArcSort = eArc3D
                self.bezSort = bez3D
                
        disjName = self.classification
        
        for ind, content in enumerate(geomSorts):
            if ind < 7:
                if content[0]:
                    fullSortName = ''      # Name of full sort 
                    try:
                        baseSort, fullSortName = self.mainSort(ind, fullSortName)        # Base sort which becomes fullSort if no attributes
                    
                    # In cases where a sort type is not yet enabled in the sortal library (cArc and eArc parametric, 2D/3D)
                    except:
                        continue
  
                    attributes = content[1:]
                    attribSort, fullAttribSort = self.attributeSorts(attributes, baseSort)
                    if self.classification == '3D':
                        if 'point' in fullSortName:
                            fullSortName = 'points'
                        elif 'lineSeg' in fullSortName:
                            fullSortName = 'line_segments'
                        elif 'planeSeg' in fullSortName:
                            fullSortName = 'plane_segments'
                        elif 'circle' in fullSortName:
                            fullSortName = 'circles'
                        elif 'ellipse' in fullSortName:
                            fullSortName = 'ellipses'
                        elif 'cArc' in fullSortName:
                            fullSortName = 'arcs'
                        elif 'bez' in fullSortName:
                            fullSortName = 'bezier_curves'

                    # base sort - geometric base sort
                    # fullsortname - base sort + attribute sort (primitive or disjunctive)
                    
                    # attribName - name of attribute sort 
                    # attribSort - attribute sort components
                    # fullAttribSort - full disjunctive attribute sort
                    
                    # Checks if full attribute sort already exists
                    # If yes, takes previously determined attribute sort and assigns it to fullSort
                    # If no, creates attributeSort to assign to fullSort
                    if attributeSort.checkSort(fullSortName) == False:
                        fullSort = attributeSort(fullSortName, baseSort, fullAttribSort)
                    else:
                        fullSort = attributeSort.register[fullSortName]

                    mainSorts.append(fullSort)
                    attribSorts.append(attribSort)

            else:
                for descName in content:
                    baseSort, fullSortName = self.mainSort(ind, descName)        # Base sort which becomes fullSort if no attributes
                    mainSorts.append(baseSort)
                        
        # Creates disjunctive sort for metaform if it does not already in the register of sort type names

        if disjunctiveSort.checkSort(disjName) == False:             # If value of disjName corresponding sort not yet in sort register, creates sort
            fullDisjSort = disjunctiveSort(disjName, mainSorts)      # If already in sort register, retrieves sort
            
        else:
            fullDisjSort = disjunctiveSort.register[disjName]   
        
        disjList.append(fullDisjSort)
        
        __dummyDesc1__ = primitiveSort('__dummyDesc1__', description)
        __dummyDesc2 = primitiveSort('__dummyDesc2__', description)
        __dummyDisjSort__ = disjunctiveSort('__dummyDescDisjunctive__', [__dummyDesc1__, __dummyDesc2])
        disjList.append(__dummyDisjSort__)
        
        # Creates compound sort
        # Stores compound sort name in sc.sticky for easy retrieval in later functions
        self.createCompound(disjList)
        sc.sticky['mainSort'] = 'rhino_shapes_c'
        
        return fullDisjSort, mainSorts, attribSorts   
        
    # Assigns base sort and individual name
    def mainSort(self, index, fullSortName):
        # create recognition of pre-existing sorts!!! 
        # needed for composite sort creation
        if index == 0:
            baseSort = self.pointSort
            fullSortName += 'point'
        elif index == 1:
            baseSort = self.lineSegSort
            fullSortName += 'lineSeg'
        elif index == 2:
            baseSort = self.planeSegSort
            fullSortName += 'planeSeg'
        elif index == 3:
            baseSort = self.cSort
            fullSortName += 'circle'
        elif index == 4:
            baseSort = self.eSort
            fullSortName += 'ellipse'
        elif index == 5:
            if 'P' not in self.classification:
                baseSort = self.cArcSort
                fullSortName += 'cArc'
            else:
                print ('Parametric circular arcs are not yet enabled')
                return
        elif index == 6:
            baseSort = self.bezSort
            fullSortName += 'bezCurve'
        elif index == 7:
            if fullSortName not in sort.register:
                baseSort = primitiveSort(fullSortName, description)
            else:
                baseSort = sort.register[fullSortName]

        fullSortName = fullSortName + '-'+ self.classification
        return baseSort, fullSortName
        
    # Assigns attribute sort and attribute name
    def attributeSorts(self, attributes, baseSort):
        attribSort = []
        attribName = '' # Name for attribute sort
        fullAttribSort = None

        for indiv in range(len(attributes)): 
            if attributes[indiv]:
                if indiv == 0:
                    attribSort.append(labelSort)
                    attribName = attribName + 'label'
                    
                elif indiv == 1:
                    if 'point' in baseSort.name:
                        newDescSort = ptDSort
                        descName = 'ptDescription'
                    elif 'lineSeg' in baseSort.name:
                        newDescSort = lsDSort
                        descName = 'lsDescription'
                    elif 'planeSeg' in baseSort.name:
                        newDescSort = plDSort
                        descName = 'plDescription'
                    elif baseSort.name[0] == 'c' and 'arc' not in baseSort.name and 'Arc' not in baseSort.name:
                        newDescSort = cDSort
                        descName = 'cDescription'
                    elif baseSort.name[0] == 'e' and 'arc' not in baseSort.name and 'Arc' not in baseSort.name:
                        newDescSort = eDSort
                        descName = 'eDescription'
                    elif baseSort.name[0] == 'c' and ('arc' in baseSort.name or 'Arc' in baseSort.name):
                        newDescSort = cArcDSort
                        descName = 'cArcDescription'
                    elif baseSort.name[0] == 'e' and ('arc' in baseSort.name or 'Arc' in baseSort.name):
                        newDescSort = eArcDSort
                        descName = 'eArcDescription'
                    elif 'bezier' in baseSort.name:
                        newDescSort = bezDSort
                        descName = 'bezDescription'
                        
                    attribSort.append(newDescSort)
                    
                    if len(attribName) != 0:
                        attribName = attribName + '_' + descName
                    else:
                        attribName = attribName + descName

                elif indiv == 2:
                    attribSort.append(enumSort)
                    if len(attribName) != 0:
                        attribName = attribName + '_enumerative'
                    else:
                        attribName = attribName + 'enumerative'
                
                elif indiv == 3:
                    attribSort.append(weightSort)
                    if len(attribName) != 0:
                        attribName = attribName + '_weight'
                    else:
                        attribName = attribName + 'weight'
                        
                elif indiv == 4:
                    attribSort.append(weight255Sort)
                    if len(attribName) != 0:
                        attribName = attribName + '_weight255'
                    else:
                        attribName = attribName + 'weight255'
                        
                elif indiv == 5:
                    attribSort.append(colorSort)
                    
                    if len(attribName) != 0:
                        attribName = attribName + '_color'
                    else:
                        attribName = attribName + 'color'
            
        # Creates disjunctive sort if attributes > 1
        # Else, returns attribute sort already inside attribSort
        if len(attribSort) > 1:
            if disjunctiveSort.checkSort(attribName) == False:
                fullAttribSort =  disjunctiveSort(attribName, attribSort)
            else: 
                fullAttribSort = disjunctiveSort.register[attribName]
        else:
            fullAttribSort =  attribSort[0]

        return attribSort, fullAttribSort
    
    # Creates compound sort from first disjunctive sort/existing compound sort and new disjunctive sort
    def createCompound(self, newComponents):
        if 'rhino_shapes_c' in sort.register:
            if sort.register['rhino_shapes_c'].component:
                del sort.register['rhino_shapes_c']
                compoundSort('rhino_shapes_c', newComponents)
        else:
            compoundSort('rhino_shapes_c', newComponents)

# GUIDE
# The input of tuples is as follows
# <discursiveSort instance>.parseDiscSort( <point>, <line segment>, <plane>, <volume> )

# For any given tuple of True/False values, the order of value setting is as follows
# 0. Existence of individual
# 1. Label
# 2. Description
# 3. Enumerative - values and sorting order need to be specified by user/developer; prompt is built into code
# 4. Weight - weight sort, which accepts values from 0 to 1
# 5. Weight255 - weight sort, which accepts integer values from 0 to 255
# 6. Color - color sort, which accepts RGB values (e.g. (120, 30, 40))

# EXAMPLE
# For a point with label, description, weight255 and enumerative enabled
# (True, True, True, True, True, False)
