# -*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

"""
====
NOTE
====

Please include all these imports in any new .py files you create, as they are necessary for the Sortal GI functions
"""

# Rhino Scripting
import rhinoscriptsyntax as rs
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import time

# Import for transformations
#from mpmath import pi, sin, cos
from copy import deepcopy

# Sortal GI methods import
import sortalgi as sgi

# Rhino Methods import
from sortalgi import rhino_methods as rm

# Sortal Library import
from sortal import sdlParser, sort, rule, form

# Clears the sort and rule registers in the back end to prevent any sort/rule name conflicts
sort.register.clear()
rule.register.clear()

"""
==================================
DEMO STRUCTURE: SIERPINSKI PYRAMID
==================================
1. Sortal Setup: Sortal setup of default settings
2. Creating sortal shapes: Creating shapes from Rhino GUIDs and text data
    a. Without labels/label descriptions
    b. Adding labels/label descriptions
3. Creating rules: Creating rules from Rhino GUIDs
4. Rule Application: Applying rules
    a. Entire shape
    b. Subshape
    c. Moving the results
5. SDL Writing: Saving result to SDL file

Additional:
    
1. Maximalize: Maximalizing a shape

NOTE: Please ensure that the sortalgi folder is on the file hierarchy level as this code
"""

"""
============
SORTAL SETUP
============
Sets up default sort types in sort register, as well as two additional external description sort types named 'label1' and 'label2'
Sets up compound sort 'rhino_shapes_c' in sort register

To learn how to hardcode sort types yourself from scratch, please refer to the code of 'sortal_setup' within '__init__.py' file inside first level of sortalgi folder
"""
sgi.sortal_setup(['label1', 'label2'])

"""
======================
CREATING SORTAL SHAPES
======================
Creates a shape using a shape name, a list of Rhino GUIDs that will compose the shape geometry, a reference point, and external description outputs
"""

shape_name = "my_first_sortal_shape"

shape = rs.GetObjects("Select geometry that will compose sortal shape (shape with the header Initial Shape)")
refPt = rs.GetObject("Select origin point/reference point for the shape")
print ('Showing shape without descriptions')
# a. Create a shape without descriptions
# Usually, when creating a shape, there must be an input called 'classification' where the user must stipulate the disjunctive sort type that will
# be used to map the Rhino GUIDs to
# However, if there is only one disjunctive sort with geometric sort types active in the compound sort 'rhino_shapes_c', 
# then the value of 'classification' may be left blank
sgi.create_shape(shape_name, shape, refPoint = refPt)

# This prints the shape data by retrieving the shape from the form register using its name, and then showing its geometry and attributes (if any)
# Keys for predicate and directive tagging are not shown
form.formRegister[shape_name].printE()
print ()

# NOTE: If the shape printed in the Rhino command line does not show any geometry, then please exit the code and use the Reset and Debug option (the little 
# black arrow next to the Start Debugging green arrow to run the code.

# b. Create a shape with descriptions
# This input illustrates two description inputs: 'label1' (followed by an ampersand “@”; it shows 2 individuals separated by a semicolon “;”), 
# and 'label2' (2 individuals); declaration of different description types and their individuals is separated by a vertical dash “|”
# When separating different description types and their individuals, use a vertical dash '|'
# When separating description individuals of the same sort type, use a semicolon ';'
print ('Showing shape with descriptions included. The following description string is used as input to sgi.create_shape to include shapes in the final sortal form:')
descriptions = 'label1@("A2", 1, 1234);("A3", 4, 1234)|label2@("A1", 1, 1234)|label2@("A9", 8, 1234)'
print (descriptions)
time.sleep(5)

# Create shape with descriptions
sgi.create_shape(shape_name, shape, descriptions, refPoint = refPt)

# This prints the shape data, showing its geometry and attributes (if any)
# Note that this shape will include descriptions in its individuals
print ('\nShape with descriptions\n')
form.formRegister[shape_name].printE()
print ()

"""
==============
CREATING RULES
==============
Creates a rule from two shapes
"""

# 1. Create the LHS of the rule
shape_name_lhs = "shapeLHS"
shape_lhs = rs.GetObjects("Select geometry that will compose the LHS (shape with the header LHS)")
refPtL = rs.GetObject("Select origin point/reference point for the LHS shape")
shape_lhs_description = 'label1@("A".l?=2, count?>0, 1234)|label2@("A".i?>0, 1, 1234);("A".b?>0, 8, 1234)'
sgi.create_shape(shape_name_lhs, shape_lhs, shape_lhs_description, refPoint = refPtL)

# 2. Create the RHS of the rule
shape_name_rhs = "shapeRHS"
shape_rhs = rs.GetObjects("Select geometry that will compose the RHS (shape with the header RHS)")
refPtR = rs.GetObject("Select origin point/reference point for the RHS shape")
shape_rhs_description = 'label1@("A3", count + 1, 1234);("A3", 4, 1234)|label2@("A".(i + 1), 1, 1234)|label2@("A9", 8, 1234)'
sgi.create_shape(shape_name_rhs, shape_rhs, shape_rhs_description, refPoint = refPtR)

# 3. Create the rule itself
rule_name = 'rule_pyramid'
rule_description = 'test rule'
sgi.create_rule(rule_name, rule_description, shape_name_lhs, shape_name_rhs)

print ('\nRules present in rule register:')
print (rule.register)
print ()
"""
================
RULE APPLICATION
================
Generates rule applications for a given rule based on:
    a. the entire shape
    b. a part of the shape (subshape)
    
Shows how this can be done using legacy methods, and how rule applications may be moved visually in the Rhino workspace
"""

# a. The Entire Shape

# Select reference point referring to results of entire shape (top pink point)
refPoint = rs.GetObject ('Select point that will serve as reference point for the rule application results using the entire shape')

# Note that by default, the function 'find_rule_appns' hides the geometry it produces from the Rhino viewport
# This can be changed to showing the geometry by setting the variable 'hide' to True
shape_results = sgi.find_rule_appns(rule_name, shape_name, refPt = refPoint, hide = False, prnt = True)

# Moves the results of rule application on the entire shape along the x axis
# This function call shows you the default spacing and behavior of using sgi.move on a list of rule application Rhino GUIDs
# The function 'move' returns a list of the GUIDs of the moved Rhino objects; the second input is the axis of alignment, which be default is (0,1,0) - the y-axis
shape_results = sgi.move(shape_results)

# b. Part of the shape (a subshape)

# Create subshape first
# Select yellow pyramid inside main shape
subshape_name = shape_name + '_subshape'
subshape = rs.GetObjects("Select geometry that will compose subshape (the yellow pyramid inside the shape named 'Initial Shape')")
refPt = rs.GetObject("\nSelect origin point/reference point for the subshape (this should be the same as the initial shape's reference point)")

# Create a shape with descriptions (the descriptions are necessary because the LHS of the rule we just created matches descriptions as well)
subshape_descriptions = 'label1@("A2", 1, 1234)|label2@("A1", 1, 1234)|label2@("A9", 8, 1234)'
sgi.create_shape(subshape_name, subshape, subshape_descriptions, refPoint = refPt)
refPoint = rs.GetObject ('Select point that will serve as reference point for the rule application results using the subshape')

# Finds rule applications using the subshape to match onto the main shape
# Of course, the subshape must first be checked to see if it is a part of the initial shape
subshape_results = sgi.find_rule_appns(rule_name, shape_name, subshape_name, refPoint, prnt = True)

# Moves the results of rule application on a subshape along the x-axis (1,0,0)
# The function 'move' returns a list of the GUIDs of the moved Rhino objects
subshape_results = sgi.move(subshape_results, (1,0,0))


"""
===========
SDL WRITING
===========
Saves the rule and initial shape to a SDL file
This will create an SDL file called 'my_first_sdl_file' in the same file hierarchy level as this demo file
"""
sdl_file_name = 'my_first_sdl_file.sdl'
sgi.save_sdl(sdl_file_name, [rule_name], shape_name)

"""
===============
SDL OVERWRITING
===============
Adds one or some of the rule application results to the same SDL file
Append must be set to True for the function to consider this an additive procedure (adding new rules, shapes to the SDL file) 
rather than a complete file overwrite (i.e. all previous data in the SDL file is lost)
In this case, the old contents of the SDL file are maintained, and the form corresponding to the subshape is added to the SDL file
"""
sgi.overwrite_sdl(sdl_file_name, shapes = [subshape_name], append = True) 

"""
===================
MAXIMALIZING SHAPES
===================
Maximalizes a shape, deletes the original shape and returns the GUIDs of the maximalized version of the shape
"""
shape = rs.GetObjects('\nSelect the pyramid with the header "Maximalize Sample"')
# Because the input is a list of Rhino GUIDs, the input will also be a list of Rhino GUIDs
maximal_shape = sgi.maximalize(shape)
sgi.move(maximal_shape)

"""
NOTE: 
    Alternatively, if the input is the name of a shape, the user may select whether to draw the maximalized shape in the Rhino viewport or not
    This is by setting the input 'rhino' in the function to True (False by default)
    The drawing may be hidden by setting the input 'hide' to True (False by default)
    i.e. sgi.maximalize(shape_name, rhino = True, hide = True)
"""