#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
@version: 1.181101
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# Sortal Library Imports
from sortal import (primitiveSort, sort, coordinate, transformSet, rule,
                    disjunctiveSort, attributeSort, affineTransform,
                    affineTransform3D, description, graphMap3D, sdlParser, 
                    compoundSort, disjunctiveSort, graphMap, form, key, flow)
            
# Import for transformations
from mpmath import pi, sin, cos
from copy import deepcopy

# Agnostic API support code imports
from .sortal_api.convert.convert_sortal_to_agnostic import convertS
from .sortal_api.convert.convert_agnostic_to_sortal import convertA
from .sortal_api.convert.convert_rhino_to_agnostic import convertR

# Sortal API function imports
from .sortal_api.apply_all_together import apply_all_together
from .sortal_api.apply_flow import apply_flow
from .sortal_api.convert_sort import convert_sort
from .sortal_api.convert_shape import convert_shape
from .sortal_api.create_flow import create_flow
from .sortal_api.create_shape import create_shape
from .sortal_api.create_rule import create_rule
from .sortal_api.draw_rule import draw_rule
from .sortal_api.draw_shape import draw_shape
from .sortal_api.extract_shape import extract_shape
from .sortal_api.find_rule_appns import find_rule_appns
from .sortal_api.create_flow import create_flow
from .sortal_api.get import get_rule_description, get_rule_lhs, get_rule_rhs
from .sortal_api.maximalize import maximalize
from .sortal_api.move import move
from .sortal_api.overwrite_sdl import overwrite_sdl
from .sortal_api.part_of import part_of
from .sortal_api.precision import check_precision, default_precision, set_precision
from .sortal_api.read_sdl import read_sdl
from .sortal_api.redraw import redraw
from .sortal_api.save_sdl import save_sdl
from .sortal_api.set import set_rule_name, set_rule_description, set_shape_name, set_flow_name, set_flow_description

# sort type imports from sort_types
from .sortal_api.sort_types import (pointSort2D, pointSort3D, pointSortP2D, pointSortP3D, 
                                          lineSegSort2D, lineSegSort3D, lineSegSortP2D, lineSegSortP3D,
                                          planeSegSort2D, planeSegSort3D, planeSegSortP2D, planeSegSortP3D, 
                                          c2D, c3D, cP2D, cP3D,
                                          e2D, e3D, eP2D, eP3D,
                                          cArc2D, cArc3D, eArc2D, eArc3D, 
                                          bez2D, bezP2D, bez3D, bezP3D,
                                          ptDSort, lsDSort, plDSort, labelSort, cDSort, cArcDSort, eDSort, eArcDSort, bezDSort,
                                          weightSort, weight255Sort, colorSort, descSort, enumSort)

# Rhino methods import (for use alongside SortalGI)
from .sortal_api.rhino_specific_methods import rhino_methods

# Rhino imports
import Rhino.Geometry as rg

convertA = convertA()
convertS = convertS()
convertR = convertR()

sc.sticky['fixed_factor'] = 2.0
sc.sticky['keyCount'] = 0
sc.sticky['groupCount'] = 0
sc.sticky['voidCount'] = 0

"""
==============
ABOUT SORTALGI
==============
The functions imported into this __init__ file under the comment 'API function imports' comprise the SortalGI API. While all publicly available functions
are listed and briefly described in this documentation, functions whose code is defined in this file have legacy behavior 
and will eventually be phased out from the API.

Legacy functions:
    1. sortal_setup: sets up default primitive, disjunctive sorts and the overarching compound sort ('rhino_shapes_c') in the sort register through hard code
    2. create_rule_ag: creates a sortal rule for as given rule name, rule description and LHS and RHS agnostic 
    3. find_rule_appns_ag: generates the rule applications for a given rule and shape (input as either Rhino GUIDs or an agnostic dictionary)
    4. get_rule: retrieves the corresponding rule object based on the input rule name from the sortal library
    5. get_rule_name: returns the name of a rule instance as a text string
    
API functions:
    1	apply_flow:	Applies a flow present in the flow register onto a shape present in the form register; returns the final result of the flow application 
        as a list of Rhino GUIDs, which are by default hidden from the Rhino viewport; an optional Boolean value called hide may be set to False 
        to render the resulting list of Rhino GUIDs visible in the viewport
    2	check_precision:	Returns current precision (as number of decimal places) of sortal library
    3	convert_shape:	Converts a sortal shape to the target sort type (this target sort type may be a product of convert_sort or retrieved 
        directly from the sort register)
    4	convert_sort:	Determines the structure of a shape's top sort type and creates the target equivalent sort type based on user input; 
        all external description types and attributes attached to geometries are reflected in the newly constructed sort
        (i.e. 2D -> 3D, P2D -> 2D)
    5	create_flow:	Creates a flow based on a text string elaborating the order and use of existing rules in the sortal rule register
    6	create_rule:	Creates a rule object from the inputs rule name, rule description, LHS shape name and RHS shape name
    7	create_shape:	Creates a shape object from the following inputs: shape name, Rhino geometry, target sort type; 
        the target sort type may be left blank if there is only one functioning geometric disjunctive sort active
    8	default_precision:	Resets precision (as number of decimal places) of sortal library to 5
    9	draw_rule:	Draws the sides of a rule as Rhino geometry, side by side one another; the drawing may be moved to a different location 
        by inputting a reference point (this may be a tuple/list of three numbers or a Rhino geometry point or GUID), 
        where the reference point serves as the new 'origin' for the drawing
    10	draw_shape:	Draws the shape retrieved from the form register as Rhino geometry; the drawing may be moved to a different location 
        by inputting a reference point (this may be a tuple/list of three numbers or a Rhino geometry point or GUID), 
        where the reference point serves as the new 'origin' for the drawing
    11	extract_shape:	Extracts a subshape from a sortal shape based on the target sort type provided by the user
    12	find_rule_appns:	Generates the rule applications from a given rule-shape combination; takes as input rule object or rule name, 
        subshape name and mainshape name and an optional Boolean value as to whether to hide the resulting Rhino GUIDs corresponding to the results; 
        returns list of lists of GUIDs corresponding to resulting shapes after rule applications, with the GUIDs hidden from the Rhino viewport by default
    13	get_lhs:	Returns LHS of rule instance as a list of Rhino GUIDs
    14	get_rhs:	Returns RHS of rule instance as a list of Rhino GUIDs
    15	get_rule_description:	Returns description of rule instance as a text string
    16	maximalize:	Maximalizes a sortal shape (called by its name from the form register)/list of Rhino geometries 
        (if the latter, redraws the Rhino geometries and returns them)
    17	move:	Moves lists of GUIDs in a list apart from another based on the dimensions of each list's collective bounding box, 
        and the axis of movement provided by the user; the default is to move the shapes to the right (based on the x-axis) and upwards
    18	overwrite_sdl: 	Overwrites a pre-existing SDL file with the inputted list of rules, list of shapes and list of flows. 
        If append is set to True, then the overwrite targets rules, shapes and flows within the file that have the same names as the inputs, overwrites them,
        while keeping prior rules, shapes and flows with different names in the SDL file; any new rules, shapes and flows are also added to the SDL file.
        If append is set to False (default), then the entire file is overwritten.
    19	part_of:	Checks if a subshape agnostic object/set of Rhino geometries is part of a potentially larger shape agnostic object/set of Rhino geometries
    20	read_sdl: 	Sets up rules and shapes from SDL file in sortal library; adds shapes to form register; converts SDL file rules and shapes 
        to prevailing 'rhino_shapes_c' sort type; rewrites contents of read SDL file to reflect prevailing 'rhino_shapes_c' structure, if necessary
    21	redraw:	Deletes inputted Rhino geometry and replaces them with their geometric counterparts and corresponding labels, descriptions, 
        predicates and directives based on data stored inside the user text of the original geometry
    22	save_sdl:	Creates a new SDL file or overwrites a pre-existing one using the inputted file name and the listed rules, shapes and flows
    23	set_flow_desc:	Changes the description of the given flow object
    24	set_flow_name:	Changes the flow name of the given flow instance; if the new flow name matches that of a pre-existing flow, 
        prompts user for overwrite or renaming
    25	set_precision:	Sets precision (as number of decimal places) of sortal library to user input
    26	set_rule_desc:	Changes the description of the given rule object
    27	set_rule_name:	Changes rule name of rule instance; if new rule name matches that of a pre-existing rule, 
        prompts user for overwrite or change of rule name input
    28	set_shape_name:	Changes name of shape object and updates its name in the form register
"""

"""
===================
ABOUT RHINO METHODS
===================
These refer to the functions imported from the 'rhino_specific' folder, through this line of code:
    
from .sortal_api.rhino_specific_methods import rhino_methods

They are used to impart attribute information such as labels, descriptions, tags, predicates and directives onto Rhino geometry in the viewport.
This data is extracted when the Rhino geometry is given as input to SortalGI functions like create_shape and redraw, 
and included in the resulting sortal structures.

    1	add_bound_line:	Adds bound line predicate tag to line segment/polyline; indicated as 'bound' within predicates text ('#( )')
    2	add_color:	Adds color to Rhino Object (one or several objects may be selected) using RGB values; this color is not visible in Print Preview Mode
    3	add_description:	Adds description to Rhino Object's user text, visualized as text dot with extra 'd( )' enclosing text in visual text dot assigned to geometry
    4	add_distance:	Adds distance directive tag to geometries; first selected geometry must be found in LHS, second selected geometry must be found in RHS; 
        range of values for length is inputted to function call; directive is indicated as 'distance' within directives text ('#( )') in visual text dot assigned to geometry
    5	add_embed:	Adds embed predicate tag to geometries; first selected geometry must be the larger geometry (either a plane segment or a straight line segment) 
        and the second selected geometry  must be the smaller geometry that will be contained inside the first selected geometry; indicated as 'embeds' within predicates text ('#( )') in visual text dot assigned to geometry
    6	add_enum:	Adds enumerative value to Rhino Object's user text, not visualized
    7	add_label:	Adds label data to Rhino Object's user text, visualized as text dot with extra 'l( )' enclosing text in visual text dot assigned to geometry
    8	add_longest_line:	Adds longest line predicate tag to line segment/polyline; indicated as 'longest' within predicates text ('#( )')  in visual text dot 
        assigned to geometry
    9	add_max_line:	Adds maximum line predicate tag to line segment/polyline; indicated as 'max' within predicates text ('#( )') in visual text dot assigned to geometry
    10	add_no_label:	Adds no-label predicate tag to geometry to constrain matching to not-labelled geometries; indicated as 'no_label' 
        within predicates text ('#( )')  in visual text dot assigned to geometry
    11	add_normal:	Adds normal directive tag to line segments (if 2D), plane segments (if 3D); indicated as 'distance' within directives text ('#( )')  
        in visual text dot assigned to geometry
    12	add_point_on_line:	Adds point-on-line directive tag to line segment/polyline; range of values for point placement is inputted to function call; 
        indicated as 'point_on_line' within directives text ('#( )') followed by coordinates of point in visual text dot assigned to geometry
    13	add_shortest_line:	Adds shortest line predicate tag to line segment/polyline; indicated as 'short' within predicates text ('#( )') in visual text dot 
        assigned to geometry in visual text dot assigned to geometry
    14	add_void:	Adds void predicate tag to collection of points/text dots/line segments/plane segments that form a closed polygon; indicated as 'void' 
        within predicates text ('#( )'), and indicates which void group it belongs to in visual text dot assigned to geometry in visual text dot assigned to geometry
    15	add_weight:	Adds width (weight) to line or grayscale (weight255) to point or text dot in visual text dot assigned to geometry in visual text dot 
        assigned to geometry
    16	analysis:	Analyses a list of Rhino geometry, and moves points/extends lines to intersect directly with each other (for lines to intersect with each other, 
        for points to line on lines) in visual text dot assigned to geometry in visual text dot assigned to geometry
    17	attach_label:	Attaches text dot to a particular line segment; this stores the text dot's GUID in the line segment's user text and moves the text dot 
        in visual text dot assigned to geometry
    18	clear_shape:	Clears a list of Rhino objects; takes as input a list/dictionary of Rhino GUIDs OR single Rhino GUID
    19	delete_description:	Removes description data from Rhino Object, if any are present, and in visual text dot assigned to geometry
    20	delete_label:	Removes label from Rhino Object's user text and in visual text dot assigned to geometry; deletes text dot 'attached' to line segment, 
        or reverts text dot to a point
    21	delete_label:	Removes label data from Rhino Object, if any are present, and in visual text dot assigned to geometry
    22	delete_pred_dir:	Removes predicates/directives from Rhino Object, if any are present, and in visual text dot assigned to geometry
    23	delete_tag:	Removes geometry predicate/directive tag from Rhino Object, if any are present, and in visual text dot assigned to geometry
    24	delete:	Uses Rhino's DeleteObjects to delete all Rhino objects in window, clears sort and rule registers, sets up new grammar frames
    25	tag: 	Adds or changes the tag data stored in a Rhino geometry and changes the visual text dot 'attached' to the geometry accordingly
"""

"""
=============
FUNCTION INFO
=============
'sortal_setup' sets up the sort types in the sortal library, as well as any external description sort types, 
according to the list of string inputs in the variable 'descriptions'. The code in this function also serves as an example 
on how to create sorts through hardcode, from primitive to compound sort types and to imbue geometric primitive sorts with attributes.
 
It may be run without any inputs; this sets up a default compound sort with two disjunctive sorts inside, one of which is a dummy disjunctive sort composed
of two dummy description sorts. This dummy disjunctive sort is removed from the compound sort 'rhino_shapes_c' once other sort types are introduced, when
using 'read_sdl' to import SDL file data.

All geometries (points, line segments, plane segments, circles, ellipses, circular arcs, Bezier curves) are enabled in this default disjunctive sort 
as non-parametric geometries, if no input is given. Due to the limitations of the library, parametric 2D and 3D circular and elliptical arcs are not available.

Labels, weights, descriptions (one type called 'desc'), weight255 (for grayscale), enumeratives and 
colors are enabled as attributes to all the non-parametric geometries. 

The names of the default sort types in any resulting agnostic dictionary
that follows from using sgi.sortal_setup are as follows:
    
    Default: N3D, with extra descriptions
    Compound sort: rhino_shapes_c
    Disjunctive sort: rhino_shapes
    Attribute sorts under disjunctive sort:
        1. Points - 'points'
        2. Line segments - 'line_segments'
        3. Plane Segments - 'plane_segments'
        4. Circles - 'circles'
        5. Ellipse - 'ellipses'
        6. Circular arcs - 'arcs'
        7. Bezier Curves - 'bezier_curves'
    Attributes enabled for all sorts: label, description, color, weight, weight255, enumerative
        
INPUTS
descriptions: List; list of description sort type names
parametric: Boolean; indicates if sort types are to be non-parametric or parametric; default is False (non-parametric)
dimension: Boolean; indicates if sort types are to be 2D or 3D; default is False (3D)

OUTPUTS
True: If setup was successful; 'rhino_shapes_c' and component disjunctive sort are declared in the sort register if this function is successful.

WARNINGS / ERRORS
TypeError: If any value inside the list of description sort type names, descriptions, is not a text string
"""

def sortal_setup(descriptions = ['descLabel'], parametric = False, dimension = False):
    # Clears all registers before proceeding with setup
    sort.register.clear()
    form.formRegister ={}
    rule.register.clear()
    flow.register.clear()
        
    # Common attributes' disjunctive sort name
    sortName = 'label_enumerative_weight_weight255_color_'
    
    # Creation of disjunctive sort types for each type of geometry-centric description
    # Points
    ptAtt = disjunctiveSort(sortName + 'ptDescription', [labelSort, enumSort, weightSort, weight255Sort, colorSort, ptDSort]) 
    # Line segments
    lsAtt = disjunctiveSort(sortName + 'lsDescription', [labelSort, enumSort, weightSort, weight255Sort, colorSort, lsDSort]) 
    # Plane segments
    psAtt = disjunctiveSort(sortName + 'psDescription', [labelSort, enumSort, weightSort, weight255Sort, colorSort, plDSort]) 
    # Circles
    circleAtt = disjunctiveSort(sortName + 'cDescription', [labelSort, enumSort, weightSort, weight255Sort, colorSort, cDSort]) 
    # Ellipses
    ellipseAtt = disjunctiveSort(sortName + 'eDescription', [labelSort, enumSort, weightSort, weight255Sort, colorSort, eDSort]) 
    # Circular arcs
    cArcAtt = disjunctiveSort(sortName + 'cArcDescription', [labelSort, enumSort, weightSort, weight255Sort, colorSort, cArcDSort])
    # Elliptical Arcs
    eArcAtt = disjunctiveSort(sortName + 'eArcDescription', [labelSort, enumSort, weightSort, weight255Sort, colorSort, eArcDSort])  
    # Bezier curves
    bezAtt = disjunctiveSort(sortName + 'bezDescription', [labelSort, enumSort, weightSort, weight255Sort, colorSort, bezDSort]) 
    
    # 3D SORTS
    if parametric:
        if dimension:
            # Parametric 2D sorts
            pointAtt = attributeSort('points', pointSortP2D, ptAtt)
            lineSegAtt = attributeSort('line_segments', lineSegSortP2D, lsAtt)
            planeSegAtt = attributeSort('plane_segments', planeSegSortP2D, psAtt)
            circleAtt = attributeSort('circles', cP2D, circleAtt)
            ellipseAtt = attributeSort('ellipses', eP2D, ellipseAtt)
            bezierAtt = attributeSort('bezier_curves', bezP2D, bezAtt)
            components = [pointAtt, lineSegAtt, planeSegAtt, circleAtt, ellipseAtt, bezierAtt]
            sortName = 'P2D'
        else:
            # Parametric 3D sorts
            pointAtt = attributeSort('points', pointSortP3D, ptAtt)
            lineSegAtt = attributeSort('line_segments', lineSegSortP3D, lsAtt)
            planeSegAtt = attributeSort('plane_segments', planeSegSortP3D, psAtt)
            circleAtt = attributeSort('circles', cP3D, circleAtt)
            ellipseAtt = attributeSort('ellipses', eP3D, ellipseAtt)
            bezierAtt = attributeSort('bezier_curves', bezP3D, bezAtt)
            components = [pointAtt, lineSegAtt, planeSegAtt, circleAtt, ellipseAtt, bezierAtt]
            sortName = 'P3D'
    else:
        if dimension:
            # Non-parametric 2D sorts
            pointAtt = attributeSort('points', pointSort2D, ptAtt)
            lineSegAtt = attributeSort('line_segments', lineSegSort2D, lsAtt)
            planeSegAtt = attributeSort('plane_segments', planeSegSort2D, psAtt)
            circleAtt = attributeSort('circles', c2D, circleAtt)
            ellipseAtt = attributeSort('ellipses', e2D, ellipseAtt)
            bezierAtt = attributeSort('bezier_curves', bez2D, bezAtt)
            cArcAtt = attributeSort('arcs', cArc2D, cArcAtt)
            eArcAtt = attributeSort('ell_arcs', eArc2D, eArcAtt)
            components = [pointAtt, lineSegAtt, planeSegAtt, circleAtt, ellipseAtt, bezierAtt, cArcAtt, eArcAtt]
            sortName = 'N2D'
        else:
            # Non-parametric 3D sorts
            pointAtt = attributeSort('points', pointSort3D, ptAtt)
            lineSegAtt = attributeSort('line_segments', lineSegSort3D, lsAtt)
            planeSegAtt = attributeSort('plane_segments', planeSegSort3D, psAtt)
            circleAtt = attributeSort('circles', c3D, circleAtt)
            ellipseAtt = attributeSort('ellipses', e3D, ellipseAtt)
            bezierAtt = attributeSort('bezier_curves', bez3D, bezAtt)
            cArcAtt = attributeSort('arcs', cArc3D, cArcAtt)
            eArcAtt = attributeSort('ell_arcs', eArc3D, eArcAtt)
            components = [pointAtt, lineSegAtt, planeSegAtt, circleAtt, ellipseAtt, bezierAtt, cArcAtt, eArcAtt]
            sortName = 'N3D'

    if type(descriptions) == type([]):
        for desc_name in descriptions:
            if type(desc_name) == type('a'):
                components.append(primitiveSort(desc_name, description))
            else:
                raise TypeError('Input for description type name ', desc_name, ' is not a text string')
    
    # Creation of final compound sort
    completeSort = disjunctiveSort(sortName, components)
    __dummyDesc1__ = primitiveSort('__dummyDesc1__', description)
    __dummyDesc2 = primitiveSort('__dummyDesc2__', description)
    __dummyDisjSort__ = disjunctiveSort('__dummyDescDisjunctive__', [__dummyDesc1__, __dummyDesc2])
    compoundSort('rhino_shapes_c', [completeSort, __dummyDisjSort__])
    
    # Creating the full disjunctive sort housing all geometric sorts with their attributes
    return True            

"""
=============
FUNCTION INFO
=============
'create_rule_ag' creates a new rule instance; it takes as input the rule name, rule description, LHS and RHS agnostic shapes, 
predicates (for LHS) and directives (RHS). Since create_rule() is linked to the sortal library, the act of creating a rule object checks 
if a rule with the same name as the new rule being created already exists. If yes, then the rule is not created and the function is exited. 

SUPPORT FUNCTIONS
add_predicates: Adds predicates to rule object
add_directives: Adds directives to rule object

INPUT
ruleName: Name of rule to be created (text string) 
ruleDesc: Description of rule to be created (text string)
lhs: An agnostic dictionary that will compose the LHS shape
rhs: An agnostic dictionary that will compose the RHS shape
predLHS: Agnostic dictionary of predicates (optional); usually, this is obtained from sgi.create_shape
dirLHS: Agnostic dictionary of directives (optional); usually, this is obtained from sgi.create_shape
prnt: Boolean value, corresponding to whether to print if rule was created successfuly, and as to which predicates/directives were added to the rule successfully 

OUTPUT
Sortal rule data structure, if successful
None, if unsuccessful

WARNINGS / ERRORS
MessageBox: If a rule with the same name already exists in the sortal register, then the user is prompted for whether they would like to overwrite 
            the pre-existing rule or to give the rule to be created a different name or to exit the create_rule function; 
            alternatively, if the geometry within the rule object's LHS is insufficient
TypeError: If the input for rule name or rule description or is not a text string or if LHS and RHS are not agnostic shapes
KeyError: If 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up

Data Flow: rule name, rule description (optional), LHS shape name, RHS shape name, predicates dictionary, agnostic dictionary
           -> checks if rule with the same name already exists (overwrite/rewrite)
           -> creates rule
           -> adds predicates/directives into rule object
           -> checks validity of rule geometry
           -> returns rule object as sortal rule data structure, if successful
"""

def create_rule_ag(ruleName, ruleDesc, lhs, rhs, predLHS = None, dirRHS = None, prnt = False):
    if 'rhino_shapes_c' in sort.register:
        if ruleName in rule.register:
            go_ahead = rs.MessageBox("A rule with the same name as " + ruleName + " already exists in the rule register. Overwrite pre-existing rule?",  3, "Rule Overwrite")
        
            if go_ahead == 6:
                del rule.register[ruleName]
                
            elif go_ahead == 7:
                newName = raw_input('Please input new rule name for the rule to be created')
                ruleName = newName
                
            elif go_ahead == 2:
                warnings.warn('Exiting create_rule function')
                return None
        
        if type(ruleName) != type('a'):
            raise TypeError('Rule name input is not a string')
            
    	if type(lhs) != type({}) or type(rhs) != type({}):    
    		raise TypeError('Content/s of LHS or RHS are not agnostic shapes')
    	
		# Converts LHS and RHS agnostic tuples to sortal metaforms
    	if len(lhs) > 0:
        	lhsMetaform = convertA.toSortal(lhs)
    	else:
        	raise ValueError('LHS is empty')
            
    	rhsMetaform = convertA.toSortal(rhs)

        # Creates rule instance using sortal library
        ruleNew = rule(ruleName, ruleDesc, lhsMetaform, rhsMetaform, keyCheck = True)  
        
        if predLHS != None and predLHS != False:
            if type(predLHS) == type([]):
                for pred in predLHS:
                    add_predicates(ruleNew, pred, prnt)
            else:
                add_predicates(ruleNew, predLHS, prnt)
        
        if dirRHS != None and dirRHS != False:
            if type(dirRHS) == type([]):
                for dir in dirRHS:
                    add_directives(ruleNew, dirRHS, prnt)
            else:
                add_directives(ruleNew, dirRHS, prnt)
        
        if ruleNew.isValid():
            if prnt:
                print (ruleName + ' has been created.')
            return ruleNew, ruleName
        else:
            message = ('Rule was not created since there was insufficient information for LHS Shape.\nThe minimum requirements for defining the LHS are:\n'+
                                    '- 3 points (not coincident)\n'+
                                    '- 1 point + 1 line (not co-linear)\n'+
                                    '- 3 lines (no two parallel: all lines intersect one another)\n'+
                                    '- 3 lines (2 lines parallel with each other, the other line intersecting both)\n'+
                                    '- 2 lines (skew): However, this may have unexpected results; try not to have rules with only skew lines if possible')
                                    
            warnings.warn(message)
            del rule.register[ruleName]
            return None
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')

"""
=============
FUNCTION INFO
=============
'find_rule_appns_ag' finds rule applications from a subshape and shape, or from a set of GUIDs 

INPUT
chosenRule: Rule name as text string
shape: Name of main shape as text string as recorded in sortal library's form register OR agnostic dictionary pertaining to shape 
       OR Rhino GUIDs composing shape (predicates and directives will not be taken into account)
subshape: Name of subshape as text string as recorded in sortal library's form register (optional) OR agnostic dictionary pertaining to subshape 
          OR Rhino GUIDs composing subshape (predicates and directives will not be taken into account)
refPt: GUID of point, tuple, or Rhino Geometry point of reference point which will serve as the 'origin' with which shapes 
       will be plotted in respect to (in principle, a vector; optional)

OUTPUT
List of applications (in tuple form) as based on match generated by subshape or main shape, if no subshape input is given
None, if unsuccessful

Each element of the list is a tuple in the following format:
    a. Non-parametric:
    (<transformation matrix>, <LHS match in agnostic form based on subshape>, <RHS result in agnostic form>, <Main Shape after rule application in agnostic form>)
    
    b. Parametric:
    (<LHS match in agnostic form based on subshape>, <RHS result in agnostic form>, <Main Shape after rule application in agnostic form>)

WARNINGS / ERRORS
ValueError: If rule name is not present in the rule register or if the input for chosenRule is neither a rule object nor a rule name
TypeError: If initial shape or subshape is not a list of GUIDs or a text string or an agnostic dictionary or is empty
KeyError: If initial shape or subshape input's name is not present in the form register
          OR if 'rhino_shapes_c' is not present in the sort register
Warning: If subshape is not part of shape -> exits function and returns False

Data Flow:  GUIDs 
            -> agnostic object
            -> sortal 
            -> rule applications based on subshape, subshapeAgnostic
"""

def find_rule_appns_ag(chosenRule, shape, subshape = None, refPointMain = r.Geometry.Point3d(0,0,0)):
    if 'rhino_shapes_c' in sort.register:
        # RETRIEVAL OF RULE OBJECT
        # Checks if rule input is a sortal rule object or a valid string input that corresponds to a rule in the rule register
        if type(chosenRule) == type('a'):
            if chosenRule in rule.register:
                chosenRule = rule.register[chosenRule]
            else:
                raise ValueError('Rule name is not present in the rule register.')
        else:
            raise ValueError('Rule input is neither a rule object or a rule name.')
        
        agsRuleAppns = []       # List of rule applications in agnostic form
        ruleAppns = []          # List of rule applications to be returned (either as agnostic dict or Rhino GUIDs)
        subshapeSortal = None
        
        # RETRIEVAL OF INITIAL SHAPE
        # Checks if inputs are of the type agnostic dictionary
        if type(shape) == type({}) or type(shape) == type([]):
            # Shape - List of Rhino GUIDs
            if type(shape) == type([]):
                try:
                    rs.ObjectType(shape[0])
                except:
                    if len(shape) > 0:
                        raise TypeError('Initial shape input is not a list of Rhino GUIDs')
                    else:
                        raise TypeError('Initial shape input is empty')
                        
                shapeAgs, shapeIDs, pred, dir = convertR.toAgnostic(shape, 'N3D')
                shapeSortal = convertA.toSortal(shapeAgs)
                
            # Shape - Agnostic object dictionary
            elif type(shape) == type({}):
                # Conversion from agnostic to sortal form
                shapeSortal = convertA.toSortal(shape) 
                
        # Text string reference to shape
        # Retrieves shape of the corresponding name from the sortal form register
        elif type(shape) == type('a'):
            if shape in form.formRegister:
                shapeSortal = form.formRegister[shape]
            else:
                raise KeyError(('Shape with the name ' + shape + ' does not exist in the sortal form register.'))
        
        else:
            raise TypeError(('Shape input is not a list of Rhino GUIDs, an agnostic dictionary, or a text string.'))
            
        # RETRIEVAL OF SUBSHAPE
        # Checks if subshape is valid (optional)
        # Retrieves subshape from sortal form register if necessary
        if subshape != None and (type(subshape) == type({}) or type(subshape) == type([]) or type(subshape) == type('a')):
            if type(subshape) == type('a'):
                if subshape in form.formRegister:
                    subshapeSortal = form.formRegister[subshape]
                else:
                    raise KeyError('Shape with the name', subshape, 'does not exist in the sortal form register.')
            
            elif type(subshape) == type([]):
                # Checks if subshape is a list of Rhino GUIDs
                try:
                    rs.ObjectType(subshape[0])
                except:
                    if len(subshape) > 0:
                        raise TypeError('Subshape input is not a list of Rhino GUIDs')
                    else:
                        raise TypeError('Subshape input is empty')
                
                subshapeAgs, subshapeIDs, pred, dir = convertR.toAgnostic(subshape, 'N3D')
                subshapeSortal = convertA.toSortal(subshapeAgs)
           
            elif type(subshape) == type({}):
                subshapeSortal = convertA.toSortal(subshape)
                
        else:
            if subshape != None:
                raise TypeError('Subshape input is not a list of Rhino GUIDs, an agnostic dictionary, or a text string.')
        
        # COMPATIBILITY BETWEEN SUBSHAPE AND SHAPE
        # If subshape is provided
        if subshapeSortal != None:
            
            # Checks if subshape is part of main shape before proceeding
            if subshapeSortal.partOf(shapeSortal):
                subshapeApps = chosenRule.detect(subshapeSortal)       # Generates matches, results and rule applications for subshape
                
                for index1 in range(len(subshapeApps)):
                    match = convertS.toAgnostic(subshapeApps[index1].matchForm)                         # Match in shape to LHS
                    result = convertS.toAgnostic(subshapeApps[index1].resultForm)                       # Match after rule application
                    after = convertS.toAgnostic(subshapeApps[index1].perform(shapeSortal.duplicate()))  # Entire shape after rule application
                    
                    # Matrix is only generated in non-parametric cases
                    try:
                        matrix = subshapeApps[index1].transformSet.retrieveTransform(affineTransform3D).matrix
                        agsRuleAppns.append((matrix, match, result, after,))
                    except:
                        agsRuleAppns.append((match, result, after,))
                        
            else:   
                warnings.warn('Subshape is not part of main shape. Matches cannot be found.')
                return None
                
        # Subshape is not provided
        else:
            shapeApps = chosenRule.detect(shapeSortal)       # Generates matches, results and rule applications for subshape
    
            for index1 in range(len(shapeApps)):
                match = convertS.toAgnostic(shapeApps[index1].matchForm)                        # Match in shape to LHS
                result = convertS.toAgnostic(shapeApps[index1].resultForm)                      # Match after rule application
                after = convertS.toAgnostic(shapeApps[index1].perform(shapeSortal.duplicate())) # Entire shape after rule application
                
                # Matrix is only generated in non-parametric cases
                try:
                    matrix = shapeApps[index1].transformSet.retrieveTransform(affineTransform3D).matrix
                    agsRuleAppns.append((matrix, match, result, after,))
                except:
                    agsRuleAppns.append((match, result, after,))
    
        # Returns agnostic dictionary forms of match, result and shape after rule application
        # Format: (<transformation matrix>, <match agnostic dict>, <result agnostic dict>, <shape after application agnostic dict>)
        ruleAppns = agsRuleAppns
                
        return ruleAppns
        
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')

"""
=============
FUNCTION INFO
=============
'get_rule' retrieves the corresponding rule object based on the input rule name from the sortal library

INPUT
ruleName: name of the rule

OUTPUT
ruleObject: the rule object from the sortal library
    
Data Flow: rule name -> rule object
"""  

def get_rule(ruleName):
    if type(ruleName) == type('a'):
        if ruleName in rule.register:
            return rule.register[ruleName]
        else:
            message = ('Rule does not exist.')
            rs.MessageBox(message)    
    else:
        message = ('Name entered is not a string.')
        rs.MessageBox(message)
        
    return None

"""
=============
FUNCTION INFO
=============
'get_rule_name' returns the name of the rule object

INPUT
ruleObject: rule object

OUTPUT
ruleName: the name of the rule object
    
Data Flow: rule object -> rule name
"""  

def get_rule_name(ruleObject):
    return ruleObject.name
