#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# API support code imports
from ..sortal_api.convert.convert_sortal_to_agnostic import convertS
from ..sortal_api.convert.convert_agnostic_to_sortal import convertA
from ..sortal_api.convert.convert_rhino_to_agnostic import convertR
from ..sortal_api.convert.visualAgnostic import visualA
from vis_result import vis_result, vis_results
from ..sortal_api.rhino_specific_methods import rhino_methods as rm

# Rhino import/s
import Rhino as r
import rhinoscriptsyntax as rs
import scriptcontext as sc

# Conversion code initialization
convertA = convertA()
convertR = convertR()
convertS = convertS()
           
"""
=============
FUNCTION INFO
=============
'draw_result' draws a single result (identified by the input 'choice') in the Rhino workspace. The default is the first item of the ruleAppns list.
When a reference point is given (tuple/list/vector/point), the shape GUIDs are moved to the reference point.
The default reference point is the origin. This function relies on the back-end function vis_result (inside sortal_api) to draw the geometry.

INPUT
choice: Index (integer) of chosen result from list of rule applications (tuple of agnostic dictionaries coming from find_rule_appns_ag) 
ruleAppns: List of rule application tuples of agnostic dictionaries (Non-parametric: (transformation matrix, match, result, shape after application), 
           Parametric: (match, result, shape after application)) to retrieve the shape to be drawn from
refPointMain (optional): Tuple / list of integers or floats / vectors / points (3Ds)), 
hide: Boolean Value; True (default) hides the resulting Rhino geometries; False leaves them visible in the Rhino Viewport
layerNameMain: name of layer to draw Rhino GUIDs of result on, 
shapeIDs: List of shape GUIDs of original shape (if these are to be deleted and replaced)

OUTPUT
List of Rhino Geometry [shape GUIDs], if successful
None, if unsuccessful or inputs are incorrect

WARNINGS / ERRORS
Warning: Choice input is an integer greater than number of available rule applications
ValueError: No rule applications entered OR target layer name is not present in layer names OR shapeIds is not a list OR value of hide is not a Boolean value
TypeError: choice input is not an integer OR ruleAppns is not a list OR data type of reference point is not a tuple/list of three numbers/Rhino point
KeyError: Number of elements in the reference point tuple or list is greater than /  less than 3

Data Flow: ruleAppn tuple 
           -> retrieve agnostic object of resulting shape after rule application 
           -> draw in Rhino workspace 
           -> hide geometries from viewport if hide set to True
           -> return list of Rhino GUIDs
"""

def draw_result(choice = 0, ruleAppns = [], refPointMain = r.Geometry.Point3d(0,0,0), hide = True, layerNameMain = 'Default', shapeIds = []):
    if layerNameMain not in rs.LayerNames() and layerNameMain != 'Default':
        raise ValueError('Layer name ' + layerNameMain + ' is not present in the Rhino workspace')
        
    if type(shapeIds) != type([]):
        raise ValueError('Input for shape GUIDs to be deleted is not a list')
        
    if type(hide) != type(True):
        raise ValueError('Input for hiding GUIDs after drawing the geometry is not a Boolean value')
        
    if type(refPointMain) != type([]) and type(refPointMain) != type(()):
        try:
            point = rs.PointCoordinates(refPointMain)
        except:
            raise TypeError('Input for reference point is not a Rhino point, tuple or list')
    elif type(refPointMain) == type([]) or type(refPointMain) == type(()):
        if len(refPointMain) != 3:
            raise KeyError('Number of elements in reference point tuple or list is greater than/less than 3')
        
    if type(choice) == type(1) and type(ruleAppns) == type([]):
        if len(ruleAppns) > 0:
            if choice < len(ruleAppns):
                geometry = vis_result(choice, ruleAppns, refPointMain, layerNameMain, shapeIds)
                if hide:
                    rs.HideObjects(geometry)
                return geometry
            else:
                warnings.warn('Choice input is greater than number of applications available')
        else:
            raise ValueError('No rule applications entered')
    else:
        if type(choice) != type(1):
            raise TypeError('Choice input is not an integer')
        
        if type(ruleAppns) != type([]):
            raise TypeError('Input for rule applications is not a list')
            
    return None
    
"""
=============
FUNCTION INFO
=============
'draw_results' draws all the resulting agnostic objects of the ruleAppns list in the Rhino workspace. 
Each result's drawing is drawn on top of one another, with the location of the plotting based on the location of the original shape the rule was applied onto.
When a reference point is given (tuple/list/vector/point), the shape GUIDs are moved to the reference point, but are still drawn on top of one another
The default reference point is the origin. It is based on the function vis_results. 
The Rhino GUIDs are, by default, hidden as soon as they are drawn in the Rhino viewport.

INPUT
ruleAppns: List of rule applications as agnostic objects in tuples resulting from use of find_rule_appns_ag
refPointMain: Rhino GUID, tuple or Rhino point geometry where results will be plotted in reference to
hide: Boolean value; True (default) - hide Rhino geometries from viewport, False - show Rhino geometries in viewport
layerNameMain: Text string; layer name of layer where resulting geometries are to be placed
shapeIDs: List of Rhino GUIDs to delete and to be replaced by Rhino GUIDs coming from draw_results
prnt: Boolean value; True - print description individuals belonging to resulting shape, False (default) - description individuals not printed 

OUTPUT
List of Rhino Geometry [GUIDs]: If successful; geometries are hidden from the viewport itself
None, if function is unsuccessful or inputs are incorrect

WARNINGS / ERRORS
ValueError: No rule applications entered OR target layer name is not present in layer names OR shapeIds is not a list OR value of hide is not a Boolean value
TypeError: If ruleAppns is not a list OR data type of reference point is not a tuple/list of three numbers/Rhino point
KeyError: If the list of rule applications is empty

Data Flow: each ruleAppn tuple 
           -> retrieve agnostic object of resulting shape after rule application 
           -> draw in Rhino workspace 
           -> Rhino GUIDs 
           -> hide Rhino GUIDs 
           -> return list of Rhino GUIDs corresponding to results
"""

def draw_results(ruleAppns, refPointMain = r.Geometry.Point3d(0,0,0), hide = True, layerNameMain = 'Default', shapeIds = [], prnt = False):    
    if layerNameMain not in rs.LayerNames() and layerNameMain != 'Default':
        raise ValueError('Layer name ' + layerNameMain + ' is not present in the Rhino workspace')
        
    if type(shapeIds) != type([]):
        raise ValueError('Input for shape GUIDs to be deleted is not a list')
        
    if type(hide) != type(True):
        raise ValueError('Input for hiding GUIDs after drawing the geometry is not a Boolean value')
        
    if type(refPointMain) != type([]) and type(refPointMain) != type(()):
        try:
            point = rs.PointCoordinates(refPointMain)
        except:
            raise TypeError('Input for reference point is not a Rhino point, tuple or list')
    elif type(refPointMain) == type([]) or type(refPointMain) == type(()):
        if len(refPointMain) != 3:
            raise KeyError('Number of elements in reference point tuple or list is greater than/less than 3')

    if type(ruleAppns) == type([]):
        if len(ruleAppns) > 0:
            geometry = vis_results(ruleAppns, refPointMain, layerNameMain, shapeIds, prnt)
            if hide:
                for result in geometry:
                    rs.HideObjects(result)
            return geometry
            
        else:
            raise KeyError('Empty list of rule applications')
    else:
        if type(ruleAppns) != type([]):
            raise TypeError('Input for rule applications is not a list')
            
    return None
