# -*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs

# Python built-ins
from builtins import range

# Sortal Library Imports
from sortal import (transformSet, rule, primitiveSort, point3D, point2D, coordinate, lineSegment3D, planeSegment2D, 
                    lineSegment2D, line3D, label, rWeight, disjunctiveSort, attributeSort, affineTransform,  
                    vector3D, vector2D, planeSegment3D, description, color, label, enumerative, sort, key, sdlParser,
                    compoundSort, disjunctiveSort)
                    
# Import for transformations
from mpmath import pi, sin, cos

# Scriptcontext Dictionary Import
import scriptcontext as sc

# API Imports
from ..sortal_shortcuts.create import create
from ..sort_types import (pointSort2D, pointSort3D, pointSortP2D, pointSortP3D, 
                                          lineSegSort2D, lineSegSort3D, lineSegSortP2D, lineSegSortP3D,
                                          planeSegSort2D, planeSegSort3D, planeSegSortP2D, planeSegSortP3D, 
                                          c2D, c3D, cP2D, cP3D,
                                          e2D, e3D, eP2D, eP3D,
                                          cArc2D, cArc3D, eArc2D, eArc3D, 
                                          bez2D, bezP2D, bez3D, bezP3D,
                                          ptDSort, lsDSort, plDSort, labelSort, cDSort, cArcDSort, eDSort, eArcDSort, bezDSort,
                                          weightSort, weight255Sort, colorSort, descSort, enumSort)

# API Initialization
create = create()

"""
ConvertA CLASS
Sorts objects into points, line segments, labels
Performs conversion of agnostic dictionary to sortal grammar

INPUT
Agnostic dictionary

OUTPUT
Sortal metaform

CURRENTLY SUPPORTS
Individuals: Points/text dots, lines and polylines, plane segments
Attributes: Labels, descriptions, enumerative, weights (0 to 1, line widths), weights (grayscale from 0 to 255), color (RGB)
"""

class convertA(object):
    # Input: agnostic form, disjunctive sort dictionary
    # Converts agnostic form to sortal metaform

    def __init__(self):
        pass
    # Question: how to check for 2d or 3d parameters in switchable dimension, parametricness?
    # pd - refers to predicate/directive text
    def toSortal(self, agnosticObject, classification = ''):
        self.agnosticObject = agnosticObject                        # Contains agnostic object (compound/disjunctive)       
        self.descSorts = {}                                         # Dictionary of sort types for external descriptions
        self.planeSegment = {}                                      # Dictionary of sort types for plane segments
        self.lineSegment = {}                                       # Dictionary of sort types for line segments
        self.point = {}                                             # Dictionary of sort types for points           
        self.circle = {}                                            # Dictionary of sort types for circles  
        self.ellipse = {}                                           # Dictionary of sort types for ellipses  
        self.eArc = {}                                              # Dictionary of sort types for elliptical arcs  
        self.cArc = {}                                              # Dictionary of sort types for circular arcs
        self.bezier = {}                                            # Dictionary of sort types for bezier curves
        self.tagCount = 0                                           # Count for void
        self.individuals = {'descriptions':[]}                      # Container for sortal individuals before creation of forms
        self.forms = []                                             # List of metaforms to create compound sort

        # print(self.agnosticObject)
        
        # Checks if convert class received any objects
        if self.agnosticObject == None:
            ('No input')
            return None 
        
        for sortType in self.agnosticObject:
            if sort.register[sortType] == 'sort-disjunctive':
                self.mainSort = sortType
                self.mainSorts = self.mainSort.component

                # Retrieves which sorts are available for use in individual creation
                self.determineMainSorts(self.mainSorts)
                
                for disjSort in self.agnosticObject[sortType]:
                    self.makeIndividuals(self.agnosticObject[sortType], self.mainSort)

                sortal_struct = self.createCompoundForm([self.makeForm()])
                break
                
            elif sort.register[sortType].type == 'sort-attribute' or sort.register[sortType].type == 'sort-primitive':
                # Retrieve disjunctive sort which contains entire metaform
                # Works in either case, if compound sort is self-declared by user or declared by system 
                
                for subsort in sort.register['rhino_shapes_c'].component:
                    descCount = 0
                    for subsortType in subsort.component:
                        if 'description' in subsortType.baseSort().characteristic.type:
                            descCount += 1
                            
                    if descCount != len(subsort.component):
                        self.mainSort = subsort
                        self.mainSorts = subsort.component
                        
                        break
                        
                # Assigns label description sort types based on prevailing highest tier sort
                self.determineMainSorts(self.mainSorts)

                # Transforms agnostic objects into sortal individuals
                # Creates metaform and inserts forms of individuals inside
                self.makeIndividuals(self.agnosticObject)
                self.forms.append(self.makeForm())

                sortal_struct = self.createCompoundForm(self.forms)
                break
       
        return sortal_struct

    """
    =================================================
    ASSIGNING MAIN GEOMETRIC AND ATTRIBUTE SORT TYPES
    =================================================
    """
    # Identifies which standalone description sort types are available in 'rhino_shapes'
    # Identifies and extracts attribute sort types of main geometric sort for easy retrieval during individual creation
    def determineMainSorts(self, mainSorts):
        for sortType in mainSorts:
            attribute = None

            # Obtains attribute sorts from geometric attribute sort
            if sortType.type == 'sort-attribute':
                attribute = sortType.weight
            
            if sortType.baseSort().characteristic.type == 'ind-description':
                self.descSorts[sortType.name] = sortType
                
            elif 'lineSegment' in sortType.baseSort().characteristic.type:
                self.lineSegment[sortType.name] = {}
                sortDict = self.lineSegment
                    
            elif 'planeSegment' in sortType.baseSort().characteristic.type: 
                self.planeSegment[sortType.name] = {}
                sortDict = self.planeSegment
                    
            elif 'point' in sortType.baseSort().characteristic.type:
                self.point[sortType.name] = {}
                sortDict = self.point
            
            elif 'circle' in sortType.baseSort().characteristic.type and 'Arc' not in sortType.baseSort().characteristic.type:
                self.circle[sortType.name] = {}
                sortDict = self.circle
            
            elif 'ellipse' in sortType.baseSort().characteristic.type and 'Arc' not in sortType.baseSort().characteristic.type:
                self.ellipse[sortType.name] = {}
                sortDict = self.ellipse
            
            elif 'Arc' in sortType.baseSort().characteristic.type or 'arc' in sortType.baseSort().characteristic.type:
                if 'circular' in sortType.baseSort().characteristic.type:
                    self.cArc[sortType.name] = {}
                    sortDict = self.cArc
            
                elif 'elliptical' in sortType.baseSort().characteristic.type:
                    self.eArc[sortType.name] = {}
                    sortDict = self.eArc
            
            elif 'bezier' in sortType.baseSort().characteristic.type:
                self.bezier[sortType.name] = {}
                sortDict = self.bezier
            
            # Adding attribute key name entries to attribute dictionary of main sort dictionary entry
            if attribute != None:
                # Determines if attribute sort is disjunctive or primitive
                # Goes through attribute sort if disjunctive
                if 'disjunctive' in attribute.baseSort().type:
                    for i in attribute.component:
                        self.determineAttributes(i, sortDict, sortType)
                elif 'primitive' in attribute.baseSort().type:
                    self.determineAttributes(attribute, sortDict, sortType)

        return
    
    def determineAttributes(self, attribute, sortDict, sortType):
        # Stores attribute sorts used by specific sort type passed in variable sortType
        if 'description' in attribute.baseSort().characteristic.type:
            sortDict[sortType.name]['label_descriptions'] = attribute
        else:
            if sortDict is self.lineSegment:
                self.lineSegment[sortType.name]['label_descriptions'] = lsDSort
            elif sortDict is self.planeSegment:
                self.planeSegment[sortType.name]['label_descriptions'] = plDSort
            elif sortDict is self.point:
                self.point[sortType.name]['label_descriptions'] = ptDSort
            elif sortDict is self.circle:
                self.circle[sortType.name]['label_descriptions'] = cDSort
            elif sortDict is self.ellipse:
                self.ellipse[sortType.name]['label_descriptions'] = eDSort
            elif sortDict is self.cArc:
                self.cArc[sortType.name]['label_descriptions'] = cArcDSort
            elif sortDict is self.eArc:
                self.eArc[sortType.name]['label_descriptions'] = eArcDSort
            elif sortDict is self.bezier:
                self.bezier[sortType.name]['labelDescriptions'] = bezDSort
                
        if 'label' in attribute.baseSort().characteristic.type:
            sortDict[sortType.name]['label'] = attribute
            
        if 'rWeight' in attribute.baseSort().characteristic.type:
            if 'weight255' in attribute.name:
                sortDict[sortType.name]['weight255'] = attribute
            elif 'weight' in attribute.name:
                sortDict[sortType.name]['weight'] = attribute
        
        if 'color' in attribute.baseSort().characteristic.type:
            sortDict[sortType.name]['color'] = attribute
   
        return
    
    """
    ======================================
    CONVERTING GUIDs to SORTAL INDIVIDUALS 
    ======================================
    """

    # Creates sortal individuals from extracted data in agnostic object 
    def makeIndividuals(self, agnosticObject):
        revisedMainSorts = []

        for i in agnosticObject: 
            if i != 'descriptions':
                # Creates point individuals with corresponding attributes
                if 'point' in sort.register[i].baseSort().characteristic.type:
                    if type(agnosticObject[i]) == type([]):
                        self.makePoint(sort.register[i].name, sort.register[i], agnosticObject[i])
                    
                    elif type(agnosticObject[i]) == type({}):
                        for a in agnosticObject[i]:
                            self.makePoint(sort.register[a].name, sort.register[a], agnosticObject[i][a])
    
                # Creates line segment individuals with corresponding attributes
                elif 'lineSeg' in sort.register[i].baseSort().characteristic.type:
                    if type(agnosticObject[i]) == type([]):
                        self.makeLine(sort.register[i].name, sort.register[i], agnosticObject[i])
                    
                    elif type(agnosticObject[i]) == type({}):
                        for a in agnosticObject[i]:
                            self.makeLine(sort.register[a].name, sort.register[a], agnosticObject[i][a])
    
                # Creates plane segment individuals with corresponding attributes
                elif 'planeSeg' in sort.register[i].baseSort().characteristic.type:
                    if type(agnosticObject[i]) == type([]):
                        self.makePlane(sort.register[i].name, sort.register[i], agnosticObject[i])
                    
                    elif type(agnosticObject[i]) == type({}):
                        for a in agnosticObject[i]:
                            self.makePlane(sort.register[i].name, sort.register[a], agnosticObject[i][a])
                
                elif 'circle' in sort.register[i].baseSort().characteristic.type:
                    if type(agnosticObject[i]) == type([]):
                        self.makeCircle(sort.register[i].name, sort.register[i], agnosticObject[i])
                    
                    elif type(agnosticObject[i]) == type({}):
                        for a in agnosticObject[i]:
                            self.makeCircle(sort.register[i].name, sort.register[a], agnosticObject[i][a])
                
                elif 'ellipse' in sort.register[i].baseSort().characteristic.type:
                    if type(agnosticObject[i]) == type([]):
                        self.makeEllipse(sort.register[i].name, sort.register[i], agnosticObject[i])
                    
                    elif type(agnosticObject[i]) == type({}):
                        for a in agnosticObject[i]:
                            self.makeEllipse(sort.register[i].name, sort.register[a], agnosticObject[i][a])
                
                elif 'Arc' in sort.register[i].baseSort().characteristic.type:
                    if 'circular' in sort.register[i].baseSort().characteristic.type:
                        if type(agnosticObject[i]) == type([]):
                            self.makeCArc(sort.register[i].name, sort.register[i], agnosticObject[i])
                        
                        elif type(agnosticObject[i]) == type({}):
                            for a in agnosticObject[i]:
                                self.makeCArc(sort.register[i].name, sort.register[a], agnosticObject[i][a])
                            
                    elif 'elliptical' in sort.register[i].baseSort().characteristic.type:
                        if type(agnosticObject[i]) == type([]):
                            self.makeEArc(sort.register[i].name, sort.register[i], agnosticObject[i])
                        
                        elif type(agnosticObject[i]) == type({}):
                            for a in agnosticObject[i]:
                                self.makeEArc(sort.register[i].name, sort.register[a], agnosticObject[i][a])
                
                elif 'bezier' in sort.register[i].baseSort().characteristic.type:
                    if type(agnosticObject[i]) == type([]):
                        self.makeBezier(sort.register[i].name, sort.register[i], agnosticObject[i])
                    
                    elif type(agnosticObject[i]) == type({}):
                        for a in agnosticObject[i]:
                            self.makeBezier(sort.register[i].name, sort.register[a], agnosticObject[i][a])
                            
            # Handles descriptions
            else:
                if 'descriptions' in agnosticObject:
                    for obj in agnosticObject['descriptions']:
                        attributes = agnosticObject['descriptions'][obj]
                        self.makeDescription(obj, attributes)
                   
        return
    
    def makePoint(self, sortName, mainSortIndex, ptsList):
        for obj in ptsList:
            coords = obj[0]
            attributes = obj[1]
            
            if len(coords) == 2:
                individual = mainSortIndex.baseSort().characteristic(mainSortIndex, coords[0], coords[1]) # individual w/o attributes
            elif len(coords) == 3:
                individual = mainSortIndex.baseSort().characteristic(mainSortIndex, coords[0], coords[1], coords[2]) # individual w/o attributes
            
            att = []
            # Checks if attributes are present
            if len (attributes) > 0:
                for attrib in attributes:
                    att.append(self.makeAttribute(self.point[sortName], attrib, attributes[attrib], individual))

                if len(att) != 0:
                    for attribute in att:
                        for a in attribute:
                            individual.addAttribute(a)
            
            if sortName not in self.individuals:
                self.individuals[sortName] = []
            
            self.individuals[sortName].append(individual)
        
        return
            
    def makeLine(self, sortName, mainSortIndex, lineSegments):
        polylineCheck = False

        for obj in lineSegments:
            if type(obj[0]) == type (()):
                info = obj[0]
                attributes = obj[1]
                if len(info) == 2:
                    if '2' in mainSortIndex.baseSort().characteristic.type:
                        v1 = vector2D(info[0][0], info[0][1])
                        v2 = vector2D(info[1][0], info[1][1])
                    else:
                        if len(info[0]) == 2:
                            v1 = vector2D(info[0][0], info[0][1], 0)
                            v2 = vector2D(info[1][0], info[1][1], 0)
                        elif len(info[0]) == 3:
                            v1 = vector3D(info[0][0], info[0][1], info[0][2])
                            v2 = vector3D(info[1][0], info[1][1], info[1][2])
                   
                    individual = mainSortIndex.baseSort().characteristic(mainSortIndex, v1, v2)
                    
                else:
                    lineList = obj[0]
                    attributes = obj[1]
                    individual = []
                    polylineCheck = True
                    for ind, pt in enumerate(lineList):
                        if ind == (len(lineList)-1):
                            v1 = vector3D(pt[0], pt[1], pt[2])
                            v2 = vector3D(lineList[0][0], lineList[0][1], lineList[0][2])
                        else:
                            v1 = vector3D(pt[0], pt[1], pt[2])
                            v2 = vector3D(lineList[ind+1][0], lineList[ind+1][1], lineList[ind+1][2])
                            
                        indiv = mainSortIndex.baseSort().characteristic(mainSortIndex, v1, v2)
                        individual.append(indiv)

                att = []

            # Checks if attributes are present
            if len (attributes) > 0:
                for attrib in attributes:
                    att.append(self.makeAttribute(self.lineSegment[sortName], attrib, attributes[attrib], individual))

                # Inserts attributes into sortal individual if any
                if len(att) > 0:
                    if type(individual) != type([]):
                        for attribute in att:
                            for a in attribute:
                                individual.addAttribute(a)
                    elif type(individual) == type([]):
                        for attribute in att:
                            for a in attribute:
                                for indiv in individual:
                                    indiv.addAttribute(a)
            
            if sortName not in self.individuals:
                self.individuals[sortName] = []
            
            if type(individual) == type([]):
                for indiv in individual:
                    self.individuals[sortName].append(indiv)
            else:
                self.individuals[sortName].append(individual)
       
        return
       
    def makePlane(self, sortName, mainSortIndex, planesList):
        for obj in planesList:
            info = obj[0]
            attributes = obj[1]
            outerBounds = []
            innerBounds = []
            pOut = None
            pIn = []
            
            if type(info) == type ([]):
                for curve in info:
                    if '2D' in mainSortIndex.name:
                        v1 = vector2D(curve[0], curve[1])
                       
                    elif '3D' in mainSortIndex.name:
                        if len(curve) == 2:
                            v1 = vector3D(curve[0], curve[1], 0)
                        else:
                            v1 = vector3D(curve[0], curve[1], curve[2])         
                    outerBounds.append(v1)
                
                # Appends first bound coordinate to complete plane segment coordinates
                outerBounds.append(outerBounds[0])
                
                # Constructs plane segment
                pOut = mainSortIndex.baseSort().characteristic(mainSortIndex, outerBounds)
                
            elif type(info) == type({}):
                outerCoords = info["outer"]
                innerCoords = info["inner"]
                
                for curve in outerCoords:
                    if '2D' in mainSortIndex.baseSort().characteristic.type:
                        v1 = vector2D(curve[0], curve[1])
                       
                    elif '3D' in mainSortIndex.baseSort().characteristic.type:
                        if len(curve) == 2:
                            v1 = vector3D(curve[0], curve[1], 0)
                        else:
                            v1 = vector3D(curve[0], curve[1], curve[2]) 

                    outerBounds.append(v1)
    
                outerBounds.append(outerBounds[0])
                pOut = mainSortIndex.baseSort().characteristic(mainSortIndex, outerBounds)
                
                if len(innerCoords) > 0:
                    for curveSet in innerCoords:
                        oneBound = []
    
                        for curve in curveSet:
                            if '2D' in mainSortIndex.baseSort().characteristic.type:
                                v1 = vector2D(curve[0], curve[1])
                                    
                            elif '3D' in mainSortIndex.baseSort().characteristic.type:
                                if len(curve) == 2:
                                    v1 = vector3D(curve[0], curve[1], 0)
                                else:
                                    v1 = vector3D(curve[0], curve[1], curve[2])     
                            oneBound.append(v1)
                        
                        oneBound.append(oneBound[0])
                        innerBounds.append(oneBound)

                    if len(innerBounds) > 0:
                        for bound in innerBounds:
                            pIn.append(mainSortIndex.baseSort().characteristic(mainSortIndex, bound))

            # Sets value of individual to pOut for attribute adding
            individual = pOut
            tempHoldForm = individual.sort.newForm()
            tempHoldForm.insert(individual)
            
            # if len(pIn)>0, this means that there are holes in the plane segment
            if len(pIn) > 0:
                form2 = individual.sort.newForm()
                for obj in pIn:    
                    form2.insert(obj)
                    
                # Subtract form2 planes from form1
                tempHoldForm.difference(form2)

            # Checks if attributes are present
            if len (attributes) > 0:
                att = []
                
                for attrib in attributes:
                    att.append(self.makeAttribute(self.planeSegment[sortName], attrib, attributes[attrib], individual))

                # Inserts attributes into sortal individual if any
                if len(att) != 0:
                    # Problem with planeSeg2D vs planeSeg3D (former is not areal, does not require as many breakdown statements)
                    
                    for attribute in att:
                        for a in attribute:
                            if '2D' in tempHoldForm.sort.baseSort().characteristic.type:
                                tempHoldForm.elements[0].addAttribute(a)
                            elif '3D' in tempHoldForm.sort.baseSort().characteristic.type:
                                tempHoldForm.elements[0].elements[0].addAttribute(a)
                
            if sortName not in self.individuals:
                self.individuals[sortName] = []
                
            self.individuals[sortName].append(tempHoldForm)
        
        return
        
    def makeCircle(self, sortName, mainSortIndex, circles):
        for obj in circles:
            info = obj[0]
            attributes = obj[1]
            individual = None
            center = info[0]    # The center coordinates must always be the first item
            
            if len(center) == 2:
                center_vec = vector2D(center[0], center[1])
            elif len(center) == 3:
                center_vec = vector3D(center[0], center[1], center[2])
                
            # Creates circle individuals
            # For 2D/3D circles
            # (center, radius)
            if len(info) == 2:
                radius = info[1]
                # Generates random angle to generate two points on circle
                angle1 = math.random()*math.PI*2
                angle2 = math.random()*math.PI*2
                # First point
                x1 = math.cos(angle1)*radius
                y1 = math.sin(angle1)*radius
                # Second point
                x2 = math.cos(angle2)*radius
                y2 = math.sin(angle2)*radius
                
                # 2D Circles
                if len(center) == 2:
                    # [ Ax * (By - Cy) + Bx * (Cy - Ay) + Cx * (Ay - By) ] / 2
                    triArea = (x1 * (y2 - center_vec[1]) + x2 * (center_vec[1] - y1) + center_vec[0] * (y1 - y2))/ 2
                    if triArea > 0:
                        pt1 = vector3D(x1, y1)
                        individual = mainSortIndex.baseSort().characteristic(mainSortIndex, center_vec, pt1)
                
                # 3D Circles
                elif len(center) == 3:
                    triArea = (x1 * (y2 - center_vec[1]) + x2 * (center_vec[1] - y1) + center_vec[0] * (y1 - y2))/ 2
                    if triArea > 0:
                        pt1 = vector3D(x1, y1,) # what goes here)
                        pt2 = vector3D(x2, y2,) # what goes here)
                        individual = mainSortIndex.baseSort().characteristic(mainSortIndex, center_vec, pt1, pt2)
            
            # 2D Circles
            # (center, radius, point on circle)
            elif len(info) == 3:
                pt1 = vector2D(info[2][0], info[2][1]) 
                individual = mainSortIndex.baseSort().characteristic(mainSortIndex, center_vec, pt1)
            
            # For 3D circles
            # (center, radius, point1 on circle, point2 on circle)
            elif len(info) == 4 or len(info) == 5:
                radius = info[1]
                if len(info[2]) == 2:
                    pt1 = vector2D(info[2][0], info[2][1])
                elif len(center) == 3:
                    pt1 = vector3D(info[2][0], info[2][1], info[2][2])

                if len(info[3]) == 2:
                    pt2 = vector2D(info[3][0], info[3][1])
                elif len(center) == 3:
                    pt2 = vector3D(info[3][0], info[3][1], info[3][2])
                    
            individual = mainSortIndex.baseSort().characteristic(mainSortIndex, center_vec, pt1, pt2)

            # Checks if attributes are present
            if len (attributes) > 0:
                att = []
                
                for attrib in attributes:
                    att.append(self.makeAttribute(self.circle[sortName], attrib, attributes[attrib], individual))
                    
                # Inserts attributes into sortal individual if any
                if len(att) != 0:
                    for attribute in att:
                        for a in attribute:
                            individual.addAttribute(a)
            
            if sortName not in self.individuals:
                self.individuals[sortName] = []
                
            self.individuals[sortName].append(individual)
            
        return
    
    def makeEllipse(self, sortName, mainSortIndex, ellipses):
        for obj in ellipses:
            info = obj[0]
            attributes = obj[1]
            individual = None
            foc1 = info[1]    # The foci are always the first two items after the center
            foc2 = info[2]    
            pin_pt = info[3]
            
            if len(foc1) == 2:
                foc1_vec = vector2D(foc1[0], foc1[1])
                foc2_vec = vector2D(foc2[0], foc2[1])
                pin_pt_vec = vector2D(pin_pt[0], pin_pt[1])
            elif len(foc1) == 3:
                foc1_vec = vector3D(foc1[0], foc1[1], foc1[2])
                foc2_vec = vector3D(foc2[0], foc2[1], foc2[2])
                pin_pt_vec = vector3D(pin_pt[0], pin_pt[1], pin_pt[2])
                
            individual = mainSortIndex.baseSort().characteristic(mainSortIndex, foc1_vec, foc2_vec, pin_pt_vec)

            # Checks if attributes are present
            if len (attributes) > 0:
                att = []
                
                for attrib in attributes:
                    att.append(self.makeAttribute(self.ellipse[sortName], attrib, attributes[attrib], individual))
                    
                # Inserts attributes into sortal individual if any
                if len(att) != 0:
                    for attribute in att:
                        for a in attribute:
                            individual.addAttribute(a)
            
            if sortName not in self.individuals:
                self.individuals[sortName] = []
                
            self.individuals[sortName].append(individual)
            
        return
    
    def determineAxes(self, foc1, foc2, pt_a, pt_b):
        if foc1[0] == foc2[0]:
            other_pt = pt_a
            if foc1[0] == pt_a[0]:
                other_pt = pt_b
        
        elif foc1[1] == foc1[1]:
            other_pt = pt_a
            if foc1[1] == pt_a[1]:
                other_pt = pt_b
            
        return other_pt
    
    def makeCArc(self, sortName, mainSortIndex, cArcs):
        for obj in cArcs:
            info = obj[0]
            attributes = obj[1]
            individual = None
            center = info[0]       # The center point is always the first item
            edge_pt1 = info[1]     # The edge pts (start and end of the arc) are always the next two items
            edge_pt2 = info[2]     
            pin_pt = info[3]       # The pin pt is always the last item
            
            if len(info[0]) == 2:
                center_vec = vector2D(center[0], center[1])
                edge_pt1_vec = vector2D(edge_pt1[0], edge_pt1[1])
                edge_pt2_vec = vector2D(edge_pt2[0], edge_pt2[1])
                pin_pt_vec = vector2D(pin_pt[0], pin_pt[1])
                
            elif len(info[0]) == 3:
                center_vec = vector3D(center[0], center[1], center[2])
                edge_pt1_vec = vector3D(edge_pt1[0], edge_pt1[1], edge_pt1[2])
                edge_pt2_vec = vector3D(edge_pt2[0], edge_pt2[1], edge_pt2[2])
                pin_pt_vec = vector3D(pin_pt[0], pin_pt[1], pin_pt[2])
            
            individual = mainSortIndex.baseSort().characteristic(mainSortIndex, center_vec, edge_pt1_vec, edge_pt2_vec, pin_pt_vec)

            # Checks if attributes are present
            if len (attributes) > 0:
                att = []
                
                for attrib in attributes:
                    att.append(self.makeAttribute(self.cArc[sortName], attrib, attributes[attrib], individual))
                    
                # Inserts attributes into sortal individual if any
                if len(att) != 0:
                    # Problem with planeSeg2D vs planeSeg3D (former is not areal, does not require as many breakdown statements)
                    
                    for attribute in att:
                        for a in attribute:
                            individual.addAttribute(a)
            
            if sortName not in self.individuals:
                self.individuals[sortName] = []
                
            self.individuals[sortName].append(individual)
            
        return
        
    def makeEArc(self, sortName, mainSortIndex, eArcs):
        for obj in eArcs:
            info = obj[0]
            center = info[0]
            f1 = info[1] 
            f2 = info[2]
            start = info[3]
            end = info[4]
            pin_pt = info[5]
            attributes = obj[1]
            individual = None
            
            # If minor axis point is available for reference
            if len(info) > 6:
                short_pt = info[6]
            else:
                short_pt = None
                
            if len(info[0]) == 2:
                f1_vec = vector2D(f1[0], f1[1])
                f2_vec = vector2D(f2[0], f2[1])
                start_vec = vector2D(start[0], start[1])
                end_vec = vector2D(end[0], end[1])
                pin_pt_vec = vector2D(pin_pt[0], pin_pt[1])
                short_pt_vec = vector2D(short_pt[0], short_pt[1])
                
            elif len(info[0]) == 3:
                f1_vec = vector3D(f1[0], f1[1], f1[2])
                f2_vec = vector3D(f2[0], f2[1], f2[2])
                start_vec = vector3D(start[0], start[1], start[2])
                end_vec = vector3D(end[0], end[1], end[2])
                pin_pt_vec = vector3D(pin_pt[0], pin_pt[1], pin_pt[2])
                short_pt_vec = vector3D(short_pt[0], short_pt[1], short_pt[2])
                
            """
            rs.AddTextDot('center', (center[0], center[1], center[2]))
            rs.AddTextDot('f1', (f1[0], f1[1], f1[2]))
            rs.AddTextDot('f2', (f2[0], f2[1], f2[2]))
            rs.AddTextDot('start', (start[0], start[1], start[2]))
            rs.AddTextDot('end', (end[0], end[1], end[2]))
            rs.AddTextDot('pin', (pin_pt[0], pin_pt[1], pin_pt[2]))
            """
            
            if sortName not in self.individuals:
                self.individuals[sortName] = []

            if short_pt != None:
                individual = mainSortIndex.baseSort().characteristic(mainSortIndex, f1_vec, f2_vec, start_vec, end_vec, pin_pt_vec, short_pt_vec)
            else:
                individual = mainSortIndex.baseSort().characteristic(mainSortIndex, f1_vec, f2_vec, start_vec, end_vec, pin_pt_vec)
                
            # Checks if attributes are present
            if len (attributes) > 0:
                att = []
                
                for attrib in attributes:
                    att.append(self.makeAttribute(self.eArc[sortName], attrib, attributes[attrib], individual))
                    
                # Inserts attributes into sortal individual if any
                if len(att) != 0:
                    # Problem with planeSeg2D vs planeSeg3D (former is not areal, does not require as many breakdown statements)
                    
                    for attribute in att:
                        for a in attribute:
                            individual.addAttribute(a)
            
            if sortName not in self.individuals:
                self.individuals[sortName] = []
                
            self.individuals[sortName].append(individual)
            
        return
        
    def makeBezier(self, sortName, mainSortIndex, beziers):
        for obj in beziers:
            info = obj[0]
            attributes = obj[1]
            individual = None
            start = info[0] # The end pts of the bezier curve are always the first and last info bits
            ctrl = info[1]  # The ctrl point of the bezier curve is always index 1, the middle bit
            end = info[2]
            
            if len(info[0]) == 2:
                start_vec = vector2D(start[0], start[1])
                ctrl_vec = vector2D(ctrl[0], ctrl[1])
                end_vec = vector2D(end[0], end[1])
                
            elif len(info[0]) == 3:
                start_vec = vector3D(start[0], start[1], start[2])
                ctrl_vec = vector3D(ctrl[0], ctrl[1], ctrl[2])
                end_vec = vector3D(end[0], end[1], end[2])
                
            individual = mainSortIndex.baseSort().characteristic(mainSortIndex, start_vec, ctrl_vec, end_vec)
            
            # Checks if attributes are present
            if len (attributes) > 0:
                att = []
                
                for attrib in attributes:
                    att.append(self.makeAttribute(self.bezier[sortName], attrib, attributes[attrib], individual))
                    
                # Inserts attributes into sortal individual if any
                if len(att) != 0:
                    # Problem with planeSeg2D vs planeSeg3D (former is not areal, does not require as many breakdown statements)
                    
                    for attribute in att:
                        for a in attribute:
                            individual.addAttribute(a)
            
            if sortName not in self.individuals:
                self.individuals[sortName] = []
                
            self.individuals[sortName].append(individual)
            
        return
         
    def makeDescription(self, sortType, descriptions):
        # Converts description text to description individuals
        for count, individual in enumerate(descriptions):
            individual.replace('\u201d', '"')
            individual.replace('\u201c', '"')
            self.individuals['descriptions'].append(description(sort.register[sortType], '`' + individual + '`'))
        
        return
    
    def makeAttribute(self, geomDict, attName, attribute, individual = None):
        att = []
        # labels 
        if attName == 'labels' and 'label' in geomDict:
            if len(attribute) > 0: 
                for content in attribute:
                    att.append(label(geomDict['label'], content))
        
        # descriptions
        elif attName == 'label_descriptions':
            if len(attribute) > 0: 
                for content in attribute:
                    if content != '':
                        att.append(description(geomDict['label_descriptions'], '`'+content+'`'))

        # enums
        elif attName == 'enums' and 'enum' in geomDict:
            if len(attribute) > 0: 
                for content in attribute:
                    att.append(enumerative(geomDict['enum'], content))
        
        # weight - line width
        elif attName == 'weight' and 'weight' in geomDict: 
            att.append(rWeight(geomDict['weight'], attribute))
        
        # weight 255 - white to black (grayscale)
        elif attName == 'weight255' and 'weight255' in geomDict:  
            if attribute[0] != 0 and attribute[0] != 0.0:
                att.append(rWeight(geomDict['weight255'], attribute[0]))
        
        # color - rgb values
        elif attName == 'color' and 'color' in geomDict:
            rgb = attribute
            att.append(color(geomDict['color'], int(rgb[0]), int(rgb[1]), int(rgb[2])))
       
        elif attName == 'tags':
            tagNames = attribute
            for tagName in tagNames:
                keyIndex = None
                
                if 'void' not in tagName:
                    if type(individual) == type([]):
                        keyIndex = []
                        count = 0
                        for indiv in individual:
                            if len(name) > 2:
                                newName = tagName + '_' + count
                                keyIndex.append(key.createKey(newName, indiv))
                                count+=1
                            else:
                                key.createKey(tagName, individual)
                    else:
                        keyIndex = key.createKey(tagName, individual)
                        att.append(keyIndex)
        return att
       
    """
    =================
    CREATING METAFORM
    =================
    """
    
    # Creates forms to be inserted into metaform
    def makeForm(self):
        forms = []

        for geomName in self.individuals:
            if geomName != 'descriptions':
                # Creates one form per sublist in individuals dictionary
                geometry = self.individuals[geomName]
                newForm = sort.register[geomName].newForm() 
                
                # Inserts individuals into form
                for object in geometry: 
                    # If there are several individuals, then treats it as a list
                    if type(object) == type([]):                
                        for obj in object:
                            newForm.insert(obj)
                            
                    # Check if attribute sort or if already form (for plane segments, mostly)
                    #elif 'ind' in object.type:
                    else:
                        # If there is only one individual, then it is inserted into the form
                        newForm.insert(object)

                forms.append(newForm)     # Appends finished subform to list of forms that will make up the metaform

        formM1 = create.metaForm(self.mainSort, forms)           # Creates metaform and inserts forms into metaform     
        formM1.maximalize()                                      # Maximalizes form geometry
       
        # Inserts description individuals into metaform
        for indiv in self.individuals['descriptions']:
            formM1.insert(indiv)
 
        return formM1
    
    """
    ======================
    CREATING COMPOUND FORM
    ======================
    """
    
    # Creates and returns compound metaform based on compound sort in sort.register and inputted disjunctive metaforms
    def createCompoundForm(self, forms):
        # Checks if rhino_shapes_c is already in sort.register
        # Obtains 'rhino_shapes_c' sort type
        if 'rhino_shapes_c' in sort.register:
            compSort = sort.register['rhino_shapes_c']
                    
        return create.metaFormC(compSort, forms)           # Creates metaform and inserts forms into metaform
        
   
    


