#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# Sortal library imports
from sortal import sort, rule, sdlParser, form

# API support code imports
from .convert.convert_sortal_to_agnostic import convertS
from .convert.visualAgnostic import visualA
from .move import translation_vector

# Conversion class initialization
convertS = convertS()

"""
=============
FUNCTION INFO
=============
'draw_rule' draws the sides of a rule in the Rhino workspace. This function supposes that the rule is already present in the rule register.
The rule sides are drawn apart from each other, with the space between them
dictated by the value of 'fixed_factor' or the size of the bounding box of the LHS GUIDs.
When a reference point is given (tuple/list/vector/point), the rule shapes are moved to the reference point. 
The default reference point is the origin (0,0,0). 

SUPPORT FUNCTIONS
convertS: Converts a sortal shape to an agnostic dictionary to serve as input for the drawing function, visualA
visualA: Draws an agnostic dictionary in the Rhino workspace
translation_vector: Moves shape to frame of reference with refPt input as origin

INPUT
ruleName: Name of rule object to draw from inside the rule register (text string)
refPt: Reference point (optional - tuple / list of integers or floats / vectors / points (3Ds)); relocates
       drawings of rule sides with reference point serving as the origin; default origin is (0,0,0)

OUTPUT
[lhsRhino, rhsRhino] or [lhsMoved, rhsMoved]: List of lists of Rhino Geometry [[lhs shape GUIDs], [rhs shape GUIDs]]; an empty list is
                                              returned if the rule cannot be found in the rule register

WARNINGS / ERRORS
KeyError: Rule name does not exist in the rule register
          OR if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: Rule name input is not a text string

Data Flow: rule name in rule register  
           -> rule object retrieved from SDL 
           -> rule sides converted to agnostic dictionaries
           -> agnostic dictionary of rule 
           -> lists of Rhino GUIDs (visible in Rhino viewport) stored in one list and returned
"""

def draw_rule(ruleName, refPt = r.Geometry.Point3d(0,0,0)):
    if 'rhino_shapes_c' in sort.register:
        translateL = None
        translateR = None
        
        if type(ruleName) == type('a'):
            if ruleName in rule.register:
                ruleLHS = rule.register[ruleName].lhs
                ruleRHS = rule.register[ruleName].rhs
                
                LHSagnostic = convertS.toAgnostic(ruleLHS)
                RHSagnostic = convertS.toAgnostic(ruleRHS)
        
                visLHS = visualA(LHSagnostic)           
                visRHS = visualA(RHSagnostic)
        
                lhsRhino = visLHS.plotAgnostic()
                rhsRhino = visRHS.plotAgnostic()
                
                bboxL = rs.BoundingBox(lhsRhino)
                bboxR = rs.BoundingBox(rhsRhino)
                
                start_distance_xL = (bboxL[1].X- bboxL[0].X)
                start_distance_xR = (start_distance_xL + (bboxR[1].X- bboxR[0].X))
        
                if start_distance_xR == 0.0:
                    start_distance_xR = sc.sticky['fixed_factor']
        
                if type(refPt) == type(()):
                    if len(refPt) == 3:
                        translateL = r.Geometry.Vector3d(refPt[0] - bboxL[0].X, refPt[1] - bboxL[0].Y, refPt[2] - bboxL[0].Z)
                        translateR = r.Geometry.Vector3d(refPt[0] + start_distance_xR - bboxR[0].X, refPt[1] - bboxR[0].Y, refPt[2] - bboxR[0].Z)
                        
                elif type(refPt) == r.Geometry.Vector3d:
                    translateL = r.Geometry.Vector3d(refPt.X + start_distance_xL - bboxL[0].X , refPt.Y - bboxL[0].Y, refPt.Z - bboxL[0].Z)
                    translateR = r.Geometry.Vector3d(refPt.X + start_distance_xR - bboxR[0].X, refPt.Y - bboxR[0].Y, refPt.Z- bboxR[0].Z)
                    
                else:
                    try:
                        if rs.ObjectType(refPt) == 1:
                            coords = rs.PointCoordinates(refPt)
                            translateL = r.Geometry.Vector3d(coords.X - bboxL[0].X, coords.Y - bboxL[0].Y, coords.Z - bboxL[0].Z)
                            translateR = r.Geometry.Vector3d(coords.X + start_distance_xR - bboxR[0].X, coords.Y - bboxR[0].Y, coords.Z - bboxR[0].Z)
                    except:
                        raise TypeError('Reference point input is not a Rhino point, tuple, list or Rhino geometry.')
        
                if translateL != None and translateR != None:
                    LHSmoved = rs.MoveObjects(lhsRhino, translateL)
                    RHSmoved = rs.MoveObjects(rhsRhino, translateR)
                    return [LHSmoved, RHSmoved]
                    
                else:
                    return [lhsRhino, rhsRhino]
            else:
                warnings.warn('Rule does not exist in the rule register.') 
                return []
        else:
            raise TypeError('Input for rule name is not a text string.')
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
