#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r

# Sortal library imports
from sortal import sort, rule, flow, form, sdlParser

# Support function import
from .read_sdl import update_sort

# Python imports
from copy import deepcopy
import os
import warnings

"""
=============
FUNCTION INFO
=============
'overwrite_sdl' stores rules, flows and an initial shape in an sdl file based on the variable fileName, by accepting their name references,
and retrieving the relevant sortal objects. If an sdl file with the same name already exists, 
then the user is prompted if they want to overwrite the file or give a new value to fileName.
The input for 'fileName' can also include the address of the intended location.

SUPPORT FUNCTIONS
determine_overwrite: Determines which rules, shapes or flows are already present in the SDL file and will be overwritten, and which of these will be 
                     newly added to the SDL file

INPUT
fileName: Text string of file name (for .sdl file; this may include the intended location's full address)
rules: List of sortal rule object names (optional, this may be an empty list)
shape: One shape object name (optional, this may be ignored)
flows: List of sortal flow object names (optional, this may be an empty list)

Note: If including the full address of the target SDL file, the following format must be observed:
     'C:\\Users\AKIDRIBM\\AppData\\Roaming\\Grasshopper\\Libraries\\source_code_gh_dev\\'+ <sdl file name> +'.sdl'
     where every slash is doubled.

OUTPUT
If successful, an sdl file of name 'fileName' in the intended location with rule(s) and shape stored in it is created, and True is returned. Otherwise -
False: Unsuccessful operation

WARNINGS/ERRORS
Warning: If a rule, form (shape) or flow does not exist in their corresponding register, or if the SDL file with the target file name does not exist,
         or if the SDL file name does not have the '.sdl' extension
TypeError: If the input for rules, shapes, flows are not lists, or if there is an invalid input (not a text string) within the 
           lists
KeyError: If a rule, form or flow name within a list does not exist in the corresponding register
          or if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up

Data Flow: file name, list of rule names, list of shape names, list of flow names, append (Boolean value), prnt (Boolean value)
           -> checking validity of input types
           -> checking pre-existence of SDL file with the same name
           -> if it exists, retrieval of sortal objects corresponding to rule names, shape name, flow names (otherwise, the function is exited)
           -> Clearing of sortal registers and SDL Parser registers to make way for file to be overwritten
           -> Determining which rules/shapes/forms are to be overwritten and which ones are to be added to the file
           -> Overwriting of SDL file
"""

def overwrite_sdl(fileName, rules = [], shapes = [], flows = [], append = False, prnt = True):      
    if 'rhino_shapes_c' in sort.register:
        if type(rules) == type([]) and type(shapes) == type([]) and type(flows) == type([]) and type(append) == type(False) and type(prnt) == type(False):
            if len(rules) > 0 or len(shapes) > 0 or len(flows) > 0:
                # Checks validity of file name
                if '.sdl' in fileName:
                    # Checks if file exists
                    if os.path.exists(fileName):
                        # Retrieval of sortal data structures
                        rule_objects = []
                        shape_objects = []
                        flow_objects = []
                        
                        for ruleName in rules:
                            if type(ruleName) == type('a'):
                                if ruleName in rule.register:
                                    rule_objects.append(rule.register[ruleName])
                                else:
                                    warnings.warn(ruleName + ' does not exist in rule register.')
                            else:
                                try:
                                    if ruleName.type == 'rule':
                                        rule_objects.append(ruleName)
                                except:
                                    raise TypeError(('Invalid rule input:', ruleName))
                        
                        for shapeName in shapes:
                            if type(shapeName) == type('a'):
                                if shapeName in form.formRegister:
                                    shape_objects.append(form.formRegister[shapeName])
                                else:
                                    if shapeName == '':
                                        pass
                                            
                                    else:
                                        warnings.warn(shapeName + ' does not exist in form register.')
                            else:
                                try: 
                                    if 'form' in shapeName.type:
                                        shape_objects.append(shapeName)
                                except:
                                    raise TypeError(('Invalid shape input:', shapeName))
                         
                        for flowName in flows:
                            if type(flowName) == type('a'):
                                if flowName in flow.register:
                                    flow_objects.append(flow.register[flowName])
                                else:
                                    warnings.warn(flowName + ' does not exist in the flow register.')
                            else:
                                try:
                                    if flowName.type == 'flow':
                                        flow_objects.append(flowName)
                                except:
                                    raise TypeError(('Invalid flow input', flowName))
                        
                        # Stores registier and SDL Parser dictionary information prior to opening the target file
                        temp_rule_dict = {}
                        temp_shape_dict = {}
                        temp_sort_dict = {}
                        temp_flow_dict = {}
                        temp_sdl_rule_dict = {}
                        temp_sdl_shape_dict = {}
                        temp_sdl_flow_dict = {}
                        
                        for ruleName in rule.register:
                            temp_rule_dict[ruleName] = deepcopy(rule.register[ruleName])
                    
                        for formName in form.formRegister:
                            temp_shape_dict[formName] = form.formRegister[formName].duplicate()
                        
                        for sortType in sort.register:
                            temp_sort_dict[sortType] = deepcopy(sort.register[sortType])
                            
                        for flowItem in flow.register:
                            temp_flow_dict[flowItem] = deepcopy(flow.register[flowItem])
                            
                        for ruleName in sdlParser.rules:
                            temp_sdl_rule_dict[ruleName] = deepcopy(sdlParser.rules[ruleName])
                            
                        for shapeName in sdlParser.forms:
                            temp_sdl_shape_dict[shapeName] = deepcopy(sdlParser.forms[shapeName])

                        for flowName in sdlParser.flows:
                            temp_sdl_flow_dict[flowName] = deepcopy(sdlParser.flows[flowName])
                            
                        rule.register.clear()
                        sort.cleanup()
                        form.formRegister = {}
                        flow.register.clear()
                        sdlParser.rules.clear()
                        sdlParser.forms.clear()
                        sdlParser.flows.clear()
                        written = True
                        
                        # Opens SDL file to be overwritten
                        sdlParser.parse(fileName)
                        
    
                        if append:
                            over_rules_list = []
                            over_shapes_list = []
                            over_flows_list = []
                            add_rules_list = []
                            add_shapes_list = []
                            add_flows_list = []
                            final_rules_list = []
                            final_shapes_list = []
                            final_flows_list = []
                            # Determining which rules are to be overwritten, added to SDL and which are kept from the previous SDL file version
                            over_rules_list, add_rules_list, final_rules_list = determine_overwrite(rule_objects, sdlParser.rules)
                             
                            # Determining which shapes are to be overwritten, added to SDL and which are kept from the previous SDL file version
                            over_shapes_list, add_shapes_list, final_shapes_list = determine_overwrite(shape_objects, sdlParser.forms)
                                    
                            # Determining which flows are to be overwritten, added to SDL and which are kept from the previous SDL file version
                            over_flows_list, add_flows_list, final_flows_list = determine_overwrite(flow_objects, sdlParser.flows)

                            if len(final_rules_list) > 0 or len(final_shapes_list) > 0 or len(final_flows_list) > 0:
                                sdlParser.writeGrammar(fileName, final_rules_list, final_shapes_list, final_flows_list)
                                
                                if prnt:
                                    print (fileName + ' has been overwritten.')
                            
                                    if len(over_rules_list) > 0:
                                        print ('The following rule/s was/were overwritten in ' + fileName + ':')
                                        print (over_rules_list)
                                        
                                    if len(add_rules_list) > 0:
                                        print ('The following rule/s was/were added to ' + fileName + ':')
                                        print (add_rules_list)
                                        
                                    if len(over_shapes_list) > 0:
                                        print ('The following shape/s was/were overwritten in ' + fileName + ':')
                                        print (over_rules_list)
                                        
                                    if len(add_shapes_list) > 0:
                                        print ('The following shape/s was/were added to ' + fileName + ':')
                                        print (add_shapes_list)
                
                                    if len(over_flows_list) > 0:
                                        print ('The following flow/s was/were overwritten in ' + fileName + ':')
                                        print (over_rules_list)
                                        
                                    if len(add_flows_list) > 0:
                                        print ('The following flow/s was/were added to ' + fileName + ':')
                                        print (add_flows_list)
                                
                            else:
                                warnings.warn(fileName + ' was not overwritten. No rules/shapes/flows were added/overwritten. Inputs for rule lists/shape lists/flow lists were empty.')
                                written = False
                        else:
                            if len(rule_objects) > 0 or len(shape_objects) > 0 or len(flow_objects) > 0:
                                sdlParser.writeGrammar(fileName, rule_objects, shape_objects, flow_objects)
                                if prnt:
                                    print (fileName + ' has been overwritten.')
                            else:
                                warnings.warn(fileName + ' was not overwritten. Inputs for rule lists/shape lists/flow lists were empty.')
                                written = False
                                
                        # Restores previous conditions of rule register, sort register, form register, flow register, and SDL Parser dictionaries
                        rule.register.clear()
                        form.formRegister = {}
                        sort.cleanup()
                        flow.register.clear()
                        sdlParser.rules.clear()
                        sdlParser.forms.clear()
                        sdlParser.flows.clear()
                        
                        # Rule, form, sort and flow registers
                        for ruleName in temp_rule_dict:
                            rule.register[ruleName] = deepcopy(temp_rule_dict[ruleName])
                            
                        for formName in temp_shape_dict:
                            form.formRegister[formName] = temp_shape_dict[formName].duplicate()
                        
                        for sortType in temp_sort_dict:
                            sort.register[sortType] = deepcopy(temp_sort_dict[sortType])
                            
                        for flowItem in temp_flow_dict:
                            flow.register[flowItem] = deepcopy(temp_flow_dict[flowItem])
                        
                        # SDL rule, form and flow dictionaries
                        for ruleName in temp_sdl_rule_dict:
                            sdlParser.rules[ruleName] = deepcopy(temp_sdl_rule_dict[ruleName])
                            
                        for formName in temp_sdl_shape_dict:
                            sdlParser.forms[formName] = temp_sdl_shape_dict[formName].duplicate()
                        
                        for flowName in temp_sdl_flow_dict:
                            sdlParser.flows[flowName] = deepcopy(temp_sdl_flow_dict[flowName])
                        
                        return written
                        
                    else:
                        warnings.warn('SDL file does not exist. Exiting function')
                        return False
                else:
                    warnings.warn('.sdl extension not present in file name. Exiting function.')
                    return False
            else:
                raise KeyError('Rule, shape, and flow list inputs are all empty.')
        else:
            if type(rules) != type([]):
                raise TypeError('Rules input is not a list.')
                
            if type(flows) != type([]):
                raise TypeError('Flows input is not a list.')
                
            if type(shape) != type([]):
                raise TypeError('Shapes input is not a list.')
                
            if type(append) != type(False):
                raise TypeError('Input for append variable is not a Boolean value.')
                
            if type(prnt) != type(False):
                raise TypeError('Input for append variable is not a Boolean value.')
    else:
        raise KeyError("Compound sort type 'rhino_shapes_c' not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named 'rhino_shapes_c' to ensure compatibility with the API.")
        
"""
=====================
SUPPORT FUNCTION INFO
=====================
'determine_overwrite' adds the predicate (max line, shortest line, void, bound line, longest line, no label) to the LHS of the rule object.

INPUT
Sortal rule object, predicate coordinates and types (agnostic), classification (parametric, dimension)

OUTPUT
Successfully added predicates, if successful

Data Flow: sortal rule object, predicates dictionary, classification of agnostic object -> rule object with predicates inside
"""

def determine_overwrite(new_items, old_items):
    overwrite = []
    add = []
    final = []
    
    for new_item in new_items:
        if new_item.name in old_items:
            overwrite.append(new_item.name)
        else:
            add.append(new_item.name)

        final.append(new_item)

    for old_item in old_items:
        if old_items[old_item] not in final:
            final.append(old_items[old_item])

    return overwrite, add, final