#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# Sortal library imports
from sortal import sort, form, rule, flow

"""
=============
FUNCTION INFO
=============
'set_rule_name' changes the name of a rule object in the sortal rule register; however, if the new rule name matches that of a pre-existing rule, 
this will cause an error. If successful, the previous rule name is deleted from the sortal library's rule register
and is free to use for new rules.

INPUT
oldName: Name of target rule object to be altered (text string)
newName: New name of target rule object (text string)

OUTPUT
True: If successful
False: If unsuccessful

WARNINGS / ERRORS
KeyError: If the original rule name does not exist in the sort register, or if the new rule name already exists in the sort register
          or if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: If the input for the original rule name or the new rule name is not a text string
    
Data Flow: rule object -> altered rule object -> success of operation is returned
"""
    
def set_rule_name(oldName, newName):
    if 'rhino_shapes_c' in sort.register:
        if type(newName) == type('a') and type(oldName) == type('a'):
            if oldName in rule.register and newName not in rule.register:
                rule.register[oldName].setName(newName)
                return True
            else:
                if oldName not in rule.register:
                    raise KeyError('Rule named ' + oldName + ' does not exist in the rule register')
               
                if newName in rule.register:
                    raise KeyError('Rule name ' + newName + ' already exists in the rule register')
        else:
            if type(oldName) != type('a'):
                warnings.warn('Input for rule to be changed is not a string')
            if type(newName) != type('a'):
                warnings.warn('Input for new rule name is not a string')
                
        return False
   
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
        
"""
=============
FUNCTION INFO
=============
'set_rule_description' changes the description of a rule object in the sortal rule register. 

INPUT
ruleName: Name of rule object (text string)
newDesc: New description for rule object (text string)

OUTPUT
True: If successful
False: If unsuccessful

WARNINGS / ERRORS
KeyError: If the original rule name does not exist in the sort register
Error: If 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: If the input for the new description or the original rule name is not a text string
    
Data Flow: rule object -> altered rule object
"""     
   
def set_rule_description(ruleName, newDesc):
    if 'rhino_shapes_c' in sort.register:
        if type(newDesc) == type('a') and type(ruleName) == type('a'):
            if ruleName in rule.register:
                rule.register[ruleName].setDescription(newDesc)
                return True
            else:
                raise KeyError('Rule named ' + ruleName + ' does not exist in the rule register')
        else:
            if type(ruleName) != type('a'):
                warnings.warn('Input for rule to be changed is not a string')
            if type(newDesc) != type('a'):
                warnings.warn('Input for new rule description is not a string')
                
        return False
    
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                

"""
=============
FUNCTION INFO
=============
'set_shape_name' changes the name of a shape in the sortal form register. 

INPUT
shapeName: Name of shape object to be changed (text string)
newShapeName: New name of target shape object (text string)

OUTPUT
True: If successful
False: If unsuccessful

WARNINGS / ERRORS
KeyError: If the original shape name does not exist in the sort register
Error: If 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: If the input for the original shape name or the new shape name is not a text string    
    
Data Flow: rule object -> altered rule object
"""   

def set_shape_name(shapeName, newShapeName):
    if 'rhino_shapes_c' in sort.register:
        if type(shapeName) == type('a') and type(newShapeName) == type('a'):
            if shapeName in form.formRegister:
                shapeObj = form.formRegister[shapeName]
                shapeObj.register(newShapeName)
                return shapeObj
            else:
                raise KeyError('Shape named ' + shapeName + ' does not exist in the rule register')
                
        else:
            if type(shapeName) != type('a'):
                warnings.warn('Shape name is not a text string')
                
            if type(newShapeName) != type('a'):
                warnings.warn('New shape name is not a text string')
    
        return False
   
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                

"""
=============
FUNCTION INFO
=============
'set_flow_name' changes the name of a flow object in the sortal flow register; however, if the new flow name matches that of a pre-existing flow, 
this will cause an error. If successful, the previous flow name is deleted from the sortal library's flow register
and is free to use for new flows.

INPUT
oldName: Name of target flow object to be altered (text string)
newName: New name of target flow object (text string)

OUTPUT
True: If successful
False: If unsuccessful

WARNINGS / ERRORS
KeyError: If the original flow name does not exist in the sort register, or if the new flow name already exists in the sort register
Error: If 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: If the input for the original flow name or the new flow name is not a text string
    
Data Flow: flow object -> altered flow object
"""
    
def set_flow_name(oldName, newName):
    if 'rhino_shapes_c' in sort.register:
        if type(newName) == type('a') and type(oldName) == type('a'):
            if oldName in flow.register and newName not in flow.register:
                flow.register[oldName].setName(newName)
                return True
            else:
                if oldName not in flow.register:
                    raise KeyError('Flow named ' + oldName + ' does not exist in the flow register')
               
                if newName in flow.register:
                    raise KeyError('Flow name ' + newName + ' already exists in the flow register')
        else:
            if type(oldName) != type('a'):
                warnings.warn('Input for flow to be changed is not a string')
            if type(newName) != type('a'):
                warnings.warn('Input for new flow name is not a string')
                
        return False
   
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
        
"""
=============
FUNCTION INFO
=============
set_flow_description changes the description of a flow object in the sortal flow register. 

INPUT
flowName: Name of flow object (text string)
newDesc: New description for flow object (text string)

OUTPUT
True: If successful
False: If successful

WARNINGS / ERRORS
KeyError: If the original flow name does not exist in the sort register
Error: If 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: If the input for the new description or the original flow name is not a text string
    
Data Flow: flow object -> altered flow object
"""     
   
def set_flow_description(flowName, newDesc):
    if 'rhino_shapes_c' in sort.register:
        if type(newDesc) == type('a') and type(flowName) == type('a'):
            if flowName in flow.register:
                flow.register[flowName].setDescription(newDesc)
                return True
            else:
                raise KeyError('Rule named ' + flowName + ' does not exist in the flow register')
        else:
            if type(flowName) != type('a'):
                warnings.warn('Input for flow to be changed is not a string')
            if type(newDesc) != type('a'):
                warnings.warn('Input for new flow description is not a string')
                
        return False
        
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                